---
--- Run as defined rw user on newly created database
--- psql -U cdnskey_processor -d cdnskey_processor -f 0001_schema.sql
---
ALTER DEFAULT PRIVILEGES IN SCHEMA public GRANT SELECT ON TABLES TO cdnskey_processor_ro;

CREATE TABLE IF NOT EXISTS worker
(
    id BIGINT GENERATED ALWAYS AS IDENTITY PRIMARY KEY NOT NULL,
    name TEXT NOT NULL CHECK(name != '' AND name = LOWER(TRIM(name))),
    last_response_at TIMESTAMP,
    removed_at TIMESTAMP
);
CREATE UNIQUE INDEX worker_name_not_removed_idx ON worker(name) WHERE removed_at IS NULL;

CREATE TABLE IF NOT EXISTS domain
(
    id BIGINT GENERATED ALWAYS AS IDENTITY PRIMARY KEY NOT NULL,
    fqdn TEXT NOT NULL CHECK(fqdn != '' AND fqdn = LOWER(TRIM(fqdn))),
    uid TEXT NOT NULL
);
CREATE UNIQUE INDEX domain_fqdn_uid_idx ON domain(fqdn, uid);

CREATE TABLE IF NOT EXISTS nameserver
(
    id BIGINT GENERATED ALWAYS AS IDENTITY PRIMARY KEY NOT NULL,
    fqdn TEXT NOT NULL CHECK(fqdn != '' AND fqdn = LOWER(TRIM(fqdn)))
);
CREATE UNIQUE INDEX nameserver_fqdn_idx ON nameserver(fqdn);

CREATE TABLE IF NOT EXISTS scan_queue
(
    id BIGINT GENERATED ALWAYS AS IDENTITY PRIMARY KEY NOT NULL,
    import_started_at TIMESTAMP NOT NULL DEFAULT (CURRENT_TIMESTAMP),
    import_finished_at TIMESTAMP,
    solving_started_at TIMESTAMP,
    solving_finished_at TIMESTAMP
);
CREATE INDEX scan_queue_import_finished_at_idx ON scan_queue(import_finished_at);

CREATE TABLE IF NOT EXISTS scan
(
    id BIGINT GENERATED ALWAYS AS IDENTITY PRIMARY KEY NOT NULL,
    scan_queue_id BIGINT NOT NULL REFERENCES scan_queue(id),
    domain_id BIGINT NOT NULL REFERENCES domain(id)
);
CREATE UNIQUE INDEX scan_unique_idx ON scan(scan_queue_id, domain_id);
CREATE INDEX scan_domain_id_idx ON scan(domain_id);

CREATE TABLE IF NOT EXISTS insecure_scan
(
    scan_id BIGINT NOT NULL REFERENCES scan(id),
    nameserver_id BIGINT NOT NULL REFERENCES nameserver(id)
);
CREATE UNIQUE INDEX insecure_scan_unique_idx ON insecure_scan(scan_id, nameserver_id);
CREATE INDEX insecure_scan_nameserver_id_idx ON insecure_scan(nameserver_id);

CREATE TABLE IF NOT EXISTS scan_batch
(
    id BIGINT GENERATED ALWAYS AS IDENTITY PRIMARY KEY NOT NULL
);

CREATE TABLE IF NOT EXISTS scan_batch_progress
(
    scan_batch_id BIGINT NOT NULL REFERENCES scan_batch(id),
    worker_id BIGINT NOT NULL REFERENCES worker(id),
    sent_at TIMESTAMP,
    started_at TIMESTAMP,
    finished_at TIMESTAMP
);
CREATE INDEX IF NOT EXISTS scan_batch_progress_finished_at_idx ON scan_batch_progress(finished_at);
CREATE INDEX IF NOT EXISTS scan_batch_progress_scan_batch_id_idx ON scan_batch_progress(scan_batch_id);

CREATE TABLE IF NOT EXISTS scan_job
(
    id BIGINT GENERATED ALWAYS AS IDENTITY PRIMARY KEY NOT NULL,
    scan_batch_id INTEGER NOT NULL REFERENCES scan_batch(id),
    scan_id BIGINT NOT NULL REFERENCES scan(id)
);
CREATE UNIQUE INDEX scan_job_unique_idx ON scan_job(scan_batch_id, scan_id);
CREATE INDEX scan_job_scan_id_idx ON scan_job(scan_id);

CREATE TYPE cdnskey_status AS ENUM
(
    'insecure',
    'insecure-empty',
    'unresolved',
    'secure',
    'secure-empty',
    'untrustworthy',
    'unknown',
    'unresolved-ip'
);

CREATE TABLE IF NOT EXISTS scan_result
(
    id BIGINT GENERATED ALWAYS AS IDENTITY PRIMARY KEY NOT NULL,
    scan_job_id BIGINT NOT NULL REFERENCES scan_job(id),
    worker_id BIGINT NOT NULL REFERENCES worker(id),
    nameserver_id BIGINT REFERENCES nameserver(id),
    nameserver_ip INET,
    scanned_at TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP,
    scan_status cdnskey_status NOT NULL,
    cdnskey_flags INTEGER,
    cdnskey_proto INTEGER,
    cdnskey_alg INTEGER,
    cdnskey_public_key TEXT
);
CREATE INDEX scan_result_scan_job_id_idx ON scan_result(scan_job_id);
CREATE UNIQUE INDEX scan_result_unique_idx_1 ON scan_result(scan_job_id, worker_id, nameserver_id, nameserver_ip)
              WHERE scan_status NOT IN ('insecure'::cdnskey_status, 'secure'::cdnskey_status);
CREATE UNIQUE INDEX scan_result_unique_idx_2 ON scan_result(scan_job_id, worker_id, nameserver_id, nameserver_ip, cdnskey_flags, cdnskey_proto, cdnskey_alg, cdnskey_public_key)
              WHERE scan_status IN ('insecure'::cdnskey_status, 'secure'::cdnskey_status);
CREATE INDEX scan_result_nameserver_id_idx ON scan_result(nameserver_id);
CREATE INDEX scan_result_scanned_at_ok_idx ON scan_result(scanned_at) WHERE scan_status IN ('insecure'::cdnskey_status, 'secure'::cdnskey_status);