---
--- Run as defined rw user on newly created database
--- psql -U cdnskey_processor -d cdnskey_processor -f 0004_scan_result_evaluation.sql
---

DROP FUNCTION IF EXISTS raw_result_evaluation(scan_type TEXT, required_number_of_iterations BIGINT);
DROP FUNCTION IF EXISTS result_evaluation_insecure(required_number_of_iterations BIGINT, result_validity_timeout INTERVAL);
DROP FUNCTION IF EXISTS result_evaluation(required_number_of_iterations BIGINT, result_validity_timeout INTERVAL);
DROP FUNCTION IF EXISTS result_evaluation(acceptance_period INTERVAL, result_validity_timeout INTERVAL);

CREATE FUNCTION result_evaluation (acceptance_period INTERVAL, result_validity_timeout INTERVAL)
        RETURNS TABLE (
               domain_uid TEXT,
               domain_fqdn TEXT,
               dnskey_flags INTEGER,
               dnskey_proto INTEGER,
               dnskey_alg INTEGER,
               dnskey_public_key TEXT
        )
AS $$

WITH
scan_queue_in_process AS NOT MATERIALIZED (
     SELECT id
       FROM (SELECT *
               FROM scan_queue
              WHERE start_at IS NOT NULL
              ORDER BY start_at DESC
              LIMIT 1) AS latest_started_queue
       WHERE finished_at IS NULL
),
relevant_queues AS NOT MATERIALIZED (
      SELECT id
        FROM scan_queue
       WHERE start_at > CURRENT_TIMESTAMP - acceptance_period
             AND id NOT IN (SELECT id FROM scan_queue_in_process)
),
relevant_jobs_insecure AS NOT MATERIALIZED (
      SELECT j.id as scan_job_id, nameserver, domain_uid, domain_fqdn, scan_batch_id,
             COUNT(*) OVER w AS expected_result_count_per_domain_nameserver_worker_ip
        FROM scan_job AS j
             INNER JOIN relevant_queues AS q
                ON j.scan_queue_id = q.id
       WHERE j.scan_type = 'insecure'::SCAN_TYPE
      WINDOW w AS (PARTITION BY nameserver, domain_uid, domain_fqdn)
),
expected_results AS NOT MATERIALIZED (
      SELECT scan_job_id, nameserver, domain_uid, domain_fqdn, b.worker_id,
             expected_result_count_per_domain_nameserver_worker_ip,
             COUNT(*) OVER w AS expected_result_count_per_domain
        FROM scan_batch_progress AS b
             INNER JOIN relevant_jobs_insecure AS j
                ON j.scan_batch_id = b.scan_batch_id
       WHERE domain_uid NOT IN -- all results for remaining uids...
             (SELECT DISTINCT domain_uid
                FROM relevant_jobs_insecure
               WHERE scan_batch_id IN -- ...have all batches for all workers finished
                     (SELECT DISTINCT id
                        FROM scan_batch_progress
                       WHERE finished_at IS NULL
                     )
                  OR scan_job_id IN -- ...have some cdnskey
                     (SELECT DISTINCT scan_job_id
                        FROM scan_result
                       WHERE cdnskey_status != 'insecure'::CDNSKEY_STATUS
                     )
             )
      WINDOW w AS (PARTITION BY domain_uid)
),
raw_results_pass_1 AS NOT MATERIALIZED (
      SELECT e.scan_job_id, e.nameserver, e.domain_uid, e.domain_fqdn, e.worker_id,
             e.expected_result_count_per_domain_nameserver_worker_ip,
             e.expected_result_count_per_domain,
             r.scan_at, r.nameserver_ip,
             r.cdnskey_status, r.cdnskey_flags, r.cdnskey_proto, r.cdnskey_alg, r.cdnskey_public_key,
             COALESCE(LEAD(scan_at, -1)  OVER w_time, CURRENT_TIMESTAMP - acceptance_period) as prev_scan,
             COALESCE(LEAD(scan_at,  1)  OVER w_time, CURRENT_TIMESTAMP) as next_scan,
             COUNT(r.cdnskey_public_key) OVER w_time AS actual_result_count_per_domain_nameserver_worker_ip,
             (DENSE_RANK() OVER w_ns_ip_per_domain_ns = 1)::INTEGER     AS ip_mention_per_domain_ns,
             (DENSE_RANK() OVER w_ns_ip_per_domain_ns_key = 1)::INTEGER AS ip_mention_per_domain_ns_key,
             (DENSE_RANK() OVER w_results_per_domain_key = 1)::INTEGER  AS ns_mention_per_domain_key
        FROM expected_results AS e
             LEFT OUTER JOIN scan_result AS r
               ON r.scan_job_id = e.scan_job_id
                  AND r.worker_id = e.worker_id
       WHERE r.scan_at > CURRENT_TIMESTAMP - acceptance_period
      WINDOW w_time AS
             (PARTITION BY e.nameserver,
                           e.domain_uid,
                           e.domain_fqdn,
                           r.worker_id,
                           r.nameserver_ip,
                           r.cdnskey_status,
                           r.cdnskey_flags,
                           r.cdnskey_proto,
                           r.cdnskey_alg,
                           r.cdnskey_public_key
                  ORDER BY r.scan_at
                  RANGE BETWEEN
                        UNBOUNDED PRECEDING AND
                        UNBOUNDED FOLLOWING
             ),
             w_ns_ip_per_domain_ns_key AS
             (PARTITION BY e.nameserver,
                           e.domain_uid,
                           e.domain_fqdn,
                           r.worker_id,
                           r.cdnskey_status,
                           r.cdnskey_flags,
                           r.cdnskey_proto,
                           r.cdnskey_alg,
                           r.cdnskey_public_key,
                           r.nameserver_ip -- now at the end
                  ORDER BY r.id
             ),
             w_ns_ip_per_domain_ns AS
             (PARTITION BY e.nameserver,
                           e.domain_uid,
                           e.domain_fqdn,
                           r.nameserver_ip
                  ORDER BY r.id
             ),
             w_results_per_domain_key AS
             (PARTITION BY e.scan_job_id,
                           r.worker_id,
                           r.cdnskey_status,
                           r.cdnskey_flags,
                           r.cdnskey_proto,
                           r.cdnskey_alg,
                           r.cdnskey_public_key
                  ORDER BY r.id
             )
),
raw_results_pass_2 AS NOT MATERIALIZED (
      SELECT *,
             SUM (ip_mention_per_domain_ns)     OVER w_domain_ns     AS expected_ip_count_per_domain_key,
             SUM (ip_mention_per_domain_ns_key) OVER w_domain_ns_key AS actual_ip_count_per_domain_key,
             SUM (ns_mention_per_domain_key)    OVER w_domain_key    AS actual_result_count_per_domain
        FROM raw_results_pass_1
      WINDOW w_domain_ns AS
             (PARTITION BY nameserver,
                           domain_uid,
                           domain_fqdn
             ),
             w_domain_key AS
             (PARTITION BY domain_uid,
                           domain_fqdn,
                           cdnskey_status,
                           cdnskey_flags,
                           cdnskey_proto,
                           cdnskey_alg,
                           cdnskey_public_key
             ),
             w_domain_ns_key AS
             (PARTITION BY nameserver,
                           domain_uid,
                           domain_fqdn,
                           worker_id,
                           cdnskey_status,
                           cdnskey_flags,
                           cdnskey_proto,
                           cdnskey_alg,
                           cdnskey_public_key
             )
),
raw_results_pass_3 AS NOT MATERIALIZED (
      SELECT domain_uid, domain_fqdn,
             cdnskey_flags, cdnskey_proto, cdnskey_alg, cdnskey_public_key,
             BOOL_AND(cdnskey_status IS NOT NULL) OVER w_domain AS domain_passed,
             BOOL_AND
             (actual_result_count_per_domain_nameserver_worker_ip = expected_result_count_per_domain_nameserver_worker_ip
              AND actual_result_count_per_domain = expected_result_count_per_domain
              AND actual_ip_count_per_domain_key = expected_ip_count_per_domain_key
              AND scan_at - prev_scan < result_validity_timeout
              AND next_scan - scan_at < result_validity_timeout
             ) OVER w_domain_key AS key_passed
        FROM raw_results_pass_2
      WINDOW w_domain_key AS
             (PARTITION BY domain_uid,
                           domain_fqdn,
                           cdnskey_status,
                           cdnskey_flags,
                           cdnskey_proto,
                           cdnskey_alg,
                           cdnskey_public_key
             ),
             w_domain AS
             (PARTITION BY domain_uid,
                           domain_fqdn
             )
),
evaluated_insecure AS NOT MATERIALIZED (
      SELECT DISTINCT domain_uid, domain_fqdn,
             cdnskey_flags, cdnskey_proto, cdnskey_alg, cdnskey_public_key
        FROM raw_results_pass_3
       WHERE domain_passed AND key_passed
),
latest_finished_scan_queue AS NOT MATERIALIZED (
      SELECT id
        FROM scan_queue
       WHERE id NOT IN (SELECT id FROM scan_queue_in_process)
       ORDER BY start_at DESC
       LIMIT 1
),
relevant_jobs_secure AS NOT MATERIALIZED (
      SELECT id as scan_job_id, domain_uid, domain_fqdn, scan_batch_id
        FROM scan_job
       WHERE scan_type = 'secure'::SCAN_TYPE
         AND scan_queue_id = (SELECT id FROM latest_finished_scan_queue)
),
evaluated_secure AS NOT MATERIALIZED (
      SELECT DISTINCT domain_uid, domain_fqdn,
             cdnskey_flags, cdnskey_proto, cdnskey_alg, cdnskey_public_key
        FROM scan_batch_progress AS b
             INNER JOIN relevant_jobs_secure AS j
                ON j.scan_batch_id = b.scan_batch_id
             LEFT OUTER JOIN scan_result AS r
               ON r.scan_job_id = j.scan_job_id
       WHERE b.finished_at IS NOT NULL
             AND cdnskey_status = 'secure'::CDNSKEY_STATUS
             AND CURRENT_TIMESTAMP - scan_at < result_validity_timeout
)
SELECT * FROM evaluated_insecure
 UNION ALL
SELECT * FROM evaluated_secure

$$ LANGUAGE SQL;

-- get secure scan keys
WITH ls AS
(
    SELECT d.fqdn,
           MAX(p.finished_at) AS finished_at
      FROM domain d
      JOIN scan s ON s.domain_id = d.id
      JOIN scan_job j ON j.scan_id = s.id
      JOIN scan_batch_progress p ON p.scan_batch_id = j.scan_batch_id
     WHERE p.finished_at IS NOT NULL
  GROUP BY 1
),
s AS
(
    SELECT d.id AS domain_id,
           d.fqdn,
           p.scan_batch_id,
           NOT EXISTS(SELECT FROM insecure_scan i WHERE i.scan_id = s.id) AS is_secure
      FROM ls
      JOIN domain d ON d.fqdn = ls.fqdn
      JOIN scan s ON s.domain_id = d.id
      JOIN scan_job j ON j.scan_id = s.id
      JOIN scan_batch_progress p ON p.scan_batch_id = j.scan_batch_id AND p.finished_at = ls.finished_at
 LEFT JOIN scan_result r ON r.scan_job_id = j.id
  GROUP BY 1, 3, 4
    HAVING BOOL_OR(r.scan_status IN ('insecure'::cdnskey_status, 'secure'::cdnskey_status))
)
SELECT d.id,
       d.fqdn,
       d.uid,
       r.cdnskey_flags,
       r.cdnskey_proto,
       r.cdnskey_alg,
       r.cdnskey_public_key
  FROM s
  JOIN scan_job j ON j.scan_batch_id = s.scan_batch_id
  JOIN scan ON scan.id = j.scan_id
  JOIN domain d ON d.fqdn = s.fqdn AND d.id = scan.domain_id
  JOIN scan_result r ON r.scan_job_id = j.id
 WHERE s.is_secure AND
       r.scan_status = 'secure'::cdnskey_status
 ORDER BY 2;

WITH t AS -- scan results in 7 days: <min, max)
(
    SELECT max - '7DAYS'::INTERVAL AS min,
           today,
           max
      FROM (SELECT ((MAX(import_finished_at) + '1DAY'::INTERVAL)::DATE)::TIMESTAMP AS max,
                   (MAX(import_finished_at)::DATE)::TIMESTAMP AS today
              FROM scan_queue) t
),
ic AS -- insecure candidates: domains with at least one today scan with CDNSKEY record
(
    SELECT s.domain_id, MAX(r.id) AS scan_result_id
      FROM t,
           scan_result r
      JOIN scan_job j ON j.id = r.scan_job_id
      JOIN scan s ON s.id = j.scan_id
     WHERE r.scan_status = 'insecure'::cdnskey_status AND
           t.today <= r.scanned_at AND r.scanned_at < t.max
  GROUP BY 1
),
r AS -- scan results by domain, day, nameserver, nameserver ip
(
    SELECT d.fqdn AS domain,
           r.scanned_at::DATE AS day,
           n.fqdn AS nameserver,
           r.nameserver_ip,
           COALESCE(BOOL_OR(r.scan_status = lr.scan_status AND
                            r.cdnskey_flags = lr.cdnskey_flags AND
                            r.cdnskey_proto = lr.cdnskey_proto AND
                            r.cdnskey_alg = lr.cdnskey_alg AND
                            r.cdnskey_public_key = lr.cdnskey_public_key),
                            FALSE) AS is_correct
      FROM t,
           ic
      JOIN scan_result lr ON lr.id = ic.scan_result_id
      JOIN domain d ON d.id = ic.domain_id
      JOIN scan s ON s.domain_id = d.id
      JOIN insecure_scan i ON i.scan_id = s.id
      JOIN nameserver n ON n.id = i.nameserver_id
      JOIN scan_job j ON j.scan_id = s.id
      JOIN scan_result r ON r.scan_job_id = j.id AND
                            r.nameserver_id = n.id                             
     WHERE t.min <= r.scanned_at AND r.scanned_at < t.max
  GROUP BY 1, 2, 3, 4
),
rd AS -- scan results by domain, day
(
    SELECT domain, day, BOOL_AND(is_correct) AS is_correct
      FROM r
  GROUP BY 1, 2
)
SELECT domain,
       COALESCE(SUM(1) FILTER(WHERE is_correct), 0) AS total_correct_days,
       ARRAY_AGG(day) FILTER(WHERE is_correct) AS correct_days
  FROM rd
 GROUP BY 1
 ORDER BY 1;

WITH t AS -- scan results today: <min, max)
(
    SELECT ((MAX(import_finished_at) + '1DAY'::INTERVAL)::DATE)::TIMESTAMP AS max,
           (MAX(import_finished_at)::DATE)::TIMESTAMP AS min
      FROM scan_queue t
),
sc AS -- secure candidates: domains with at least one today scan with CDNSKEY record
(
    SELECT s.domain_id, MAX(r.id) AS scan_result_id
      FROM t,
           scan_result r
      JOIN scan_job j ON j.id = r.scan_job_id
      JOIN scan s ON s.id = j.scan_id
     WHERE r.scan_status = 'secure'::cdnskey_status AND
           t.min <= r.scanned_at AND r.scanned_at < t.max
  GROUP BY 1
)
SELECT d.fqdn AS domain,
       r.cdnskey_flags,
       r.cdnskey_proto,
       r.cdnskey_alg,
       r.cdnskey_public_key
  FROM sc
  JOIN scan_result r ON r.id = sc.scan_result_id
  JOIN domain d ON d.id = sc.domain_id
 ORDER BY 1;

WITH param AS
(
    SELECT (NOW() - max_scan_period)::TIMESTAMP AS oldest_last_scan,
           (NOW() - max_scan_period - min_period_fixed_cdnskey)::TIMESTAMP AS oldest_first_scan,
           max_scan_period,
           min_period_fixed_cdnskey
      FROM (SELECT '2DAYS'::INTERVAL, '1DAYS'::INTERVAL) AS c(max_scan_period, min_period_fixed_cdnskey)
),
ic AS -- insecure candidates: domains with at least one CDNSKEY record found in the last period
(
    SELECT s.domain_id
      FROM param,
           scan_result r
      JOIN scan_job j ON j.id = r.scan_job_id
      JOIN scan s ON s.id = j.scan_id
     WHERE r.scan_status = 'insecure'::cdnskey_status AND
           param.oldest_last_scan <= r.scanned_at
  GROUP BY 1
),
r AS -- scan results by domain, scan_id, nameserver, nameserver ip
(
    SELECT d.fqdn AS domain,
           d.uid AS domain_uid,
           s.id AS scan_id,
           n.fqdn AS nameserver,
           r.nameserver_ip,
           r.scanned_at,
           r.scan_status,
           r.cdnskey_flags,
           r.cdnskey_proto,
           r.cdnskey_alg,
           r.cdnskey_public_key
      FROM ic
      JOIN domain d ON d.id = ic.domain_id
      JOIN scan s ON s.domain_id = d.id
      JOIN scan_queue q ON q.id = s.scan_queue_id
      JOIN param ON param.oldest_first_scan <= q.import_finished_at
      JOIN insecure_scan i ON i.scan_id = s.id
      JOIN nameserver n ON n.id = i.nameserver_id
      JOIN scan_job j ON j.scan_id = s.id
 LEFT JOIN scan_result r ON r.scan_job_id = j.id AND
                            r.nameserver_id = n.id
),
ds AS -- domain scan
(
    SELECT domain,
           domain_uid,
           scan_id,
           MAX(scanned_at) AS scanned_at,
           MIN(cdnskey_flags) AS cdnskey_flags,
           MIN(cdnskey_proto) AS cdnskey_proto,
           MIN(cdnskey_alg) AS cdnskey_alg,
           MIN(cdnskey_public_key) AS cdnskey_public_key
      FROM r
  GROUP BY 1, 2, 3
    HAVING BOOL_AND(scan_status = 'insecure'::cdnskey_status) AND
           ARRAY_LENGTH(ARRAY_AGG(DISTINCT ROW(cdnskey_flags,
                                               cdnskey_proto,
                                               cdnskey_alg,
                                               cdnskey_public_key)), 1) = 1
),
dss AS -- domain scan serie
(
    SELECT domain,
           domain_uid,
           scan_id,
           scanned_at,
           cdnskey_flags,
           cdnskey_proto,
           cdnskey_alg,
           cdnskey_public_key,
           last_value(ROW(cdnskey_flags,
                          cdnskey_proto,
                          cdnskey_alg,
                          cdnskey_public_key)) OVER d =
           first_value(ROW(cdnskey_flags,
                           cdnskey_proto,
                           cdnskey_alg,
                           cdnskey_public_key)) OVER d AS is_the_same_key,
           last_value(scanned_at) OVER d - first_value(scanned_at) OVER d AS total_diff,
           last_value(scanned_at) OVER d - lead(scanned_at, -1) OVER d AS prev_diff
      FROM ds
    WINDOW d AS (PARTITION BY domain ORDER BY scanned_at)
),
sc AS -- secure candidates
(
    SELECT s.domain_id,
           MAX(r.id) AS scan_result_id
      FROM scan_queue q
      JOIN scan s ON s.scan_queue_id = q.id AND
                     NOT EXISTS(SELECT FROM insecure_scan i WHERE i.scan_id = s.id)
      JOIN scan_job j ON j.scan_id = s.id
      JOIN scan_result r ON r.scan_job_id = j.id
     WHERE (NOW() - '2DAYS'::INTERVAL)::TIMESTAMP <= q.import_finished_at
  GROUP BY 1
)
SELECT domain,
       domain_uid,
       MIN(cdnskey_flags) AS cdnskey_flags,
       MIN(cdnskey_proto) AS cdnskey_proto,
       MIN(cdnskey_alg) AS cdnskey_alg,
       MIN(cdnskey_public_key) AS cdnskey_public_key
  FROM dss,
       param
 GROUP BY 1, 2, param.min_period_fixed_cdnskey, param.max_scan_period
HAVING BOOL_AND(is_the_same_key) AND
       param.min_period_fixed_cdnskey <= MAX(total_diff) AND
       MAX(prev_diff) <= param.max_scan_period
UNION
SELECT d.fqdn,
       d.uid,
       r.cdnskey_flags,
       r.cdnskey_proto,
       r.cdnskey_alg,
       r.cdnskey_public_key
  FROM ssc
  JOIN domain d ON d.id = ssc.domain_id
  JOIN scan_result r ON r.id = ssc.scan_result_id AND
                        r.scan_status = 'secure'::cdnskey_status
 ORDER BY 1;

r AS
(
    SELECT day.id AS day,
           result.scan_id,
           result.domain,
           result.nameserver,
           result.nameserver_ip,
           COALESCE(result.is_correct, FALSE) AS is_correct
      FROM day
 LEFT JOIN result ON result.day = day.id
)
SELECT *
  FROM r
 ORDER BY 1, 2, 4, 5;

WITH ssc AS -- secure scan candidate
(
    SELECT s.domain_id,
           MAX(r.id) AS scan_result_id
      FROM scan_queue q
      JOIN scan s ON s.scan_queue_id = q.id AND
                     NOT EXISTS(SELECT FROM insecure_scan i WHERE i.scan_id = s.id)
      JOIN scan_job j ON j.scan_id = s.id
      JOIN scan_result r ON r.scan_job_id = j.id
     WHERE (NOW() - '2DAYS'::INTERVAL)::TIMESTAMP <= q.import_finished_at
  GROUP BY 1
)
SELECT d.fqdn,
       d.uid,
       r.cdnskey_flags,
       r.cdnskey_proto,
       r.cdnskey_alg,
       r.cdnskey_public_key
  FROM ssc
  JOIN domain d ON d.id = ssc.domain_id
  JOIN scan_result r ON r.id = ssc.scan_result_id AND
                        r.scan_status = 'secure'::cdnskey_status
 ORDER BY 1;

WITH s AS
(
    SELECT s.id
      FROM scan_queue q
      JOIN scan s ON s.scan_queue_id = q.id
     WHERE q.import_started_at < (NOW() - '11DAYS'::INTERVAL)::TIMESTAMP
),
dr AS
(
    DELETE FROM scan_result r
          USING s,
                scan_job j
          WHERE j.scan_id = s.id AND
                r.scan_job_id = j.id
),
dj AS
(
    DELETE FROM scan_job j
          USING s
          WHERE j.scan_id = s.id
),
di AS
(
    DELETE FROM insecure_scan i
          USING s
          WHERE i.scan_id = s.id
),
ds AS
(
    DELETE FROM scan
          USING s
          WHERE scan.id = s.id
),
b AS
(
    SELECT id
      FROM scan_batch b
     WHERE NOT EXISTS(SELECT FROM scan_job j WHERE j.scan_batch_id = b.id)
),
dbp AS
(
    DELETE FROM scan_batch_progress bp
          USING b
          WHERE bp.scan_batch_id = b.id
),
db AS
(
    DELETE FROM scan_batch sb
          USING b
          WHERE sb.id = b.id
),
dd AS
(
    DELETE FROM domain d
          WHERE NOT EXISTS(SELECT FROM scan s WHERE s.domain_id = d.id)
),
dn AS
(
    DELETE FROM nameserver n
          WHERE NOT EXISTS(SELECT FROM insecure_scan i WHERE i.nameserver_id = n.id) AND
                NOT EXISTS(SELECT FROM scan_result r WHERE r.nameserver_id = n.id)
)
DELETE FROM scan_queue
      WHERE import_started_at < (NOW() - '11DAYS'::INTERVAL)::TIMESTAMP
  RETURNING id;

WITH od AS
(
    SELECT domain.id, d.fqdn, d.uid
      FROM (VALUES ('uj2l3e5clv0aw-kld7gjg.cz', '37295775-9c14-42f3-840a-d9ec0fc93bf2')
                   ('eo082x-og.cz', '6032ceda-d6a7-4bfc-bb49-e55f306af626')
                   ('kif4bfab04j28.cz', 'ef333184-b421-424c-809d-71d1a9e7a2f9')
                   ('7nteo77y33.cz', 'b42695f9-19cd-4540-b311-6ddfca906152')
                   ('9aube2.cz', 'd6af73bc-2fe1-4fa5-a48e-d932d0bd2722')
                   ('6if9w8i1.cz', '23ebe4d0-e34e-4429-8dce-2a3eac9f6db3')
                   ('32f29820.cz', '501b717e-3765-420e-954a-9f5eb5538283')
                   ('rl4lpag58l3l.cz', '40625afd-1621-4a24-920e-ff1086a6ff30')
                   ('7s23ne85.cz', '80b1c0bf-e692-4a0b-a497-06955c14f405')
                   ('d0fd0hwi4wvhavafc.cz', 'cdf7dedd-f515-4b3e-be30-e7b76641caf8')
                   ('fruh80.cz', '95b82f25-f98e-49e6-94d0-26aa2c5359d2')
                   ('d654p93.cz', 'f4861a1a-ca8a-4afc-bcf0-cd5e5da2af88')
                   ('4n90.cz', '9d301bd5-5a47-4629-82c9-b65a89bff6e3')
                   ('c9biw5ubdii1.cz', 'e5da9db3-5800-4267-aa1a-447a6c546ff1')
                   ('826ku285ek4khek.cz', '639ae0ee-98da-450b-a7ba-1bea3931a10f')
                   ('ak0h58ers.cz', '0ebc0ce2-c9ae-446d-96ee-4543936f64ad')) d(fqdn, uid)
 LEFT JOIN domain ON domain.fqdn = d.fqdn AND
                     domain.uid = d.uid
     WHERE d.fqdn IS NOT NULL AND
           d.uid IS NOT NULL
),
nd AS
(
    INSERT INTO domain (fqdn, uid)
         SELECT fqdn, uid
           FROM od
          WHERE id IS NULL
      RETURNING id, fqdn, uid
),
d AS
(
    SELECT * FROM od WHERE id IS NOT NULL
    UNION
    SELECT * FROM nd
)
INSERT INTO scan (scan_queue_id, domain_id)
     SELECT 53, id
       FROM d
  RETURNING id;

WITH ns AS
(
    SELECT fqdn, UNNEST(ns) AS ns_fqdn
      FROM (VALUES ('nic.cz', ARRAY['a.ns.cz', 'b.ns.cz', NULL]),
                   ('nica.cz', ARRAY['a.ns.cz', NULL, NULL]),
                   ('nicab.cz', ARRAY['c.ns.cz', 'd.ns.cz', NULL]),
                   (NULL, ARRAY[NULL, NULL, NULL])) dn(fqdn, ns)
     WHERE dn.fqdn IS NOT NULL
)
SELECT DISTINCT fqdn, ns_fqdn FROM ns;

BEGIN;
WITH od AS
(
    SELECT domain.id, d.fqdn, d.uid, d.ns
      FROM (VALUES ('00445.cz', '0647a6e5-4210-4156-92c4-c58874f42cef', ARRAY['001.pasas.cz', '002.pasas.cz', NULL])) d(fqdn, uid, ns)
 LEFT JOIN domain ON domain.fqdn = d.fqdn AND
           domain.uid = d.uid
     WHERE d.fqdn IS NOT NULL AND
           d.uid IS NOT NULL
),
nd AS
(
    INSERT INTO domain (fqdn, uid)
         SELECT fqdn, uid
           FROM od
          WHERE id IS NULL
       RETURNING id, fqdn, uid
),
d AS
(
    SELECT * FROM od WHERE id IS NOT NULL
    UNION
    SELECT nd.id, nd.fqdn, nd.uid, od.ns
      FROM nd
      JOIN od ON od.fqdn = nd.fqdn AND od.uid = nd.uid
),
s AS
(
    INSERT INTO scan (scan_queue_id, domain_id)
         SELECT 59,
                id
           FROM d
      RETURNING id, domain_id
),
ons AS
(
    SELECT n.id, ans.fqdn
      FROM (SELECT DISTINCT UNNEST(ns) AS fqdn FROM d) ans
 LEFT JOIN nameserver n ON n.fqdn = ans.fqdn
     WHERE ans.fqdn IS NOT NULL
),
nns AS
(
    INSERT INTO nameserver (fqdn) SELECT fqdn FROM ons WHERE id IS NULL
      RETURNING id, fqdn
),
ns AS
(
    SELECT * FROM ons WHERE id IS NOT NULL
    UNION
    SELECT * FROM nns
)
INSERT INTO insecure_scan (scan_id, nameserver_id)
     SELECT s.id, ns.id
       FROM (SELECT id, fqdn, uid, UNNEST(ns)
               FROM d) AS dns(domain_id, domain_fqdn, uid, ns_fqdn)
       JOIN s ON s.domain_id = dns.domain_id
       JOIN ns ON ns.fqdn = dns.ns_fqdn
  RETURNING scan_id, nameserver_id;
ROLLBACK;

WITH i AS -- insecure
(
    SELECT s.id AS scan_id,
           n.id AS nameserver_id
      FROM scan_queue q
      JOIN scan s ON s.scan_queue_id = q.id
      JOIN insecure_scan i ON i.scan_id = s.id
      JOIN nameserver n ON n.id = i.nameserver_id
     WHERE q.import_finished_at IS NOT NULL AND
           q.id = 73
),
b AS -- buckets
(
    SELECT FIRST_VALUE(scan_id) OVER (PARTITION BY chunk ORDER BY scan_id) =
           LAST_VALUE(scan_id) OVER (PARTITION BY chunk ORDER BY scan_id) AS is_first,
           chunk,
           scan_id
      FROM (SELECT NTILE((SELECT GREATEST(1, COUNT(*) / 15000) FROM i)::INT)
                       OVER (ORDER BY scan_id, nameserver_id) AS chunk,
                   scan_id
              FROM i) tmp
),
bs AS -- bucket statistics
(
    SELECT chunk,
           MIN(scan_id) AS first,
           MAX(scan_id) AS last
      FROM b
  GROUP BY 1
),
bi AS -- bucket informations
(
    SELECT b0.chunk,
           COALESCE(b1.last = b0.first, FALSE) AS exclude_first
      FROM bs b0
 LEFT JOIN bs b1 ON b1.chunk = b0.chunk - 1
),
ib AS -- insecure buckets
(
    SELECT DISTINCT b.chunk, b.scan_id
      FROM bi
      JOIN b ON b.chunk = bi.chunk AND
                NOT (bi.exclude_first AND b.is_first)
),
s AS -- secure
(
    SELECT s.id AS scan_id
      FROM scan_queue q
      JOIN scan s ON s.scan_queue_id = q.id
 LEFT JOIN insecure_scan i ON i.scan_id = s.id
     WHERE q.import_finished_at IS NOT NULL AND
           q.id = 73 AND
           i.scan_id IS NULL
),
sb AS -- secure buckets
(
    SELECT NTILE((SELECT GREATEST(1, COUNT(*) / 15000) FROM s)::INT)
               OVER (ORDER BY scan_id) +
           (SELECT MAX(chunk) FROM ib) AS chunk,
           scan_id
      FROM s
),
batch AS
(
    SELECT * FROM ib
    UNION
    SELECT * FROM sb
),
isb AS
(
    INSERT INTO scan_batch SELECT FROM batch GROUP BY chunk
      RETURNING id
),
nb AS -- new batch
(
    SELECT ROW_NUMBER() OVER (ORDER BY id) AS chunk,
           id AS scan_batch_id
    FROM isb
),
job AS
(
    INSERT INTO scan_job (scan_batch_id, scan_id)
         SELECT nb.scan_batch_id, batch.scan_id
           FROM batch
           JOIN nb ON nb.chunk = batch.chunk
      RETURNING id, scan_batch_id, scan_id
)
SELECT COUNT(DISTINCT scan_batch_id)
  FROM job;

SELECT s.scan_queue_id,
       COUNT(*),
       COUNT(DISTINCT r.scan_job_id),
       MIN(r.scanned_at),
       MAX(r.scanned_at),
       (((SELECT COUNT(*) FROM scan WHERE scan.scan_queue_id = s.scan_queue_id) *
         EXTRACT('epoch' FROM MAX(r.scanned_at) - MIN(r.scanned_at)) /
         COUNT(DISTINCT r.scan_job_id))||'SECONDS')::INTERVAL +
       MIN(r.scanned_at)
  FROM scan s
  LEFT JOIN (scan_job j
        JOIN scan_result r ON r.scan_job_id = j.id) ON j.scan_id = s.id
 GROUP BY 1
 ORDER BY 1;

WITH s AS
(
    SELECT s.scan_queue_id,
           s.id,
           BOOL_AND(i.scan_id IS NULL) AS is_secure
      FROM scan s
 LEFT JOIN insecure_scan i ON i.scan_id = s.id
  GROUP BY 1, 2
),
ss AS
(
    SELECT s.scan_queue_id,
           COUNT(s.id) AS domains,
           COUNT(s.id) FILTER (WHERE s.is_secure) AS secure_domains
      FROM s
  GROUP BY 1
),
rs AS
(
    SELECT ss.scan_queue_id,
           COUNT(*) AS results,
           COUNT(DISTINCT r.scan_job_id) FILTER (WHERE s.is_secure) AS secure_scanned,
           COUNT(DISTINCT r.scan_job_id) FILTER (WHERE NOT s.is_secure) AS insecure_scanned,
           MIN(r.scanned_at) FILTER (WHERE s.is_secure) AS secure_started_at,
           MAX(r.scanned_at) FILTER (WHERE s.is_secure) AS secure_last_at,
           MIN(r.scanned_at) FILTER (WHERE NOT s.is_secure) AS insecure_started_at,
           MAX(r.scanned_at) FILTER (WHERE NOT s.is_secure) AS insecure_last_at
      FROM s
      JOIN ss ON ss.scan_queue_id = s.scan_queue_id
 LEFT JOIN (scan_job j
       JOIN scan_result r ON r.scan_job_id = j.id) ON j.scan_id = s.id
  GROUP BY 1
)
SELECT ss.scan_queue_id,
       ss.domains,
       ss.secure_domains,
       rs.results,
       rs.secure_scanned,
       rs.insecure_scanned,
       rs.secure_started_at,
       rs.secure_last_at,
       rs.insecure_started_at,
       rs.insecure_last_at
  FROM rs
  JOIN ss ON ss.scan_queue_id = rs.scan_queue_id
ORDER BY 1;

WITH s AS
(
    SELECT s.scan_queue_id,
           j.scan_batch_id,
           COUNT(*) AS scans,
           COUNT(DISTINCT s.domain_id) AS to_scan,
           COUNT(DISTINCT s.domain_id) FILTER (WHERE r.id IS NOT NULL) AS scanned,
           MIN(r.scanned_at) AS first,
           MAX(r.scanned_at) AS last
      FROM scan s
      JOIN scan_job j ON j.scan_id = s.id
 LEFT JOIN scan_result r ON r.scan_job_id = j.id
  GROUP BY 1, 2
),
r AS
(
    SELECT scan_queue_id,
           SUM(scans) AS scans,
           SUM(to_scan) AS domains_to_scan,
           SUM(scanned) AS domains_scanned,
           SUM(to_scan - scanned) AS domains_left,
           COUNT(*) AS jobs_to_scan,
           COUNT(*) FILTER (WHERE 0 < scanned) AS jobs_scanned,
           MAX(last) - MIN(first) AS duration,
           MIN(first) AS first,
           MAX(last) AS last
      FROM s
  GROUP BY 1
)
SELECT *,
       first + ((EXTRACT('epoch' FROM duration) * jobs_to_scan / jobs_scanned)||'SECONDS')::INTERVAL
       AS estimated_end
  FROM r
 ORDER BY 1;