/*
 * Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/cfg/api.hh"
#include "src/database.hh"
#include "src/grpc_service_report.hh"
#include "src/grpc_service_supervisor.hh"

#include "cdnskey_processor_api/service_report_grpc.grpc.pb.h"
#include "cdnskey_processor_api/service_supervisor_grpc.grpc.pb.h"

#include <grpcpp/grpcpp.h>

#include <liblog/liblog.hh>

#include <cstdlib>
#include <csignal>
#include <cstring>
#include <exception>
#include <memory>
#include <thread>

namespace {

template <int sig>
void add(::sigset_t& signals)
{
    ::sigaddset(&signals, sig);
}

template <int first, int second, int ...others>
void add(::sigset_t& signals)
{
    add<first>(signals);
    add<second, others...>(signals);
}

template <int ...signals>
void block_signals(::sigset_t& blocked_signals)
{
    // Other threads created by grpc::Server will inherit a copy of the signal mask.
    ::sigemptyset(&blocked_signals);
    add<signals...>(blocked_signals);
    const auto pthread_sigmask_result = ::pthread_sigmask(SIG_BLOCK, &blocked_signals, nullptr);
    constexpr int operation_success = 0;
    if (pthread_sigmask_result != operation_success)
    {
        struct PthreadSigmaskFailure : std::exception
        {
            explicit PthreadSigmaskFailure(int error_code) : error_code{error_code} { }
            const char* what() const noexcept override { return std::strerror(error_code); }
            int error_code;
        };
        throw PthreadSigmaskFailure{pthread_sigmask_result};
    }
}

const char* get_sig_name(int sig)
{
    switch (sig)
    {
        case SIGHUP: return "SIGHUP";
        case SIGINT: return "SIGINT";
        case SIGQUIT: return "SIGQUIT";
        case SIGTERM: return "SIGTERM";
    }
    struct UnexpectedValue : std::exception
    {
        const char* what() const noexcept override { return "unexpected signal"; }
    };
    throw UnexpectedValue{};
}

void wait_on_signal(const ::sigset_t& signals)
{
    int received_signal;
    const auto sigwait_result = ::sigwait(&signals, &received_signal);
    constexpr int operation_success = 0;
    if (sigwait_result != operation_success)
    {
        struct SigwaitFailure : std::exception
        {
            explicit SigwaitFailure(int error_code) : error_code{error_code} { }
            const char* what() const noexcept override { return std::strerror(error_code); }
            int error_code;
        };
        throw SigwaitFailure{sigwait_result};
    }
    LIBLOG_DEBUG("{} received", get_sig_name(received_signal));
}

}//namespace {anonymous}

int main(int argc, char* argv[])
{
    try
    {
        const auto cfg = CdnskeyProcessor::Cfg::get_cfg(
                argc,
                argv,
                CdnskeyProcessor::Cfg::set_up_logging,
                CdnskeyProcessor::set_up_database);
        LIBLOG_INFO("api is running");

        ::sigset_t blocked_signals;
        block_signals<SIGHUP, SIGINT, SIGQUIT, SIGTERM>(blocked_signals);

        grpc::ServerBuilder builder;
        builder.AddListeningPort(cfg.grpc_address, grpc::InsecureServerCredentials());

        using ReportService = CdnskeyProcessor::Api::Report::Report::Service;
        const auto report_service = [&]() -> std::unique_ptr<ReportService>
        {
            if (cfg.report.disabled)
            {
                LIBLOG_INFO("do not run Report service");
                return nullptr;
            }
            LIBLOG_INFO("run Report service");
            auto service = CdnskeyProcessor::make_grpc_service_report(cfg.report);
            builder.RegisterService(service.get());
            return service;
        }();

        using SupervisorService = CdnskeyProcessor::Api::Supervisor::Supervisor::Service;
        const auto supervisor_service = [&]() -> std::unique_ptr<SupervisorService>
        {
            if (cfg.supervisor.disabled)
            {
                LIBLOG_INFO("do not run Supervisor service");
                return nullptr;
            }
            LIBLOG_INFO("run Supervisor service");
            auto service = CdnskeyProcessor::make_grpc_service_supervisor(cfg.supervisor);
            builder.RegisterService(service.get());
            return service;
        }();

        if (!report_service && !supervisor_service)
        {
            LIBLOG_INFO("no service to run");
            return EXIT_SUCCESS;
        }
        auto server = builder.BuildAndStart();
        wait_on_signal(blocked_signals);
        static constexpr auto deadline = std::chrono::seconds{10};
        server->Shutdown(std::chrono::system_clock::now() + deadline);
        return EXIT_SUCCESS;
    }
    catch (const CdnskeyProcessor::Cfg::Api::Help&)
    {
        return EXIT_SUCCESS;
    }
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("unexpected error occurred: {}", e.what());
    }
    catch (...)
    {
        LIBLOG_ERROR("unexpected error occurred");
    }
    return EXIT_FAILURE;
}
