/*
 * Copyright (C) 2023-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/cfg/api.hh"
#include "src/cfg/validator.hh"

#include "config.h"

#include <boost/program_options.hpp>

#include <exception>
#include <fstream>
#include <iostream>

namespace CdnskeyProcessor {
namespace Cfg {

namespace {

struct GeneralArgs
{
    std::string config_file;
};

boost::program_options::options_description get_options(GeneralArgs& general_args)
{
    namespace po = boost::program_options;
    po::options_description options{"General options", program_options_line_length};
    options.add_options()
        ("conf,c", po::value<std::string>(&general_args.config_file)->required(),
         "configuration file")
        ("help,h",
         "display this help message");
    return options;
}

boost::program_options::options_description get_options(Api::Report& cfg)
{
    namespace po = boost::program_options;
    po::options_description options{"gRPC Report service options", program_options_line_length};
    options.add_options()
        ("report.disabled",
         po::value<bool>(&cfg.disabled)->default_value(false),
         "Do not run gRPC Report service");
    return options;
}

boost::program_options::options_description get_options(Api::Supervisor& cfg)
{
    namespace po = boost::program_options;
    po::options_description options{"gRPC Supervisor service options", program_options_line_length};
    options.add_options()
        ("supervisor.disabled",
         po::value<bool>(&cfg.disabled)->default_value(false),
         "Do not run gRPC Supervisor service")
        ("supervisor.batch_size",
         po::value<std::size_t>(&cfg.batch_size)->default_value(15'000),
         "Maximal number of scanned items in a batch")
        ("supervisor.max_scan_period",
         po::value<std::chrono::seconds>(&cfg.max_scan_period)->default_value(std::chrono::hours{2 * 24}),
         "Maximal interval between two scans")
        ("supervisor.min_period_fixed_cdnskey",
         po::value<std::chrono::seconds>(&cfg.min_period_fixed_cdnskey)->default_value(std::chrono::hours{7 * 24}),
         "Minimal interval between first and last scans");
    return options;
}

boost::program_options::options_description get_options(Api& cfg)
{
    namespace po = boost::program_options;
    po::options_description options{"gRPC General options", program_options_line_length};
    options.add_options()
        ("grpc.address",
         po::value<std::string>(&cfg.grpc_address)->required(),
         "gRPC address which the server is listening to");
    options.add(get_options(cfg.report));
    options.add(get_options(cfg.supervisor));
    return options;
}

}//namespace CdnskeyProcessor::Cfg::{anonymous}

}//namespace CdnskeyProcessor::Cfg
}//namespace CdnskeyProcessor

using namespace CdnskeyProcessor::Cfg;

Api CdnskeyProcessor::Cfg::get_cfg(
        int argc,
        char** argv,
        void(*set_up_logging)(const Log&),
        void(*set_up_database)(const Database&))
{

    const auto file_name = [&]()
    {
        GeneralArgs general_args;
        auto all_options = get_options(general_args);
        boost::program_options::variables_map vm;
        boost::program_options::store(boost::program_options::parse_command_line(argc, argv, all_options), vm);
        if (vm.count("help"))
        {
            Api cfg;
            Log log_cfg;
            Database db_cfg;
            all_options.add(get_options(log_cfg))
                       .add(get_options(db_cfg))
                       .add(get_options(cfg));
            std::cout << all_options << std::flush;
            throw Api::Help{};
        }
        notify(vm);
        return std::move(general_args.config_file);
    }();
    auto cfg_file = std::ifstream{file_name};
    if (!cfg_file.is_open())
    {
        struct UnableToOpen : std::exception
        {
            const char* what() const noexcept override { return "unable to open config file"; }
        };
        throw UnableToOpen{};
    }
    Api cfg;
    Log log_cfg;
    Database db_cfg;
    boost::program_options::options_description all_options{"cdnskey-processor-master options"};
    all_options.add(get_options(log_cfg))
               .add(get_options(db_cfg))
               .add(get_options(cfg));
    boost::program_options::variables_map vm;
    boost::program_options::store(boost::program_options::parse_config_file(cfg_file, all_options), vm);
    boost::program_options::notify(vm);
    if (set_up_logging != nullptr)
    {
        set_up_logging(log_cfg);
    }
    if (set_up_database != nullptr)
    {
        set_up_database(db_cfg);
    }
    return cfg;
}
