/*
 * Copyright (C) 2023-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/cfg/log.hh"
#include "src/cfg/validator.hh"

#include "config.h"

#include <liblog/log.hh>

#include <syslog.h>
#include <vector>

namespace CdnskeyProcessor {
namespace Cfg {

namespace {

int make_facility(int offset)
{
    switch (offset)
    {
        case 0: return LOG_LOCAL0;
        case 1: return LOG_LOCAL1;
        case 2: return LOG_LOCAL2;
        case 3: return LOG_LOCAL3;
        case 4: return LOG_LOCAL4;
        case 5: return LOG_LOCAL5;
        case 6: return LOG_LOCAL6;
        case 7: return LOG_LOCAL7;
    }
    auto exception = boost::program_options::validation_error{
            boost::program_options::validation_error::invalid_option_value};
    exception.m_error_template.append("; "
            "facility must be from range [0, 7]");
    throw exception;
};

}//namespace CdnskeyProcessor::Cfg::{anonymous}

}//namespace CdnskeyProcessor::Cfg
}//namespace CdnskeyProcessor

using namespace CdnskeyProcessor::Cfg;

boost::program_options::options_description CdnskeyProcessor::Cfg::get_options(Log& cfg)
{
    namespace po = boost::program_options;
    po::options_description log_options{"Logging options", program_options_line_length};
    static constexpr const char* min_severity_description = "do not log more trivial events; "
            "severity in descending order: critical/error/warning/info/debug/trace";
    log_options.add_options()
        ("log.min_severity",
         po::value<Log::Level>(&cfg.min_severity)->default_value(Log::Level::trace, "trace"),
         min_severity_description)

        ("log.console.enabled", po::bool_switch(&cfg.console.enabled), "enable logging on the console")
        ("log.console.min_severity",
         po::value<Log::Level>()->notifier([&cfg](const Log::Level& level)
         {
            cfg.console.min_severity = level;
         }),
         min_severity_description)

        ("log.file.enabled", po::bool_switch(&cfg.file.enabled), "enable logging to the file")
        ("log.file.file_name",
         po::value<std::string>()->notifier([&cfg](const std::string& name)
         {
            cfg.file.file_name = name;
         }),
         "what file to log into")
        ("log.file.min_severity",
         po::value<Log::Level>()->notifier([&cfg](const Log::Level& level)
         {
            cfg.file.min_severity = level;
         }),
         min_severity_description)

        ("log.syslog.enabled", po::bool_switch(&cfg.syslog.enabled), "enable logging into syslog")
        ("log.syslog.facility",
         po::value<int>()->notifier([&cfg](const int& offset)
         {
            cfg.syslog.facility = make_facility(offset);
         }),
         "what facility to log with (range 0..7 mapped to LOG_LOCAL0..7; default LOG_USER)")
        ("log.syslog.min_severity",
         po::value<Log::Level>()->notifier([&cfg](const Log::Level& level)
         {
            cfg.syslog.min_severity = level;
         }),
         min_severity_description);
    return log_options;
}

LibLog::Level to_liblog_level(Log::Level src)
{
    switch (src)
    {
        case Log::Level::critical:
            return LibLog::Level::critical;
        case Log::Level::error:
            return LibLog::Level::error;
        case Log::Level::warning:
            return LibLog::Level::warning;
        case Log::Level::info:
            return LibLog::Level::info;
        case Log::Level::debug:
            return LibLog::Level::debug;
        case Log::Level::trace:
            return LibLog::Level::trace;
    }
    struct InvalidLogLevel : std::exception
    {
        const char* what() const noexcept override { return "invalid log level"; }
    };
    throw InvalidLogLevel{};
}

LibLog::Level get_liblog_level(const boost::optional<Log::Level>& option, Log::Level default_level)
{
    if (option != boost::none)
    {
        return to_liblog_level(*option);
    }
    return to_liblog_level(default_level);
}

void CdnskeyProcessor::Cfg::set_up_logging(const Log& cfg)
{
    constexpr LibLog::ThreadMode threading = LibLog::ThreadMode::multi_threaded; // global, compile-time parameter
    auto liblog_log_config = LibLog::LogConfig();
    if (cfg.console.enabled)
    {
        liblog_log_config.add_sink_config(
                LibLog::Sink::ConsoleSinkConfig{}
                    .set_level(get_liblog_level(cfg.console.min_severity, cfg.min_severity)));
    }
    if (cfg.file.enabled)
    {
        liblog_log_config.add_sink_config(
                LibLog::Sink::FileSinkConfig{cfg.file.file_name}
                    .set_level(get_liblog_level(cfg.file.min_severity, cfg.min_severity)));
    }
    if (cfg.syslog.enabled)
    {
        liblog_log_config.add_sink_config(
                LibLog::Sink::SyslogSinkConfig{}
                .set_syslog_facility(cfg.syslog.facility == boost::none ? LOG_USER
                                                                        : *cfg.syslog.facility)
                .set_level(get_liblog_level(cfg.syslog.min_severity, cfg.min_severity)));
    }
    liblog_log_config.set_level(to_liblog_level(cfg.min_severity));
    LibLog::Log::start<threading>(liblog_log_config);
}
