/*
 * Copyright (C) 2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/cfg/validator.hh"

#include <boost/lexical_cast.hpp>
#include <boost/program_options.hpp>

#include <exception>
#include <iostream>
#include <unordered_map>

using namespace std::chrono;

void std::chrono::validate(boost::any& dst, const std::vector<std::string>& options, seconds*, int)
{
    using namespace boost::program_options;
    validators::check_first_occurrence(dst);

    const auto value = [&options]()
    {
        try
        {
            return seconds{boost::lexical_cast<seconds::duration::rep>(validators::get_single_string(options))};
        }
        catch (const boost::bad_lexical_cast&)
        {
            auto exception = validation_error{validation_error::invalid_option_value};
            exception.m_error_template += "; not a seconds";
            throw exception;
        }
    }();
    dst = boost::any{seconds{value}};
}

std::ostream& std::chrono::operator<<(std::ostream& out, const seconds& value)
{
    return out << value.count();
}

using namespace boost::asio::ip;
    
void boost::asio::ip::validate(any& dst, const std::vector<std::string>& options, address*, int)
{
    using namespace boost::program_options;
    validators::check_first_occurrence(dst);

    system::error_code error_code;
    const auto value = address::from_string(validators::get_single_string(options), error_code);

    if (error_code)
    {
        auto exception = validation_error{validation_error::invalid_option_value};
        exception.m_error_template += "; should be IPv4 or IPv6 address";
        throw exception;
    }
    dst = boost::any(value);
}

using namespace CdnskeyProcessor::Cfg;

namespace {

Log::Level make_severity(const std::string& str)
{
    static const std::unordered_map<std::string, Log::Level> str2level =
    {
        { "critical", Log::Level::critical },
        { "error", Log::Level::error },
        { "warning", Log::Level::warning },
        { "info", Log::Level::info },
        { "debug", Log::Level::debug },
        { "trace", Log::Level::trace }
    };
    const auto level_itr = str2level.find(str);
    if (level_itr != str2level.end())
    {
        return level_itr->second;
    }
    auto exception = boost::program_options::validation_error{
            boost::program_options::validation_error::invalid_option_value};
    exception.m_error_template.append("; "
            "available severity in descending order: critical/error/warning/info/debug/trace");
    throw exception;
}

}//namespace {anonymous}

void CdnskeyProcessor::Cfg::validate(boost::any& dst, const std::vector<std::string>& options, Log::Level*, int)
{
    boost::program_options::validators::check_first_occurrence(dst);
    dst = boost::any{make_severity(boost::program_options::validators::get_single_string(options))};
}
