/*
 * Copyright (C) 2023-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/cfg/worker.hh"

#include "config.h"

#include <boost/program_options.hpp>

#include <exception>
#include <fstream>
#include <iostream>

namespace CdnskeyProcessor {
namespace Cfg {

namespace {

struct GeneralArgs
{
    std::string config_file;
};

boost::program_options::options_description get_options(GeneralArgs& general_args)
{
    namespace po = boost::program_options;
    po::options_description options{"General options", program_options_line_length};
    options.add_options()
        ("conf,c", po::value<std::string>(&general_args.config_file)->required(),
         "configuration file")
        ("help,h",
         "display this help message");
    return options;
}

boost::program_options::options_description get_options(Worker::Amqp::Broker& cfg)
{
    namespace po = boost::program_options;
    po::options_description log_options{"AMQP Broker options", program_options_line_length};
    log_options.add_options()
        ("amqp.broker.hostname",
         po::value<std::string>(&cfg.hostname)->default_value("localhost"),
         "RabbitMQ broker hostname")
        ("amqp.broker.port", po::value<std::uint16_t>(&cfg.port), "RabbitMQ broker port")
        ("amqp.broker.vhost",
         po::value<std::string>(&cfg.vhost)->default_value("cdnskey-processor-dev"),
         "RabbitMQ broker virtual host")
        ("amqp.broker.username",
         po::value<std::string>(&cfg.username)->default_value("guest"), "RabbitMQ username")
        ("amqp.broker.password",
         po::value<std::string>(&cfg.password)->default_value("guest"), "RabbitMQ password");
    return log_options;
}

boost::program_options::options_description get_options(Worker::Amqp::Exchange& cfg)
{
    namespace po = boost::program_options;
    po::options_description log_options{"AMQP Exchange options", program_options_line_length};
    log_options.add_options()
        ("amqp.exchange.insecure_fanout",
         po::value<std::string>(&cfg.insecure_fanout)->default_value("exchange_insecure_fanout"),
         "AMQP Exchange for insecure(fanout) scan requests")
        ("amqp.exchange.response",
         po::value<std::string>(&cfg.response)->default_value("exchange_response"),
         "AMQP Exchange for workers' responses");
    return log_options;
}

boost::program_options::options_description get_options(Worker::Amqp::Queue& cfg)
{
    namespace po = boost::program_options;
    po::options_description log_options{"AMQP Queue options", program_options_line_length};
    log_options.add_options()
        ("amqp.queue.results",
         po::value<std::string>(&cfg.results)->default_value("queue_results"),
         "AMQP Queue for scan results (connected to 'response' exchange")
        ("amqp.queue.insecure_prefix",
         po::value<std::string>(&cfg.insecure_prefix),
         "AMQP Queue name prefix for worker to consume insecure tasks (full name appends worker's id); "
         "missing or empty value means \"do not scan insecure domains\"")
        ("amqp.queue.secure",
         po::value<std::string>(&cfg.secure),
         "AMQP Queue name prefix for worker to consume secure tasks; missing or empty value means "
         "\"do not scan secure domains\"");
    return log_options;
}

boost::program_options::options_description get_options(Worker::Amqp& cfg)
{
    namespace po = boost::program_options;
    po::options_description options{"AMQP General options", program_options_line_length};
    options.add_options()
        ("amqp.prefetch_count",
         po::value<std::uint16_t>(&cfg.prefetch_count)->default_value(1),
         "defines the max number of unacknowledged deliveries that are permitted on a channel");
    options.add(get_options(cfg.broker))
           .add(get_options(cfg.exchange))
           .add(get_options(cfg.queue));
    return options;
}

boost::program_options::options_description get_options(Worker& cfg)
{
    namespace po = boost::program_options;
    po::options_description options{"Worker options", program_options_line_length};
    options.add_options()
        ("worker.name", po::value<std::string>(&cfg.name)->required(), "Worker's name")
        ("worker.scanner", po::value<std::string>(&cfg.scanner)->required(), "Scanner path")
        ("worker.scanner_args", po::value<std::vector<std::string>>(&cfg.scanner_args)->multitoken(),
         "Scanner arguments");
    options.add(get_options(cfg.amqp));
    return options;
}

}//namespace CdnskeyProcessor::Cfg::{anonymous}

}//namespace CdnskeyProcessor::Cfg
}//namespace CdnskeyProcessor

using namespace CdnskeyProcessor::Cfg;

Worker CdnskeyProcessor::Cfg::get_cfg(
        int argc,
        char** argv,
        void(*set_up_logging)(const Log&))
{
    const auto file_name = [&]()
    {
        GeneralArgs general_args;
        auto all_options = get_options(general_args);
        boost::program_options::variables_map vm;
        boost::program_options::store(boost::program_options::parse_command_line(argc, argv, all_options), vm);
        if (vm.count("help"))
        {
            Worker cfg;
            Log log_cfg;
            all_options.add(get_options(log_cfg))
                       .add(get_options(cfg));
            std::cout << all_options << std::flush;
            throw Worker::Help{};
        }
        notify(vm);
        return general_args.config_file;
    }();
    auto cfg_file = std::ifstream{file_name};
    if (!cfg_file.is_open())
    {
        struct UnableToOpen : std::exception
        {
            const char* what() const noexcept override { return "unable to open config file"; }
        };
        throw UnableToOpen{};
    }
    Worker cfg;
    Log log_cfg;
    boost::program_options::options_description all_options{"cdnskey-processor-worker options"};
    all_options.add(get_options(log_cfg))
               .add(get_options(cfg));
    boost::program_options::variables_map vm;
    boost::program_options::store(boost::program_options::parse_config_file(cfg_file, all_options), vm);
    boost::program_options::notify(vm);
    if (set_up_logging != nullptr)
    {
        set_up_logging(log_cfg);
    }
    return cfg;
}
