/*
 * Copyright (C) 2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/grpc_service_report.hh"
#include "src/database.hh"
#include "src/storage.hh"
#include "src/unwrapper/unwrapper.hh"
#include "src/wrapper/wrapper.hh"

#include <liblog/liblog.hh>

#include <libpg/pg_exception.hh>

#include <sstream>
#include <utility>

template <typename Char>
struct fmt::formatter<LibPg::PgConnection::ExecResult::ErrorCode, Char>
{
    template <typename ParseContext>
    constexpr auto parse(ParseContext& ctx)
    {
        return ctx.begin();
    }
    template <typename FormatContext>
    auto format(const LibPg::PgConnection::ExecResult::ErrorCode& value, FormatContext& ctx) const -> decltype(ctx.out())
    {
        std::ostringstream out;
        out << value;
        return fmt::format_to(ctx.out(), "{}", out.str());
    }
};

namespace CdnskeyProcessor {

namespace {

using MaxBatchSize = LibStrong::Type<std::size_t, struct MaxBatchSizeTag_>;

class ReportService : public Api::Report::Report::Service
{
public:
    explicit ReportService();
private:
    grpc::Status raw_scan_results(
            grpc::ServerContext*,
            const Api::Report::RawScanResultsRequest*,
            grpc::ServerWriter<Api::Report::RawScanResultsReply>*) override;
};

grpc::Status on_invalid_argument(const std::string& reason)
{
    return grpc::Status{grpc::StatusCode::INVALID_ARGUMENT, reason};
}

grpc::Status on_database_error(const std::string& reason)
{
    return grpc::Status{grpc::StatusCode::FAILED_PRECONDITION, "Database error: " + reason};
}

grpc::Status on_unexpected_error()
{
    return grpc::Status{grpc::StatusCode::UNKNOWN, "UnexpectedServerError"};
}

grpc::Status on_grpc_stream_error()
{
    return grpc::Status{grpc::StatusCode::CANCELLED, "GrpcStreamError"};
}

ReportService::ReportService()
{ }

grpc::Status ReportService::raw_scan_results(
        grpc::ServerContext* context,
        const Api::Report::RawScanResultsRequest* api_request,
            grpc::ServerWriter<Api::Report::RawScanResultsReply>* api_reply_writer)
{
    LIBLOG_SET_CONTEXT(Context, ctx, __func__);
    try
    {
        const auto request = Unwrapper::make_unwrapper(*api_request);
        auto tx = make_ro_transaction();
        const auto results = [&request, &tx]()
        {
            if (request.has_domain_fqdn())
            {
                return CdnskeyProcessor::get_raw_scan_results(tx, request.domain_fqdn());
            }
            if (request.has_domain_uid())
            {
                return CdnskeyProcessor::get_raw_scan_results(tx, request.domain_uid());
            }
            throw std::runtime_error{"no domain specified"};
        }();
        commit(std::move(tx));
        Api::Report::RawScanResultsReply reply;
        auto wrapper = Wrapper::make_wrapper(reply);
        LIBLOG_TRACE("{} scan results", results.size());
        int write_count = 0;
        for (auto chunk_begin = begin(results); chunk_begin != end(results);)
        {
            static constexpr auto max_chunk_size = 1'000;
            auto chunk_end = chunk_begin;
            for (auto chunk_size = 0;
                 (chunk_size < max_chunk_size) && (chunk_end != end(results));
                 ++chunk_end, ++chunk_size)
            { }
            wrapper.set_items(chunk_begin, chunk_end);
            const auto write_options = chunk_end != end(results) ? ::grpc::WriteOptions{}
                                                                 : ::grpc::WriteOptions{}.set_last_message();
            const auto stream_ok = api_reply_writer->Write(reply, write_options);
            if (!stream_ok)
            {
                LIBLOG_INFO("Write() failed: bad stream");
                return on_grpc_stream_error();
            }
            if (context->IsCancelled())
            {
                LIBLOG_INFO("cancelled");
                return grpc::Status::CANCELLED;
            }
            ++write_count;
            chunk_begin = chunk_end;
        }
        LIBLOG_TRACE("{} successful writes", write_count);
        return grpc::Status::OK;
    }
    // database exceptions
    catch (const InvalidArgument& e)
    {
        LIBLOG_INFO("invalid argument: {}", e.what());
        return on_invalid_argument(e.what());
    }
    // database exceptions
    catch (const LibPg::ConnectFailure& e)
    {
        LIBLOG_ERROR("database connection failure: {}", e.what());
        return on_database_error("unable to connect");
    }
    catch (const LibPg::ExecFailure& e)
    {
        LIBLOG_ERROR("database exec failure {}: {}", e.error_code(), e.what());
        return on_database_error("exec failure");
    }
    catch (const LibPg::NoConnectionAvailable& e)
    {
        LIBLOG_ERROR("no connection to database is available: {}", e.what());
        return on_database_error("no connection available");
    }
    catch (const LibPg::PgException& e)
    {
        LIBLOG_ERROR("error during database operation: {}", e.what());
        return on_database_error("unexpected");
    }
    // general error
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("unexpected error occurred: {}", e.what());
        return on_unexpected_error();
    }
    catch (...)
    {
        LIBLOG_ERROR("unexpected error occurred");
        return on_unexpected_error();
    }
}

} // namespace CdnskeyProcessor::{anonymous}

} // namespace CdnskeyProcessor

using namespace CdnskeyProcessor;

GrpcServiceReport CdnskeyProcessor::make_grpc_service_report(const Cfg::Api::Report&)
{
    return std::make_unique<ReportService>();
}
