/*
 * Copyright (C) 2023-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/grpc_service_supervisor.hh"
#include "src/database.hh"
#include "src/storage.hh"
#include "src/unwrapper/unwrapper.hh"
#include "src/wrapper/wrapper.hh"

#include <liblog/liblog.hh>

#include <libpg/pg_exception.hh>

#include <utility>

namespace CdnskeyProcessor {

namespace {

using MaxBatchSize = LibStrong::Type<std::size_t, struct MaxBatchSizeTag_>;

class SupervisorService : public Api::Supervisor::Supervisor::Service
{
public:
    explicit SupervisorService(
            MaxBatchSize max_batch_size,
            MaxScanPeriod max_scan_period,
            MinPeriodFixedCdnskey min_period_fixed_cdnskey);
private:
    using ApiScanTasksRequest = Api::Supervisor::ScanTaskCollectionRequest;
    grpc::Status import_tasks(
            grpc::ServerContext* context,
            grpc::ServerReader<ApiScanTasksRequest>* request_reader,
            google::protobuf::Empty*) override;

    grpc::Status acceptable_dnskeys(
            grpc::ServerContext*,
            const google::protobuf::Empty*,
            grpc::ServerWriter<Api::Supervisor::AcceptableDnskeysReply>*) override;

    grpc::Status add_worker(
            grpc::ServerContext*,
            const Api::Supervisor::AddWorkerRequest*,
            google::protobuf::Empty*) override;

    grpc::Status remove_worker(
            grpc::ServerContext*,
            const Api::Supervisor::RemoveWorkerRequest*,
            google::protobuf::Empty*) override;

    std::size_t max_batch_size_;
    MaxScanPeriod max_scan_period_;
    MinPeriodFixedCdnskey min_period_fixed_cdnskey_;
};

class AddNameserver
{
public:
    explicit AddNameserver(std::set<Nameserver>& set) : set_{set} { }
    void operator()(Nameserver&& nameserver) const
    {
        set_.insert(std::move(nameserver));
    }
private:
    std::set<Nameserver>& set_;
};

ScanTask make_scan_task(const Unwrapper::ScanTask& scan_task)
{
    std::set<Nameserver> nameservers;
    if (scan_task.domain().scan_type() == ScanType::insecure)
    {
        scan_task.nameservers(AddNameserver{nameservers});
    }
    return ScanTask{scan_task.domain().fqdn(), scan_task.domain().uid(), nameservers};
}

class AddScanTask
{
public:
    explicit AddScanTask(ScanTasksSet& set) : set_{set} { }
    void operator()(const Unwrapper::ScanTask& scan_task) const
    {
        set_.insert(make_scan_task(scan_task));
    }
private:
    ScanTasksSet& set_;
};

grpc::Status on_database_error(const std::string& reason)
{
    return grpc::Status{grpc::StatusCode::FAILED_PRECONDITION, "Database error: " + reason};
}

grpc::Status on_unexpected_error()
{
    return grpc::Status{grpc::StatusCode::UNKNOWN, "UnexpectedServerError"};
}

grpc::Status on_grpc_stream_error()
{
    return grpc::Status{grpc::StatusCode::CANCELLED, "GrpcStreamError"};
}

SupervisorService::SupervisorService(
        MaxBatchSize max_batch_size,
        MaxScanPeriod max_scan_period,
        MinPeriodFixedCdnskey min_period_fixed_cdnskey)
    : max_batch_size_{*max_batch_size},
      max_scan_period_{max_scan_period},
      min_period_fixed_cdnskey_{min_period_fixed_cdnskey}
{ }

grpc::Status SupervisorService::import_tasks(
        grpc::ServerContext* context,
        grpc::ServerReader<ApiScanTasksRequest>* request_reader,
        google::protobuf::Empty*)
{
    LIBLOG_SET_CONTEXT(Context, ctx, __func__);
    try
    {
        ApiScanTasksRequest request;
        auto db_conn = make_connection();
        const auto scan_queue_id = [&db_conn]()
        {
            auto tx = LibPg::PgRwTransaction{std::move(db_conn)};
            auto id = open_scan_queue(tx);
            db_conn = commit(std::move(tx));
            return id;
        }();
        LIBLOG_INFO("start to read");
        while (request_reader->Read(&request))
        {
            if (context->IsCancelled())
            {
                LIBLOG_INFO("cancelled");
                return grpc::Status::CANCELLED;
            }
            LIBLOG_TRACE("chunk read");
            ScanTasksSet scan_tasks;
            Unwrapper::make_unwrapper(request).items(AddScanTask{scan_tasks});
            LIBLOG_TRACE("chunk [{}, {}].{} unwrapped", *scan_tasks.begin()->domain_fqdn, *scan_tasks.rbegin()->domain_fqdn, scan_tasks.size());
            auto tx = LibPg::PgRwTransaction{std::move(db_conn)};
            add_scan_tasks(tx, scan_queue_id, scan_tasks);
            db_conn = commit(std::move(tx));
            LIBLOG_TRACE("read next");
        }
        LIBLOG_INFO("finish to read");
        {
            auto tx = LibPg::PgRwTransaction{std::move(db_conn)};
            clean_old_records(tx, max_scan_period_, min_period_fixed_cdnskey_);
            db_conn = commit(std::move(tx));
            vacuum(db_conn);
        }
        auto tx = LibPg::PgRwTransaction{std::move(db_conn)};
        close_scan_queue(tx, scan_queue_id);
        const auto number_of_batches = split_into_batches(tx, scan_queue_id, max_batch_size_);
        LIBLOG_INFO("scan queue splited into {} batches", number_of_batches);
        commit(std::move(tx));
        return grpc::Status::OK;
    }
    // database exceptions
    catch (const LibPg::ConnectFailure& e)
    {
        LIBLOG_ERROR("database connection failure: {}", e.what());
        return on_database_error("unable to connect");
    }
    catch (const LibPg::ExecFailure& e)
    {
        LIBLOG_ERROR("database exec failure: {}", e.what());
        return on_database_error("exec failure");
    }
    catch (const LibPg::NoConnectionAvailable& e)
    {
        LIBLOG_ERROR("no connection to database is available: {}", e.what());
        return on_database_error("no connection available");
    }
    catch (const LibPg::PgException& e)
    {
        LIBLOG_ERROR("error during database operation: {}", e.what());
        return on_database_error("unexpected");
    }
    // general error
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("unexpected error occurred: {}", e.what());
        return on_unexpected_error();
    }
    catch (...)
    {
        LIBLOG_ERROR("unexpected error occurred");
        return on_unexpected_error();
    }
}

grpc::Status SupervisorService::acceptable_dnskeys(
        grpc::ServerContext* context,
        const google::protobuf::Empty*,
        grpc::ServerWriter<Api::Supervisor::AcceptableDnskeysReply>* reply_writer)
{
    LIBLOG_SET_CONTEXT(Context, ctx, __func__);
    try
    {
        auto tx = make_ro_transaction();
        const auto cdnskeys = CdnskeyProcessor::get_acceptable_cdnskey_records(
                tx,
                max_scan_period_,
                min_period_fixed_cdnskey_);
        commit(std::move(tx));
        Api::Supervisor::AcceptableDnskeysReply reply;
        auto wrapper = Wrapper::make_wrapper(reply);
        LIBLOG_TRACE("{} cdnskeys to export", cdnskeys.size());
        int write_count = 0;
        for (auto chunk_begin = begin(cdnskeys); chunk_begin != end(cdnskeys);)
        {
            static constexpr auto max_chunk_size = 1'000;
            auto chunk_end = chunk_begin;
            for (auto chunk_size = 0;
                 (chunk_size < max_chunk_size) && (chunk_end != end(cdnskeys));
                 ++chunk_end, ++chunk_size)
            { }
            wrapper.set_items(chunk_begin, chunk_end);
            const auto write_options = chunk_end != end(cdnskeys) ? ::grpc::WriteOptions{}
                                                                  : ::grpc::WriteOptions{}.set_last_message();
            const auto stream_ok = reply_writer->Write(reply, write_options);
            if (!stream_ok)
            {
                LIBLOG_INFO("Write() failed: bad stream");
                return on_grpc_stream_error();
            }
            if (context->IsCancelled())
            {
                LIBLOG_INFO("cancelled");
                return grpc::Status::CANCELLED;
            }
            ++write_count;
            chunk_begin = chunk_end;
        }
        LIBLOG_TRACE("{} successful writes", write_count);
        return grpc::Status::OK;
    }
    // database exceptions
    catch (const LibPg::ConnectFailure& e)
    {
        LIBLOG_ERROR("database connection failure: {}", e.what());
        return on_database_error("unable to connect");
    }
    catch (const LibPg::ExecFailure& e)
    {
        LIBLOG_ERROR("database exec failure: {}", e.what());
        return on_database_error("exec failure");
    }
    catch (const LibPg::NoConnectionAvailable& e)
    {
        LIBLOG_ERROR("no connection to database is available: {}", e.what());
        return on_database_error("no connection available");
    }
    catch (const LibPg::PgException& e)
    {
        LIBLOG_ERROR("error during database operation: {}", e.what());
        return on_database_error("unexpected");
    }
    // general error
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("unexpected error occurred: {}", e.what());
        return on_unexpected_error();
    }
    catch (...)
    {
        LIBLOG_ERROR("unexpected error occurred");
        return on_unexpected_error();
    }
}

bool is_valid_name(const WorkerName& name)
{
    static constexpr char allowed_characters[] =
        "abcdefghijklmnopqrstuvwxyz"
        "0123456789"
        "-_.:";
    static constexpr auto number_of_allowed_characters = sizeof(allowed_characters) - 1;
    static_assert(0 < number_of_allowed_characters && number_of_allowed_characters <= 96, "unexpected number of allowed characters");
    return name->find_first_not_of(allowed_characters, 0, number_of_allowed_characters) == std::string::npos;
}

grpc::Status SupervisorService::add_worker(
        grpc::ServerContext*,
        const Api::Supervisor::AddWorkerRequest* request,
        google::protobuf::Empty*)
{
    LIBLOG_SET_CONTEXT(Context, ctx, __func__);
    try
    {
        const auto worker_name = Unwrapper::make_unwrapper(*request).worker_name();
        if (worker_name->empty())
        {
            static constexpr char message[] = "worker name should not be empty";
            LIBLOG_DEBUG("{}", message);
            return grpc::Status{grpc::StatusCode::INVALID_ARGUMENT, "InvalidArgument", message};
        }
        if (!is_valid_name(worker_name))
        {
            static constexpr char message[] = "worker name should contain only uppercase letters, digits, hyphen, underscore, period or colon";
            LIBLOG_DEBUG("{}", message);
            return grpc::Status{grpc::StatusCode::INVALID_ARGUMENT, "InvalidArgument", message};
        }
        auto tx = make_rw_transaction();
        const bool success = CdnskeyProcessor::add_worker(tx, worker_name);
        commit(std::move(tx));
        if (!success)
        {
            static constexpr char message[] = "worker with a given name already exists";
            LIBLOG_DEBUG("{}", message);
            return grpc::Status{grpc::StatusCode::ALREADY_EXISTS, "AlreadyExists", message};
        }
        return grpc::Status::OK;
    }
    // database exceptions
    catch (const LibPg::ConnectFailure& e)
    {
        LIBLOG_ERROR("database connection failure: {}", e.what());
        return on_database_error("unable to connect");
    }
    catch (const LibPg::ExecFailure& e)
    {
        LIBLOG_ERROR("database exec failure: {}", e.what());
        return on_database_error("exec failure");
    }
    catch (const LibPg::NoConnectionAvailable& e)
    {
        LIBLOG_ERROR("no connection to database is available: {}", e.what());
        return on_database_error("no connection available");
    }
    catch (const LibPg::PgException& e)
    {
        LIBLOG_ERROR("error during database operation: {}", e.what());
        return on_database_error("unexpected");
    }
    // general error
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("unexpected error occurred: {}", e.what());
        return on_unexpected_error();
    }
    catch (...)
    {
        LIBLOG_ERROR("unexpected error occurred");
        return on_unexpected_error();
    }
}

grpc::Status SupervisorService::remove_worker(
        grpc::ServerContext*,
        const Api::Supervisor::RemoveWorkerRequest* request,
        google::protobuf::Empty*)
{
    LIBLOG_SET_CONTEXT(Context, ctx, __func__);
    try
    {
        const auto worker_name = Unwrapper::make_unwrapper(*request).worker_name();
        auto tx = make_rw_transaction();
        const bool success = CdnskeyProcessor::remove_worker(tx, worker_name);
        commit(std::move(tx));
        if (!success)
        {
            static constexpr char message[] = "no active worker with a given name";
            LIBLOG_DEBUG("{}", message);
            return grpc::Status{grpc::StatusCode::NOT_FOUND, "DoesNotExist", message};
        }
        return grpc::Status::OK;
    }
    // database exceptions
    catch (const LibPg::ConnectFailure& e)
    {
        LIBLOG_ERROR("database connection failure: {}", e.what());
        return on_database_error("unable to connect");
    }
    catch (const LibPg::ExecFailure& e)
    {
        LIBLOG_ERROR("database exec failure: {}", e.what());
        return on_database_error("exec failure");
    }
    catch (const LibPg::NoConnectionAvailable& e)
    {
        LIBLOG_ERROR("no connection to database is available: {}", e.what());
        return on_database_error("no connection available");
    }
    catch (const LibPg::PgException& e)
    {
        LIBLOG_ERROR("error during database operation: {}", e.what());
        return on_database_error("unexpected");
    }
    // general error
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("unexpected error occurred: {}", e.what());
        return on_unexpected_error();
    }
    catch (...)
    {
        LIBLOG_ERROR("unexpected error occurred");
        return on_unexpected_error();
    }
}

} // namespace CdnskeyProcessor::{anonymous}

} // namespace CdnskeyProcessor

using namespace CdnskeyProcessor;

GrpcServiceSupervisor CdnskeyProcessor::make_grpc_service_supervisor(const Cfg::Api::Supervisor& cfg)
{
    return std::make_unique<SupervisorService>(
            MaxBatchSize{cfg.batch_size},
            MaxScanPeriod{cfg.max_scan_period},
            MinPeriodFixedCdnskey{cfg.min_period_fixed_cdnskey});
}
