/*
 * Copyright (C) 2023-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/internal/subprocess.hh"

#include <cerrno>
#include <cstring>
#include <utility>

namespace CdnskeyProcessor {

namespace {

template <class Base>
class Exception : public Base
{
public:
    explicit Exception(std::string message)
        : Base{},
          message_{std::move(message)}
    { }
private:
    const char* what() const noexcept override { return message_.c_str(); }
    std::string message_;
};

using ParentException = Exception<SubprocessException>;
using ChildException = Exception<SubprocessChildException>;

}//namespace CdnskeyProcessor::{anonymous}

Subprocess::Pipe::Pipe()
{
    errno = 0;
    if (pipe(fd_) != 0)
    {
        struct PipeFailed : ParentException
        {
            explicit PipeFailed(int error_code)
                : ParentException{std::string{"pipe() failed ("} + std::strerror(error_code) + ")"}
            { }
        };
        throw PipeFailed{errno};
    }
}

Subprocess::Pipe::~Pipe()
{
    close();
}

int Subprocess::Pipe::read_fd() const
{
    return fd_[0];
}

int Subprocess::Pipe::write_fd() const
{
    return fd_[1];
}

void Subprocess::Pipe::close()
{
    ::close(read_fd());
    ::close(write_fd());
}

Subprocess::Subprocess(const std::vector<const char*>& argv)
{
    parent_pid_ = ::getpid();
    errno = 0;
    child_pid_ = ::fork();
    if (child_pid_ == -1)
    {
        struct ForkFailed : ParentException
        {
            explicit ForkFailed(int error_code)
                : ParentException{std::string{"fork failed ("} + std::strerror(error_code) + ")"}
            { }
        };
        throw ForkFailed{errno};
    }
    if (child_pid_ == 0)
    {
        /* child */
        ::dup2(write_pipe_.read_fd(), STDIN_FILENO);
        ::dup2(read_pipe_.write_fd(), STDOUT_FILENO);

        write_pipe_.close();
        read_pipe_.close();

        errno = 0;
        ::execv(argv[0], const_cast<char**>(argv.data()));
        struct ExecvFailed : ChildException
        {
            explicit ExecvFailed(int error_code)
                : ChildException{std::string{"execv failed ("} + std::strerror(error_code) + ")"}
            { }
        };
        throw ExecvFailed{errno};
    }
    ::close(write_pipe_.read_fd());
    ::close(read_pipe_.write_fd());
}

pid_t Subprocess::parent_pid() const
{
    return parent_pid_;
}

pid_t Subprocess::child_pid() const
{
    return child_pid_;
}

int Subprocess::wait()
{
    int child_status;
    errno = 0;
    const pid_t exited_child = waitpid(child_pid_, &child_status, 0);
    if (exited_child != child_pid_)
    {
        struct WaitPidFailed : ParentException
        {
            explicit WaitPidFailed(int error_code)
                : ParentException{std::string{"waitpid failed ("} + std::strerror(error_code) + ")"}
            { }
        };
        throw WaitPidFailed{errno};
    }
    const bool child_exited_successfully = WIFEXITED(child_status) && (WEXITSTATUS(child_status) == EXIT_SUCCESS);
    if (!child_exited_successfully)
    {
        struct ChildFailed : SubprocessException
        {
            const char* what() const noexcept override { return "child terminated unsuccesfully"; }
        };
        throw ChildFailed{};
    }
    return WEXITSTATUS(child_status);
}

ssize_t Subprocess::write(const std::string& _buffer)
{
    if (0 < _buffer.size())
    {
        const ssize_t ret = ::write(write_pipe_.write_fd(), _buffer.c_str(), _buffer.size());
        if (ret == -1)
        {
            struct WriteFailed : SubprocessException
            {
                const char* what() const noexcept override { return "write to pipe failed"; }
            };
            throw WriteFailed{};
        }
        if (static_cast<size_t>(ret) != _buffer.size())
        {
            struct WriteIncomplete : SubprocessException
            {
                const char* what() const noexcept override { return "write to pipe incomplete?"; }
            };
            throw WriteIncomplete{};
        }
        return ret;
    }
    return 0;
}

void Subprocess::write_eof()
{
    ::close(write_pipe_.write_fd());
}

ssize_t Subprocess::read(ReadBuffer& _buffer)
{
    return ::read(read_pipe_.read_fd(), _buffer.data(), _buffer.size());
}

}//namespace CdnskeyProcessor
