/*
 * Copyright (C) 2023-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/internal/timer.hh"

#include <liblog/liblog.hh>

#include <algorithm>
#include <exception>
#include <utility>

namespace CdnskeyProcessor {
namespace Internal {

Timer::Timer(int concurrency_hint)
    : service_{concurrency_hint}
{ }

Timer::~Timer()
{
    try
    {
        LIBLOG_INFO("{} timeouts rest", scheduler_.size());
    }
    catch (...) { }
}

Timer::CallbackHandle Timer::register_callback(TimePoint scheduled, Callback callback)
{
    LIBLOG_DEBUG("register callback function triggered by time limited polling loop");
    auto& callback_list = scheduler_[scheduled];
    return CallbackHandle{scheduled, callback_list.insert(end(callback_list), std::move(callback))};
}

void Timer::erase_callback(const CallbackHandle& callback_handle)
{
    auto scheduler_iter = scheduler_.find(callback_handle.scheduled);
    if (scheduler_iter != end(scheduler_))
    {
        scheduler_iter->second.erase(callback_handle.callback_list_iter);
        if (scheduler_iter->second.empty())
        {
            scheduler_.erase(scheduler_iter);
        }
    }
}

void Timer::loop()
{
    while (!service_.stopped())
    {
        if (scheduler_.empty())
        {
            LIBLOG_DEBUG("run infinity polling loop");
            service_.run();
            LIBLOG_DEBUG("infinity polling loop done");
            continue;
        }
        LIBLOG_DEBUG("run time limited polling loop");
        service_.run_until(begin(scheduler_)->first);
        LIBLOG_DEBUG("time limited polling finished");
        const auto now = std::chrono::steady_clock::now();
        const auto first_in_the_future = std::find_if(begin(scheduler_), end(scheduler_), [&](auto&& scheduled)
        {
            if (now < scheduled.first)
            {
                return true;
            }
            std::for_each(begin(scheduled.second), end(scheduled.second), [&scheduled](auto&& callback)
            {
                try
                {
                    callback(scheduled.first);
                }
                catch (const std::exception& e)
                {
                    LIBLOG_ERROR("scheduled callback failure: {}", e.what());
                }
                catch (...)
                {
                    LIBLOG_ERROR("scheduled callback failure");
                }
            });
            return false;
        });
        scheduler_.erase(begin(scheduler_), first_in_the_future);
    }
}

boost::asio::io_context& Timer::service()
{
    return service_;
}

} // namespace CdnskeyProcessor::Internal
} // namespace CdnskeyProcessor
