/*
 * Copyright (C) 2023-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/scan_task.hh"
#include "src/unwrapper/unwrapper.hh"
#include "src/wrapper/wrapper.hh"

#include <stdexcept>

using namespace CdnskeyProcessor;

std::string CdnskeyProcessor::serialize(const SecureScanBatch& batch)
{
    Api::Internal::ScanBatchSecure proto;
    Wrapper::make_wrapper(proto).set(batch);
    return proto.SerializeAsString();
}

NumberOfQueries CdnskeyProcessor::estimated_size(const SecureScanBatch& batch)
{
    return NumberOfQueries{batch.jobs.size()};
}

std::string CdnskeyProcessor::serialize(const InsecureScanBatch& batch)
{
    Api::Internal::ScanBatchInsecure proto;
    Wrapper::make_wrapper(proto).set(batch);
    return proto.SerializeAsString();
}

NumberOfQueries CdnskeyProcessor::estimated_size(const InsecureScanBatch& batch)
{
    std::size_t size = batch.jobs.size();
    std::for_each(begin(batch.jobs), end(batch.jobs), [&size](auto&& ns_to_jobs)
    {
        size += ns_to_jobs.second.size();
    });
    return NumberOfQueries{size};
}

bool CdnskeyProcessor::does_contain_result(CdnskeyStatus status)
{
    switch (status)
    {
        case CdnskeyStatus::insecure:
        case CdnskeyStatus::secure:
            return true;
        case CdnskeyStatus::insecure_empty:
        case CdnskeyStatus::secure_empty:
        case CdnskeyStatus::untrustworthy:
        case CdnskeyStatus::unknown:
        case CdnskeyStatus::unresolved:
        case CdnskeyStatus::unresolved_ip:
            return false;
    }
    throw std::invalid_argument{"CdnskeyStatus value out of range"};
}

bool CdnskeyProcessor::is_successful(CdnskeyStatus status)
{
    switch (status)
    {
        case CdnskeyStatus::insecure:
        case CdnskeyStatus::secure:
        case CdnskeyStatus::insecure_empty:
        case CdnskeyStatus::secure_empty:
            return true;
        case CdnskeyStatus::untrustworthy:
        case CdnskeyStatus::unknown:
        case CdnskeyStatus::unresolved:
        case CdnskeyStatus::unresolved_ip:
            return false;
    }
    throw std::invalid_argument{"CdnskeyStatus value out of range"};
}

template <>
InsecureScanBatch CdnskeyProcessor::deserialize(const char* msg, std::size_t size)
{
    Api::Internal::ScanBatchInsecure proto;
    proto.ParseFromArray(msg, size);
    const auto unwrapper = Unwrapper::make_unwrapper(proto);
    return InsecureScanBatch
    {
        unwrapper.id(),
        unwrapper.scan_tasks()
    };
}

template <>
SecureScanBatch CdnskeyProcessor::deserialize(const char* msg, std::size_t size)
{
    Api::Internal::ScanBatchSecure proto;
    proto.ParseFromArray(msg, size);
    const auto unwrapper = Unwrapper::make_unwrapper(proto);
    return SecureScanBatch
    {
        unwrapper.id(),
        unwrapper.scan_tasks()
    };
}
