/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/storage.hh"

#include <liblog/liblog.hh>

#include <libpg/query.hh>
#include <libpg/sql_state/code.hh>

#include <algorithm>
#include <cstring>
#include <exception>
#include <tuple>
#include <utility>

namespace boost {
namespace asio {
namespace ip {

std::string wrap_into_psql_representation(const boost::asio::ip::address& addr)
{
    return addr.to_string();
}

address unwrap_from_psql_representation(const address&, const char* src)
{
    return address::from_string(src);
}

}//namespace boost::asio::ip
}//namespace boost::asio
}//namespace boost

namespace CdnskeyProcessor {

CdnskeyStatus unwrap_from_psql_representation(const CdnskeyStatus&, const char* src)
{
    const auto is_src = [src](const char* status_name)
    {
        return std::strcmp(src, status_name) == 0;
    };
    if (is_src("insecure"))
    {
        return CdnskeyStatus::insecure;
    }
    if (is_src("insecure-empty"))
    {
        return CdnskeyStatus::insecure_empty;
    }
    if (is_src("secure"))
    {
        return CdnskeyStatus::secure;
    }
    if (is_src("secure-empty"))
    {
        return CdnskeyStatus::secure_empty;
    }
    if (is_src("untrustworthy"))
    {
        return CdnskeyStatus::untrustworthy;
    }
    if (is_src("unknown"))
    {
        return CdnskeyStatus::unknown;
    }
    if (is_src("unresolved"))
    {
        return CdnskeyStatus::unresolved;
    }
    if (is_src("unresolved-ip"))
    {
        return CdnskeyStatus::unresolved_ip;
    }
    throw std::runtime_error{"invalid value of CdnskeyStatus"};
}

}//namespace CdnskeyProcessor

using namespace CdnskeyProcessor;

namespace {

template <int> struct DomainFqdnAtTag_;
template <int idx>
using DomainFqdnAt = StrongString<DomainFqdnAtTag_<idx>>;
template <int idx>
using DomainFqdnOptionalAt = typename DomainFqdnAt<idx>::Optional;

template <int> struct DomainUidAtTag_;
template <int idx>
using DomainUidAt = StrongString<DomainUidAtTag_<idx>>;
template <int idx>
using DomainUidOptionalAt = typename DomainUidAt<idx>::Optional;

template <int, int> struct NameserverAtTag_;
template <int domain_idx, int nameserver_idx>
using NameserverAt = StrongString<NameserverAtTag_<domain_idx, nameserver_idx>>;
template <int domain_idx, int nameserver_idx>
using NameserverOptionalAt = typename NameserverAt<domain_idx, nameserver_idx>::Optional;

template <int idx>
using NssetAt = std::tuple<
        NameserverOptionalAt<idx, 0>,
        NameserverOptionalAt<idx, 1>,
        NameserverOptionalAt<idx, 2>,
        NameserverOptionalAt<idx, 3>,
        NameserverOptionalAt<idx, 4>,
        NameserverOptionalAt<idx, 5>,
        NameserverOptionalAt<idx, 6>,
        NameserverOptionalAt<idx, 7>,
        NameserverOptionalAt<idx, 8>,
        NameserverOptionalAt<idx, 9>>;

using InsecureScanTuple = std::tuple<
        DomainFqdnOptionalAt<0>, DomainUidOptionalAt<0>, NssetAt<0>,
        DomainFqdnOptionalAt<1>, DomainUidOptionalAt<1>, NssetAt<1>,
        DomainFqdnOptionalAt<2>, DomainUidOptionalAt<2>, NssetAt<2>,
        DomainFqdnOptionalAt<3>, DomainUidOptionalAt<3>, NssetAt<3>>;

using SecureScanTuple = std::tuple<
        DomainFqdnOptionalAt<0>, DomainUidOptionalAt<0>,
        DomainFqdnOptionalAt<1>, DomainUidOptionalAt<1>,
        DomainFqdnOptionalAt<2>, DomainUidOptionalAt<2>,
        DomainFqdnOptionalAt<3>, DomainUidOptionalAt<3>,
        DomainFqdnOptionalAt<4>, DomainUidOptionalAt<4>,
        DomainFqdnOptionalAt<5>, DomainUidOptionalAt<5>,
        DomainFqdnOptionalAt<6>, DomainUidOptionalAt<6>,
        DomainFqdnOptionalAt<7>, DomainUidOptionalAt<7>,
        DomainFqdnOptionalAt<8>, DomainUidOptionalAt<8>,
        DomainFqdnOptionalAt<9>, DomainUidOptionalAt<9>,
        DomainFqdnOptionalAt<10>, DomainUidOptionalAt<10>,
        DomainFqdnOptionalAt<11>, DomainUidOptionalAt<11>,
        DomainFqdnOptionalAt<12>, DomainUidOptionalAt<12>,
        DomainFqdnOptionalAt<13>, DomainUidOptionalAt<13>,
        DomainFqdnOptionalAt<14>, DomainUidOptionalAt<14>,
        DomainFqdnOptionalAt<15>, DomainUidOptionalAt<15>>;

template <int idx, typename T>
decltype(auto) get_fqdn(T&& tuple)
{
    return std::get<DomainFqdnOptionalAt<idx>>(std::forward<T>(tuple));
}

template <int idx, typename T>
decltype(auto) get_uid(T&& tuple)
{
    return std::get<DomainUidOptionalAt<idx>>(std::forward<T>(tuple));
}

template <int domain_idx, int nameserver_idx, typename T>
decltype(auto) get_nameserver(T&& tuple)
{
    return std::get<NameserverOptionalAt<domain_idx, nameserver_idx>>(
            std::get<NssetAt<domain_idx>>(std::forward<T>(tuple)));
}

template <int idx, int max, template<int> typename ItemAt>
struct FirstMissing
{
    template <typename T>
    static auto item_idx(const T& tuple)
    {
        const bool is_missing = std::get<typename ItemAt<idx>::Optional>(tuple) == ItemAt<idx>::nullopt;
        return is_missing ? idx
                          : FirstMissing<idx + 1, max, ItemAt>::item_idx(tuple);
    }
};

template <int max, template<int> typename ItemAt>
struct FirstMissing<max, max, ItemAt>
{
    template <typename T>
    static constexpr auto item_idx(T&&)
    {
        return max;
    }
};

static_assert(std::tuple_size<SecureScanTuple>::value % 2 == 0, "invalid length of SecureScanTuple");
constexpr auto number_of_secure_domains = std::tuple_size<SecureScanTuple>::value / 2;

auto get_first_missing(const SecureScanTuple& tuple)
{
    return FirstMissing<0, number_of_secure_domains, DomainFqdnAt>::item_idx(tuple);
}

template <int domain_idx>
struct NameserversAtRow
{
    template <int nameserver_idx>
    using ItemAt = NameserverAt<domain_idx, nameserver_idx>;
};

template <int idx>
auto get_first_missing(const NssetAt<idx>& nsset)
{
    static constexpr auto number_of_nameservers = std::tuple_size<NssetAt<idx>>::value;
    return FirstMissing<0, number_of_nameservers, NameserversAtRow<idx>::template ItemAt>::item_idx(nsset);
}

static_assert(std::tuple_size<InsecureScanTuple>::value % 3 == 0, "invalid length of InsecureScanTuple");
constexpr auto number_of_insecure_domains = std::tuple_size<InsecureScanTuple>::value / 3;

template <int> struct NumberOfNameservers;

template <>
struct NumberOfNameservers<number_of_insecure_domains>
{
    static constexpr auto get(const InsecureScanTuple&)
    {
        return 0;
    }
};

template <int idx>
struct NumberOfNameservers
{
    static auto get(const InsecureScanTuple& tuple)
    {
        const bool domain_is_missing = std::get<DomainFqdnOptionalAt<idx>>(tuple) == DomainFqdnAt<idx>::nullopt;
        return domain_is_missing
                ? 0
                : get_first_missing<idx>(std::get<NssetAt<idx>>(tuple)) + NumberOfNameservers<idx + 1>::get(tuple);
    }
};

auto get_number_of_nameservers(const InsecureScanTuple& tuple)
{
    return NumberOfNameservers<0>::get(tuple);
}

void add_scan(const LibPg::PgRwTransaction& tx, const ScanQueueId& scan_queue_id, const SecureScanTuple& scan)
{
    const auto db_res = exec(tx, LibPg::make_query(
            "WITH od AS " // old domain
            "("
                "SELECT domain.id, d.fqdn, d.uid "
                  "FROM (VALUES (") << LibPg::parameter<DomainFqdnOptionalAt<0>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<0>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<1>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<1>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<2>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<2>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<3>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<3>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<4>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<4>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<5>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<5>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<6>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<6>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<7>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<7>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<8>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<8>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<9>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<9>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<10>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<10>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<11>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<11>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<12>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<12>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<13>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<13>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<14>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<14>>().as_text() << "), "
                                "(" << LibPg::parameter<DomainFqdnOptionalAt<15>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<15>>().as_text() << ")) d(fqdn, uid) "
             "LEFT JOIN domain ON domain.fqdn = d.fqdn AND "
                                 "domain.uid = d.uid "
                 "WHERE d.fqdn IS NOT NULL AND "
                       "d.uid IS NOT NULL"
            "), "
            "nd AS " // new domain
            "("
                "INSERT INTO domain (fqdn, uid) "
                     "SELECT fqdn, uid "
                       "FROM od "
                      "WHERE id IS NULL "
                  "RETURNING id, fqdn, uid"
            "), "
            "d AS " // domain (old and new together)
            "("
                "SELECT * FROM od WHERE id IS NOT NULL "
                "UNION "
                "SELECT * FROM nd"
            ") "
            "INSERT INTO scan (scan_queue_id, domain_id) "
                 "SELECT " << LibPg::parameter<ScanQueueId>().as_big_int() << ", "
                        "id "
                   "FROM d "
              "RETURNING" << LibPg::item<std::size_t>() << "id",
            scan_queue_id,
            get_fqdn<0>(scan), get_uid<0>(scan),
            get_fqdn<1>(scan), get_uid<1>(scan),
            get_fqdn<2>(scan), get_uid<2>(scan),
            get_fqdn<3>(scan), get_uid<3>(scan),
            get_fqdn<4>(scan), get_uid<4>(scan),
            get_fqdn<5>(scan), get_uid<5>(scan),
            get_fqdn<6>(scan), get_uid<6>(scan),
            get_fqdn<7>(scan), get_uid<7>(scan),
            get_fqdn<8>(scan), get_uid<8>(scan),
            get_fqdn<9>(scan), get_uid<9>(scan),
            get_fqdn<10>(scan), get_uid<10>(scan),
            get_fqdn<11>(scan), get_uid<11>(scan),
            get_fqdn<12>(scan), get_uid<12>(scan),
            get_fqdn<13>(scan), get_uid<13>(scan),
            get_fqdn<14>(scan), get_uid<14>(scan),
            get_fqdn<15>(scan), get_uid<15>(scan));
    if (db_res.size() != LibPg::RowIndex{get_first_missing(scan)})
    {
        struct UnexpectedResult : std::exception
        {
            const char* what() const noexcept override { return "unexpected number of inserted scan rows"; }
        };
        throw UnexpectedResult{};
    }
}

void add_scan(const LibPg::PgRwTransaction& tx, const ScanQueueId& scan_queue_id, const InsecureScanTuple& scan)
{
    const auto db_res = exec(tx, LibPg::make_query(
            "WITH od AS " // old domain
            "("
                "SELECT domain.id, d.fqdn, d.uid, d.ns "
                  "FROM (VALUES (") << LibPg::parameter<DomainFqdnOptionalAt<0>>().as_text() << ", "
                                    << LibPg::parameter<DomainUidOptionalAt<0>>().as_text() << ", "
                                "ARRAY[" << LibPg::parameter<NameserverOptionalAt<0, 0>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<0, 1>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<0, 2>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<0, 3>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<0, 4>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<0, 5>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<0, 6>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<0, 7>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<0, 8>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<0, 9>>().as_text() << "]), "
                               "(" << LibPg::parameter<DomainFqdnOptionalAt<1>>().as_text() << ", "
                                   << LibPg::parameter<DomainUidOptionalAt<1>>().as_text() << ", "
                                "ARRAY[" << LibPg::parameter<NameserverOptionalAt<1, 0>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<1, 1>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<1, 2>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<1, 3>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<1, 4>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<1, 5>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<1, 6>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<1, 7>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<1, 8>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<1, 9>>().as_text() << "]), "
                               "(" << LibPg::parameter<DomainFqdnOptionalAt<2>>().as_text() << ", "
                                   << LibPg::parameter<DomainUidOptionalAt<2>>().as_text() << ", "
                                "ARRAY[" << LibPg::parameter<NameserverOptionalAt<2, 0>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<2, 1>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<2, 2>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<2, 3>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<2, 4>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<2, 5>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<2, 6>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<2, 7>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<2, 8>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<2, 9>>().as_text() << "]), "
                               "(" << LibPg::parameter<DomainFqdnOptionalAt<3>>().as_text() << ", "
                                   << LibPg::parameter<DomainUidOptionalAt<3>>().as_text() << ", "
                                "ARRAY[" << LibPg::parameter<NameserverOptionalAt<3, 0>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<3, 1>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<3, 2>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<3, 3>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<3, 4>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<3, 5>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<3, 6>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<3, 7>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<3, 8>>().as_text() << ", "
                                         << LibPg::parameter<NameserverOptionalAt<3, 9>>().as_text() << "])"
                       ") d(fqdn, uid, ns) "
             "LEFT JOIN domain ON domain.fqdn = d.fqdn AND "
                                 "domain.uid = d.uid "
                 "WHERE d.fqdn IS NOT NULL AND "
                       "d.uid IS NOT NULL"
            "), "
            "nd AS " // new domain
            "("
                "INSERT INTO domain (fqdn, uid) "
                     "SELECT fqdn, uid "
                       "FROM od "
                      "WHERE id IS NULL "
                  "RETURNING id, fqdn, uid"
            "), "
            "d AS " // domain (old and new together)
            "("
                "SELECT * FROM od WHERE id IS NOT NULL "
                "UNION "
                "SELECT nd.id, nd.fqdn, nd.uid, od.ns "
                  "FROM nd "
                  "JOIN od ON od.fqdn = nd.fqdn AND od.uid = nd.uid"
            "), "
            "s AS " // scan
            "("
                "INSERT INTO scan (scan_queue_id, domain_id) "
                     "SELECT " << LibPg::parameter<ScanQueueId>().as_big_int() << ", "
                            "id "
                       "FROM d "
                  "RETURNING id, domain_id"
            "), "
            "ons AS " // old nameserver
            "("
                "SELECT n.id, ans.fqdn "
                  "FROM (SELECT DISTINCT UNNEST(ns) AS fqdn FROM od) ans "
             "LEFT JOIN nameserver n ON n.fqdn = ans.fqdn "
                 "WHERE ans.fqdn IS NOT NULL"
            "), "
            "nns AS " // new nameserver
            "("
                "INSERT INTO nameserver (fqdn) SELECT fqdn FROM ons WHERE id IS NULL "
                  "RETURNING id, fqdn"
            "), "
            "ns AS " // nameserver (old and new together)
            "("
                "SELECT * FROM ons WHERE id IS NOT NULL "
                "UNION "
                "SELECT * FROM nns"
            ") "
            "INSERT INTO insecure_scan (scan_id, nameserver_id) "
                 "SELECT s.id, ns.id "
                   "FROM (SELECT id, fqdn, uid, UNNEST(ns) "
                           "FROM d) AS dns(domain_id, domain_fqdn, uid, ns_fqdn) "
                   "JOIN s ON s.domain_id = dns.domain_id "
                   "JOIN ns ON ns.fqdn = dns.ns_fqdn",
            scan_queue_id,
            get_fqdn<0>(scan), get_uid<0>(scan), get_nameserver<0, 0>(scan), get_nameserver<0, 1>(scan),
                                                 get_nameserver<0, 2>(scan), get_nameserver<0, 3>(scan),
                                                 get_nameserver<0, 4>(scan), get_nameserver<0, 5>(scan),
                                                 get_nameserver<0, 6>(scan), get_nameserver<0, 7>(scan),
                                                 get_nameserver<0, 8>(scan), get_nameserver<0, 9>(scan),
            get_fqdn<1>(scan), get_uid<1>(scan), get_nameserver<1, 0>(scan), get_nameserver<1, 1>(scan),
                                                 get_nameserver<1, 2>(scan), get_nameserver<1, 3>(scan),
                                                 get_nameserver<1, 4>(scan), get_nameserver<1, 5>(scan),
                                                 get_nameserver<1, 6>(scan), get_nameserver<1, 7>(scan),
                                                 get_nameserver<1, 8>(scan), get_nameserver<1, 9>(scan),
            get_fqdn<2>(scan), get_uid<2>(scan), get_nameserver<2, 0>(scan), get_nameserver<2, 1>(scan),
                                                 get_nameserver<2, 2>(scan), get_nameserver<2, 3>(scan),
                                                 get_nameserver<2, 4>(scan), get_nameserver<2, 5>(scan),
                                                 get_nameserver<2, 6>(scan), get_nameserver<2, 7>(scan),
                                                 get_nameserver<2, 8>(scan), get_nameserver<2, 9>(scan),
            get_fqdn<3>(scan), get_uid<3>(scan), get_nameserver<3, 0>(scan), get_nameserver<3, 1>(scan),
                                                 get_nameserver<3, 2>(scan), get_nameserver<3, 3>(scan),
                                                 get_nameserver<3, 4>(scan), get_nameserver<3, 5>(scan),
                                                 get_nameserver<3, 6>(scan), get_nameserver<3, 7>(scan),
                                                 get_nameserver<3, 8>(scan), get_nameserver<3, 9>(scan));
    if (db_res.get_number_of_rows_affected() != LibPg::RowIndex{get_number_of_nameservers(scan)})
    {
        struct UnexpectedResult : std::exception
        {
            const char* what() const noexcept override { return "unexpected number of inserted scan rows"; }
        };
        throw UnexpectedResult{};
    }
}

template <int domain_idx, int nameserver_idx>
void add_nameserver(NssetAt<domain_idx>& nsset, const Nameserver& nameserver)
{
    std::get<NameserverOptionalAt<domain_idx, nameserver_idx>>(nsset) =
            NameserverAt<domain_idx, nameserver_idx>(*nameserver);
}

template <int domain_idx>
void fill_nameservers(NssetAt<domain_idx>& nsset, const std::set<Nameserver>& nameservers)
{
    int nameserver_idx = 0;
    std::for_each(begin(nameservers), end(nameservers), [&](auto&& ns)
    {
        static_assert(std::tuple_size<NssetAt<domain_idx>>::value == 10, "invalid size of NssetAt<domain_idx>");
        switch (nameserver_idx++)
        {
            case 0: return add_nameserver<domain_idx, 0>(nsset, ns);
            case 1: return add_nameserver<domain_idx, 1>(nsset, ns);
            case 2: return add_nameserver<domain_idx, 2>(nsset, ns);
            case 3: return add_nameserver<domain_idx, 3>(nsset, ns);
            case 4: return add_nameserver<domain_idx, 4>(nsset, ns);
            case 5: return add_nameserver<domain_idx, 5>(nsset, ns);
            case 6: return add_nameserver<domain_idx, 6>(nsset, ns);
            case 7: return add_nameserver<domain_idx, 7>(nsset, ns);
            case 8: return add_nameserver<domain_idx, 8>(nsset, ns);
            case 9: return add_nameserver<domain_idx, 9>(nsset, ns);
        }
        throw std::runtime_error{"unexpected number of nameservers"};
    });
}

template <int idx>
void add_scan(InsecureScanTuple& tuple, const ScanTask& task)
{
    get_fqdn<idx>(tuple) = DomainFqdnAt<idx>(*task.domain_fqdn);
    get_uid<idx>(tuple) = DomainUidAt<idx>(*task.domain_uid);
    fill_nameservers<idx>(std::get<NssetAt<idx>>(tuple), task.nameservers);
}

int add_scan(InsecureScanTuple& tuple, int domain_idx, const ScanTask& task)
{
    static_assert(number_of_insecure_domains == 4, "invalid number of insecure domains");
    switch (domain_idx)
    {
        case 0:
            add_scan<0>(tuple, task);
            return 1;
        case 1:
            add_scan<1>(tuple, task);
            return 2;
        case 2:
            add_scan<2>(tuple, task);
            return 3;
        case 3:
            add_scan<3>(tuple, task);
            return 0; // tuple is full
        default:
            throw std::runtime_error{"unexpected number of insecure domains"};
    }
}

template <int idx>
void add_scan(SecureScanTuple& tuple, const ScanTask& task)
{
    get_fqdn<idx>(tuple) = DomainFqdnAt<idx>(*task.domain_fqdn);
    get_uid<idx>(tuple) = DomainUidAt<idx>(*task.domain_uid);
}

int add_scan(SecureScanTuple& tuple, int domain_idx, const ScanTask& task)
{
    static_assert(number_of_secure_domains == 16, "invalid number of secure domains");
    switch (domain_idx)
    {
        case 0:
            add_scan<0>(tuple, task);
            return 1;
        case 1:
            add_scan<1>(tuple, task);
            return 2;
        case 2:
            add_scan<2>(tuple, task);
            return 3;
        case 3:
            add_scan<3>(tuple, task);
            return 4;
        case 4:
            add_scan<4>(tuple, task);
            return 5;
        case 5:
            add_scan<5>(tuple, task);
            return 6;
        case 6:
            add_scan<6>(tuple, task);
            return 7;
        case 7:
            add_scan<7>(tuple, task);
            return 8;
        case 8:
            add_scan<8>(tuple, task);
            return 9;
        case 9:
            add_scan<9>(tuple, task);
            return 10;
        case 10:
            add_scan<10>(tuple, task);
            return 11;
        case 11:
            add_scan<11>(tuple, task);
            return 12;
        case 12:
            add_scan<12>(tuple, task);
            return 13;
        case 13:
            add_scan<13>(tuple, task);
            return 14;
        case 14:
            add_scan<14>(tuple, task);
            return 15;
        case 15:
            add_scan<15>(tuple, task);
            return 0; // tuple is full
        default:
            throw std::runtime_error{"unexpected number of secure domains"};
    }
}

std::int64_t store_insecure_result(
        const LibPg::PgRwTransaction& tx,
        ScanBatchId batch_id,
        const WorkerName& worker,
        ScanJobId job_id,
        const Nameserver& nameserver,
        const IpAddress& nameserver_ip,
        const Cdnskey& record)
{
    const auto dbres = exec(tx, LibPg::make_query(
            "WITH d AS " // data
            "("
                "SELECT w.id AS worker_id, "
                       "s.domain_id, "
                       "i.nameserver_id "
                  "FROM worker w "
                  "JOIN scan_batch_progress p ON p.worker_id = w.id "
                  "JOIN scan_job j ON j.scan_batch_id = p.scan_batch_id "
                  "JOIN scan s ON s.id = j.scan_id "
                  "JOIN insecure_scan i ON i.scan_id = s.id "
                  "JOIN nameserver n ON n.id = i.nameserver_id "
                 "WHERE w.name = ") << LibPg::parameter<WorkerName>().as_text() << " AND "
                       "w.removed_at IS NULL AND "
                       "p.scan_batch_id = " << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                       "j.id = " << LibPg::parameter<ScanJobId>().as_big_int() << " AND "
                       "n.fqdn = " << LibPg::parameter<Nameserver>().as_text() <<
            ") "
            "INSERT INTO scan_result "
            "("
                "scan_job_id, "
                "worker_id, "
                "nameserver_id, "
                "nameserver_ip, "
                "scan_status, "
                "cdnskey_flags, "
                "cdnskey_proto, "
                "cdnskey_alg, "
                "cdnskey_public_key"
            ") SELECT " << LibPg::parameter<ScanJobId>().as_big_int() << ", "
                     "d.worker_id, "
                     "d.nameserver_id, "
                        << LibPg::parameter<IpAddress>().as_type("INET") << ", "
                     "'insecure'::cdnskey_status, "
                        << LibPg::parameter<CdnskeyFlags>().as_integer() << ", "
                        << LibPg::parameter<CdnskeyProto>().as_integer() << ", "
                        << LibPg::parameter<CdnskeyAlg>().as_integer() << ", "
                        << LibPg::parameter<CdnskeyPublicKey>().as_text() << " "
                "FROM d "
              "RETURNING" << LibPg::item<std::int64_t>() << "id",
            batch_id,
            worker,
            job_id,
            nameserver,
            nameserver_ip,
            *record.flags,
            *record.proto,
            *record.alg,
            *record.public_key);
    if (dbres.size() == LibPg::RowIndex{1})
    {
        return dbres.front().get<std::int64_t>();
    }
    if (dbres.empty())
    {
        throw std::runtime_error{"no job joinable with this result"};
    }
    throw std::runtime_error{"too many jobs joinable with this result"};
}

std::int64_t store_insecure_result(
        const LibPg::PgRwTransaction& tx,
        ScanBatchId batch_id,
        const WorkerName& worker,
        ScanJobId job_id,
        const Nameserver& nameserver,
        const IpAddress& nameserver_ip,
        const char* cdnskey_status)
{
    const auto dbres = exec(tx, LibPg::make_query(
            "WITH d AS " // data
            "("
                "SELECT w.id AS worker_id, "
                       "s.domain_id, "
                       "i.nameserver_id "
                  "FROM worker w "
                  "JOIN scan_batch_progress p ON p.worker_id = w.id "
                  "JOIN scan_job j ON j.scan_batch_id = p.scan_batch_id "
                  "JOIN scan s ON s.id = j.scan_id "
                  "JOIN insecure_scan i ON i.scan_id = s.id "
                  "JOIN nameserver n ON n.id = i.nameserver_id "
                 "WHERE w.name = ") << LibPg::parameter<WorkerName>().as_text() << " AND "
                       "w.removed_at IS NULL AND "
                       "p.scan_batch_id = " << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                       "j.id = " << LibPg::parameter<ScanJobId>().as_big_int() << " AND "
                       "n.fqdn = " << LibPg::parameter<Nameserver>().as_text() <<
            ") "
            "INSERT INTO scan_result "
            "("
                "scan_job_id, "
                "worker_id, "
                "nameserver_id, "
                "nameserver_ip, "
                "scan_status"
            ") SELECT " << LibPg::parameter<ScanJobId>().as_big_int() << ", "
                     "d.worker_id, "
                     "d.nameserver_id, "
                        << LibPg::parameter<IpAddress>().as_type("INET") << ", "
                        << LibPg::parameter<const char*>().as_type("cdnskey_status") << " "
                "FROM d "
              "RETURNING" << LibPg::item<std::int64_t>() << "id",
            batch_id,
            worker,
            job_id,
            nameserver,
            nameserver_ip,
            cdnskey_status);
    if (dbres.size() == LibPg::RowIndex{1})
    {
        return dbres.front().get<std::int64_t>();
    }
    if (dbres.empty())
    {
        throw std::runtime_error{"no job joinable with this result"};
    }
    throw std::runtime_error{"too many jobs joinable with this result"};
}

std::int64_t store_secure_result(
        const LibPg::PgRwTransaction& tx,
        ScanBatchId batch_id,
        const WorkerName& worker,
        ScanJobId job_id,
        const Cdnskey& record)
{
    const auto dbres = exec(tx, LibPg::make_query(
            "WITH d AS " // data
            "("
                "SELECT w.id AS worker_id, "
                       "s.domain_id "
                  "FROM worker w "
                  "JOIN scan_batch_progress p ON p.worker_id = w.id "
                  "JOIN scan_job j ON j.scan_batch_id = p.scan_batch_id "
                  "JOIN scan s ON s.id = j.scan_id "
                 "WHERE w.name = ") << LibPg::parameter<WorkerName>().as_text() << " AND "
                       "w.removed_at IS NULL AND "
                       "p.scan_batch_id = " << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                       "j.id = " << LibPg::parameter<ScanJobId>().as_big_int() <<
            ") "
            "INSERT INTO scan_result "
            "("
                "scan_job_id, "
                "worker_id, "
                "scan_status, "
                "cdnskey_flags, "
                "cdnskey_proto, "
                "cdnskey_alg, "
                "cdnskey_public_key"
            ") SELECT " << LibPg::parameter<ScanJobId>().as_big_int() << ", "
                     "d.worker_id, "
                     "'secure'::cdnskey_status, "
                        << LibPg::parameter<CdnskeyFlags>().as_integer() << ", "
                        << LibPg::parameter<CdnskeyProto>().as_integer() << ", "
                        << LibPg::parameter<CdnskeyAlg>().as_integer() << ", "
                        << LibPg::parameter<CdnskeyPublicKey>().as_text() << " "
                "FROM d "
              "RETURNING" << LibPg::item<std::int64_t>() << "id",
            batch_id,
            worker,
            job_id,
            *record.flags,
            *record.proto,
            *record.alg,
            *record.public_key);
    if (dbres.size() == LibPg::RowIndex{1})
    {
        return dbres.front().get<std::int64_t>();
    }
    if (dbres.empty())
    {
        throw std::runtime_error{"no job joinable with this result"};
    }
    throw std::runtime_error{"too many jobs joinable with this result"};
}

std::int64_t store_secure_result(
        const LibPg::PgRwTransaction& tx,
        ScanBatchId batch_id,
        const WorkerName& worker,
        ScanJobId job_id,
        const char* cdnskey_status)
{
    const auto dbres = exec(tx, LibPg::make_query(
            "WITH d AS " // data
            "("
                "SELECT w.id AS worker_id, "
                       "s.domain_id "
                  "FROM worker w "
                  "JOIN scan_batch_progress p ON p.worker_id = w.id "
                  "JOIN scan_job j ON j.scan_batch_id = p.scan_batch_id "
                  "JOIN scan s ON s.id = j.scan_id "
                 "WHERE w.name = ") << LibPg::parameter<WorkerName>().as_text() << " AND "
                       "w.removed_at IS NULL AND "
                       "p.scan_batch_id = " << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                       "j.id = " << LibPg::parameter<ScanJobId>().as_big_int() <<
            ") "
            "INSERT INTO scan_result "
            "("
                "scan_job_id, "
                "worker_id, "
                "scan_status"
            ") SELECT " << LibPg::parameter<ScanJobId>().as_big_int() << ", "
                     "d.worker_id, "
                        << LibPg::parameter<const char*>().as_type("cdnskey_status") << " "
                "FROM d "
              "RETURNING" << LibPg::item<std::int64_t>() << "id",
            batch_id,
            worker,
            job_id,
            cdnskey_status);
    if (dbres.size() == LibPg::RowIndex{1})
    {
        return dbres.front().get<std::int64_t>();
    }
    if (dbres.empty())
    {
        throw std::runtime_error{"no job joinable with this result"};
    }
    throw std::runtime_error{"too many jobs joinable with this result"};
}

std::int64_t store_unknown_result(
        const LibPg::PgRwTransaction& tx,
        ScanBatchId batch_id,
        const WorkerName& worker,
        ScanJobId job_id)
{
    const auto dbres = exec(tx, LibPg::make_query(
            "WITH d AS " // data
            "("
                "SELECT w.id AS worker_id, "
                       "s.domain_id "
                  "FROM worker w "
                  "JOIN scan_batch_progress p ON p.worker_id = w.id "
                  "JOIN scan_job j ON j.scan_batch_id = p.scan_batch_id "
                  "JOIN scan s ON s.id = j.scan_id "
                 "WHERE w.name = ") << LibPg::parameter<WorkerName>().as_text() << " AND "
                       "w.removed_at IS NULL AND "
                       "p.scan_batch_id = " << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                       "j.id = " << LibPg::parameter<ScanJobId>().as_big_int() <<
            ") "
            "INSERT INTO scan_result "
            "("
                "scan_job_id, "
                "worker_id, "
                "scan_status"
            ") SELECT " << LibPg::parameter<ScanJobId>().as_big_int() << ", "
                     "d.worker_id, "
                     "'unknown'::cdnskey_status "
                "FROM d "
              "RETURNING" << LibPg::item<std::int64_t>() << "id",
            batch_id,
            worker,
            job_id);
    if (dbres.size() == LibPg::RowIndex{1})
    {
        return dbres.front().get<std::int64_t>();
    }
    if (dbres.empty())
    {
        throw std::runtime_error{"no job joinable with this result"};
    }
    throw std::runtime_error{"too many jobs joinable with this result"};
}

std::int64_t store_unresolved_ip_result(
        const LibPg::PgRwTransaction& tx,
        ScanBatchId batch_id,
        const WorkerName& worker,
        ScanJobId job_id,
        const Nameserver& nameserver)
{
    const auto dbres = exec(tx, LibPg::make_query(
            "WITH d AS " // data
            "("
                "SELECT w.id AS worker_id, "
                       "s.domain_id, "
                       "i.nameserver_id "
                  "FROM worker w "
                  "JOIN scan_batch_progress p ON p.worker_id = w.id "
                  "JOIN scan_job j ON j.scan_batch_id = p.scan_batch_id "
                  "JOIN scan s ON s.id = j.scan_id "
                  "JOIN insecure_scan i ON i.scan_id = s.id "
                  "JOIN nameserver n ON n.id = i.nameserver_id "
                 "WHERE w.name = ") << LibPg::parameter<WorkerName>().as_text() << " AND "
                       "w.removed_at IS NULL AND "
                       "p.scan_batch_id = " << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                       "j.id = " << LibPg::parameter<ScanJobId>().as_big_int() << " AND "
                       "n.fqdn = " << LibPg::parameter<Nameserver>().as_text() <<
            ") "
            "INSERT INTO scan_result "
            "("
                "scan_job_id, "
                "worker_id, "
                "nameserver_id, "
                "scan_status"
            ") SELECT " << LibPg::parameter<ScanJobId>().as_big_int() << ", "
                     "d.worker_id, "
                     "d.nameserver_id, "
                     "'unresolved-ip'::cdnskey_status "
                "FROM d "
              "RETURNING" << LibPg::item<std::int64_t>() << "id",
            batch_id,
            worker,
            job_id,
            nameserver);
    if (dbres.size() == LibPg::RowIndex{1})
    {
        return dbres.front().get<std::int64_t>();
    }
    if (dbres.empty())
    {
        throw std::runtime_error{"no job joinable with this result"};
    }
    throw std::runtime_error{"too many jobs joinable with this result"};
}

void finish_batch(
        const LibPg::PgRwTransaction& tx,
        ScanBatchId batch_id,
        const WorkerName& worker,
        const Timestamp& finished_at)
{
    const auto db_res = exec(tx, LibPg::make_query(
            "UPDATE scan_batch_progress "
               "SET finished_at = ") << LibPg::parameter<Timestamp>().as_timestamp() << ", "
                   "worker_id = (SELECT id "
                                  "FROM worker w "
                                 "WHERE w.name = " << LibPg::parameter<WorkerName>().as_text() << " AND "
                                       "w.removed_at IS NULL) "
             "WHERE scan_batch_id = " << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                   "finished_at IS NULL",
            batch_id,
            worker,
            finished_at);
    if (db_res.get_number_of_rows_affected() == LibPg::RowIndex{0})
    {
        struct UnexpectedResult : std::exception
        {
            const char* what() const noexcept override { return "no batch to finish found"; }
        };
        throw UnexpectedResult{};
    }
    if (LibPg::RowIndex{1} < db_res.get_number_of_rows_affected())
    {
        struct UnexpectedResult : std::exception
        {
            const char* what() const noexcept override { return "too many batches to finish found"; }
        };
        throw UnexpectedResult{};
    }
}

void clean_scan_result(
        const LibPg::PgRwTransaction& tx,
        MaxScanPeriod max_scan_period,
        MinPeriodFixedCdnskey min_period_fixed_cdnskey)
{
    exec(tx, LibPg::make_query(
            "WITH s AS " // scan
            "("
                "SELECT s.id "
                  "FROM scan_queue q "
                  "JOIN scan s ON s.scan_queue_id = q.id "
                 "WHERE q.import_started_at < (NOW() - ") <<
                                               LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") << " - " <<
                                               LibPg::parameter<MinPeriodFixedCdnskey>().as_type("INTERVAL") << " - " <<
                                               LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") <<
                                             ")::TIMESTAMP"
            ") "
            "DELETE FROM scan_result r "
                  "USING s, "
                        "scan_job j "
                  "WHERE j.scan_id = s.id AND "
                        "r.scan_job_id = j.id",
            max_scan_period,
            min_period_fixed_cdnskey);
}

void clean_scan_job(
        const LibPg::PgRwTransaction& tx,
        MaxScanPeriod max_scan_period,
        MinPeriodFixedCdnskey min_period_fixed_cdnskey)
{
    exec(tx, LibPg::make_query(
            "WITH s AS " // scan
            "("
                "SELECT s.id "
                  "FROM scan_queue q "
                  "JOIN scan s ON s.scan_queue_id = q.id "
                 "WHERE q.import_started_at < (NOW() - ") <<
                                               LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") << " - " <<
                                               LibPg::parameter<MinPeriodFixedCdnskey>().as_type("INTERVAL") << " - " <<
                                               LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") <<
                                             ")::TIMESTAMP"
            ") "
            "DELETE FROM scan_job j "
                  "USING s "
                  "WHERE j.scan_id = s.id",
            max_scan_period,
            min_period_fixed_cdnskey);
}

void clean_insecure_scan(
        const LibPg::PgRwTransaction& tx,
        MaxScanPeriod max_scan_period,
        MinPeriodFixedCdnskey min_period_fixed_cdnskey)
{
    exec(tx, LibPg::make_query(
            "WITH s AS " // scan
            "("
                "SELECT s.id "
                  "FROM scan_queue q "
                  "JOIN scan s ON s.scan_queue_id = q.id "
                 "WHERE q.import_started_at < (NOW() - ") <<
                                               LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") << " - " <<
                                               LibPg::parameter<MinPeriodFixedCdnskey>().as_type("INTERVAL") << " - " <<
                                               LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") <<
                                             ")::TIMESTAMP"
            ") "
            "DELETE FROM insecure_scan i "
                  "USING s "
                  "WHERE i.scan_id = s.id",
            max_scan_period,
            min_period_fixed_cdnskey);
}

void clean_scan(
        const LibPg::PgRwTransaction& tx,
        MaxScanPeriod max_scan_period,
        MinPeriodFixedCdnskey min_period_fixed_cdnskey)
{
    exec(tx, LibPg::make_query(
            "WITH s AS " // scan
            "("
                "SELECT s.id "
                  "FROM scan_queue q "
                  "JOIN scan s ON s.scan_queue_id = q.id "
                 "WHERE q.import_started_at < (NOW() - ") <<
                                               LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") << " - " <<
                                               LibPg::parameter<MinPeriodFixedCdnskey>().as_type("INTERVAL") << " - " <<
                                               LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") <<
                                             ")::TIMESTAMP"
            ") "
            "DELETE FROM scan "
                  "USING s "
                  "WHERE scan.id = s.id",
            max_scan_period,
            min_period_fixed_cdnskey);
}

void clean_scan_queue(
        const LibPg::PgRwTransaction& tx,
        MaxScanPeriod max_scan_period,
        MinPeriodFixedCdnskey min_period_fixed_cdnskey)
{
    clean_scan_result(tx, max_scan_period, min_period_fixed_cdnskey);
    clean_scan_job(tx, max_scan_period, min_period_fixed_cdnskey);
    clean_insecure_scan(tx, max_scan_period, min_period_fixed_cdnskey);
    clean_scan(tx, max_scan_period, min_period_fixed_cdnskey);
    exec(tx, LibPg::make_query(
            "DELETE FROM scan_queue "
                  "WHERE import_started_at < (NOW() - ") <<
                                              LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") << " - " <<
                                              LibPg::parameter<MinPeriodFixedCdnskey>().as_type("INTERVAL") << " - " <<
                                              LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") <<
                                            ")::TIMESTAMP",
            max_scan_period,
            min_period_fixed_cdnskey);
}

void clean_scan_batch_progress(const LibPg::PgRwTransaction& tx)
{
    exec(tx, LibPg::make_query(
            "WITH b AS " // batch
            "("
                "SELECT id "
                  "FROM scan_batch b "
                 "WHERE NOT EXISTS(SELECT FROM scan_job j WHERE j.scan_batch_id = b.id)"
            ") "
            "DELETE FROM scan_batch_progress bp "
                  "USING b "
                  "WHERE bp.scan_batch_id = b.id"));
}

void clean_scan_batch(const LibPg::PgRwTransaction& tx)
{
    exec(tx, LibPg::make_query(
            "WITH b AS " // batch
            "("
                "SELECT id "
                  "FROM scan_batch b "
                 "WHERE NOT EXISTS(SELECT FROM scan_job j WHERE j.scan_batch_id = b.id)"
            ") "
            "DELETE FROM scan_batch sb "
                  "USING b "
                  "WHERE sb.id = b.id"));
}

void clean_domain(const LibPg::PgRwTransaction& tx)
{
    exec(tx, LibPg::make_query(
            "DELETE FROM domain d "
                  "WHERE NOT EXISTS(SELECT FROM scan s WHERE s.domain_id = d.id)"));
}

void clean_unused(const LibPg::PgRwTransaction& tx)
{
    clean_scan_batch_progress(tx);
    clean_scan_batch(tx);
    clean_domain(tx);
    exec(tx, LibPg::make_query(
            "DELETE FROM nameserver n "
                  "WHERE NOT EXISTS(SELECT FROM insecure_scan i WHERE i.nameserver_id = n.id) AND "
                        "NOT EXISTS(SELECT FROM scan_result r WHERE r.nameserver_id = n.id)"));
}

}//namespace {anonymous}

ScanQueueId CdnskeyProcessor::open_scan_queue(const LibPg::PgRwTransaction& tx)
{
    const auto db_res = exec(tx, LibPg::make_query(
            "INSERT INTO scan_queue DEFAULT VALUES "
              "RETURNING") << LibPg::item<ScanQueueId>() << "id");
    if (db_res.size() != LibPg::RowIndex{1})
    {
        struct UnexpectedResult : std::exception
        {
            const char* what() const noexcept override { return "unexpected number of inserted scan_queue rows"; }
        };
        throw UnexpectedResult{};
    }
    return db_res.front().get<ScanQueueId>();
}

void CdnskeyProcessor::close_scan_queue(const LibPg::PgRwTransaction& tx, ScanQueueId scan_queue_id)
{
    const auto db_res = exec(tx, LibPg::make_query(
            "UPDATE scan_queue "
               "SET import_finished_at = CURRENT_TIMESTAMP "
             "WHERE id = ") << LibPg::parameter<ScanQueueId>().as_big_int() << " AND "
                   "import_finished_at IS NULL", scan_queue_id);
    if (db_res.get_number_of_rows_affected() == LibPg::RowIndex{0})
    {
        struct UnexpectedResult : std::exception
        {
            const char* what() const noexcept override { return "no open scan queue found"; }
        };
        throw UnexpectedResult{};
    }
    if (LibPg::RowIndex{1} < db_res.get_number_of_rows_affected())
    {
        struct UnexpectedResult : std::exception
        {
            const char* what() const noexcept override { return "too many scan queues found"; }
        };
        throw UnexpectedResult{};
    }
}

bool CdnskeyProcessor::LessByDomainFqdn::operator()(const ScanTask& lhs, const ScanTask& rhs) const noexcept
{
    return lhs.domain_fqdn < rhs.domain_fqdn;
}

std::size_t CdnskeyProcessor::split_into_batches(
        const LibPg::PgRwTransaction& tx,
        ScanQueueId queue_id,
        std::size_t batch_size)
{
    const auto db_res = exec(
            tx,
            LibPg::make_query(
                    "WITH i AS "// insecure
                    "("
                        "SELECT s.id AS scan_id, "
                               "n.id AS nameserver_id "
                          "FROM scan_queue q "
                          "JOIN scan s ON s.scan_queue_id = q.id "
                          "JOIN insecure_scan i ON i.scan_id = s.id "
                          "JOIN nameserver n ON n.id = i.nameserver_id "
                         "WHERE q.import_finished_at IS NOT NULL AND "
                               "q.id = ") << LibPg::parameter<ScanQueueId>().as_big_int() <<
                    "), "
                    "b AS "// buckets
                    "("
                        "SELECT FIRST_VALUE(scan_id) OVER (PARTITION BY chunk ORDER BY scan_id) = "
                               "LAST_VALUE(scan_id) OVER (PARTITION BY chunk ORDER BY scan_id) AS is_first, "
                               "chunk, "
                               "scan_id "
                          "FROM (SELECT NTILE((SELECT GREATEST(1, COUNT(*) / " << LibPg::parameter<std::size_t>().as_big_int() << ") "
                                                "FROM i)::INT) "
                                           "OVER (ORDER BY scan_id, nameserver_id) AS chunk, "
                                       "scan_id "
                                  "FROM i) tmp"
                    "), "
                    "bs AS "// bucket statistics
                    "("
                        "SELECT chunk, "
                               "MIN(scan_id) AS first, "
                               "MAX(scan_id) AS last "
                          "FROM b "
                      "GROUP BY 1"
                    "), "
                    "bi AS "// bucket informations
                    "("
                        "SELECT b0.chunk, "
                               "COALESCE(b1.last = b0.first, FALSE) AS exclude_first "
                          "FROM bs b0 "
                     "LEFT JOIN bs b1 ON b1.chunk = b0.chunk - 1"
                    "), "
                    "ib AS "// insecure buckets
                    "("
                        "SELECT DISTINCT b.chunk, b.scan_id "
                          "FROM bi "
                          "JOIN b ON b.chunk = bi.chunk AND "
                                    "NOT (bi.exclude_first AND b.is_first)"
                    "), "
                    "s AS "// secure
                    "("
                        "SELECT s.id AS scan_id "
                          "FROM scan_queue q "
                          "JOIN scan s ON s.scan_queue_id = q.id "
                     "LEFT JOIN insecure_scan i ON i.scan_id = s.id "
                         "WHERE q.import_finished_at IS NOT NULL AND "
                               "q.id = " << LibPg::parameter<ScanQueueId>().as_big_int() << " AND "
                               "i.scan_id IS NULL"
                    "), "
                    "sb AS "// secure buckets
                    "("
                        "SELECT NTILE((SELECT GREATEST(1, COUNT(*) / " << LibPg::parameter<std::size_t>().as_big_int() << ") "
                                        "FROM s)::INT) "
                                   "OVER (ORDER BY scan_id) + "
                               "(SELECT MAX(chunk) FROM ib) AS chunk, "
                               "scan_id "
                          "FROM s"
                    "), "
                    "batch AS "
                    "("
                        "SELECT * FROM ib "
                        "UNION "
                        "SELECT * FROM sb"
                    "), "
                    "isb AS "// inserted batch
                    "("
                        "INSERT INTO scan_batch SELECT FROM batch GROUP BY chunk "
                          "RETURNING id"
                    "), "
                    "nb AS "// new batch
                    "("
                        "SELECT ROW_NUMBER() OVER (ORDER BY id) AS chunk, "
                               "id AS scan_batch_id "
                        "FROM isb"
                    "), "
                    "job AS "
                    "("
                        "INSERT INTO scan_job (scan_batch_id, scan_id) "
                             "SELECT nb.scan_batch_id, batch.scan_id "
                               "FROM batch "
                               "JOIN nb ON nb.chunk = batch.chunk "
                          "RETURNING scan_batch_id"
                    ") "
                    "SELECT" << LibPg::item<std::size_t>() << "COUNT(DISTINCT scan_batch_id) "
                      "FROM job",
                    queue_id,
                    batch_size);
    return db_res.front().get<std::size_t>();
}

void CdnskeyProcessor::add_scan_tasks(
        const LibPg::PgRwTransaction& tx,
        ScanQueueId scan_queue_id,
        const ScanTasksSet& scan_tasks)
{
    InsecureScanTuple insecure;
    SecureScanTuple secure;
    int insecure_idx = 0;
    int secure_idx = 0;
    std::for_each(begin(scan_tasks), end(scan_tasks), [&](const ScanTask& scan)
    {
        const bool is_insecure = !scan.nameservers.empty();
        if (is_insecure)
        {
            insecure_idx = add_scan(insecure, insecure_idx, scan);
            if (insecure_idx == 0)
            {
                add_scan(tx, scan_queue_id, insecure);
                insecure = InsecureScanTuple{};
            }
        }
        else
        {
            secure_idx = add_scan(secure, secure_idx, scan);
            if (secure_idx == 0)
            {
                add_scan(tx, scan_queue_id, secure);
                secure = SecureScanTuple{};
            }
        }
    });
    if (insecure_idx != 0)
    {
        add_scan(tx, scan_queue_id, insecure);
    }
    if (secure_idx != 0)
    {
        add_scan(tx, scan_queue_id, secure);
    }
}

bool CdnskeyProcessor::add_worker(const LibPg::PgRwTransaction& tx, const WorkerName& worker_name)
{
    if (worker_name->empty())
    {
        return false;
    }
    const auto query = LibPg::make_query() <<
            // clang-format off
            "INSERT INTO worker (name) "
                 "VALUES (" << LibPg::parameter<WorkerName>().as_text() << ") "
            "ON CONFLICT DO NOTHING";
            // clang-format on
    const auto cmdres = exec(tx, query, worker_name);
    return cmdres.get_number_of_rows_affected() == LibPg::RowIndex{1};
}

bool CdnskeyProcessor::remove_worker(const LibPg::PgRwTransaction& tx, const WorkerName& worker_name)
{
    const auto query = LibPg::make_query() <<
            // clang-format off
            "UPDATE worker "
               "SET removed_at = CURRENT_TIMESTAMP "
             "WHERE name = " << LibPg::parameter<WorkerName>().as_text() << " AND "
                   "removed_at IS NULL";
            // clang-format on
    const auto cmdres = exec(tx, query, worker_name);
    return cmdres.get_number_of_rows_affected() == LibPg::RowIndex{1};
}

SecureScanBatch CdnskeyProcessor::get_secure_batch_to_solve(const LibPg::PgTransaction& tx)
{
    SecureScanBatch result;
    const auto dbres = exec(tx, LibPg::make_query(
            "WITH b AS " // batch
            "("
                "SELECT MIN(b.id) AS id "
                  "FROM scan_batch b "
                 "WHERE NOT EXISTS(SELECT FROM scan_batch_progress p WHERE p.scan_batch_id = b.id) AND "
                       "NOT EXISTS(SELECT FROM scan_job j "
                                         "JOIN insecure_scan i ON i.scan_id = j.scan_id "
                                        "WHERE j.scan_batch_id = b.id)"
            ") "
            "SELECT") << LibPg::item<ScanBatchId>() << "b.id"
                      << LibPg::item<ScanJobId>() << "j.id"
                      << LibPg::item<DomainFqdn>() << "d.fqdn"
                      << LibPg::item<DomainUid>() << "d.uid "
              "FROM scan_job j "
              "JOIN scan s ON s.id = j.scan_id "
              "JOIN domain d ON d.id = s.domain_id "
              "JOIN b ON b.id = j.scan_batch_id "
          "ORDER BY j.id");
    if (!dbres.empty())
    {
        result.batch_id = dbres.front().get<ScanBatchId>();
        result.jobs.reserve(*dbres.size());
        std::for_each(begin(dbres), end(dbres), [&result](auto&& row)
        {
            result.jobs.push_back(
                    ScanJob
                    {
                        row.template get<ScanJobId>(),
                        row.template get<DomainFqdn>(),
                        row.template get<DomainUid>()
                    });
        });
    }
    return result;
}

InsecureScanBatch CdnskeyProcessor::get_insecure_batch_to_solve(const LibPg::PgTransaction& tx)
{
    InsecureScanBatch result;
    const auto dbres = exec(tx, LibPg::make_query(
            "WITH b AS " // batch
            "("
                "SELECT MIN(b.id) AS id "
                  "FROM scan_batch b "
                 "WHERE NOT EXISTS(SELECT FROM scan_batch_progress p WHERE p.scan_batch_id = b.id) AND "
                       "EXISTS(SELECT FROM scan_job j "
                                     "JOIN insecure_scan i ON i.scan_id = j.scan_id "
                                    "WHERE j.scan_batch_id = b.id)"
            ") "
            "SELECT") << LibPg::item<ScanBatchId>() << "b.id"
                      << LibPg::item<ScanJobId>() << "j.id"
                      << LibPg::item<DomainFqdn>() << "d.fqdn"
                      << LibPg::item<DomainUid>() << "d.uid"
                      << LibPg::item<Nameserver>() << "n.fqdn "
              "FROM scan_job j "
              "JOIN scan s ON s.id = j.scan_id "
              "JOIN domain d ON d.id = s.domain_id "
              "JOIN b ON b.id = j.scan_batch_id "
              "JOIN insecure_scan i ON i.scan_id = s.id "
              "JOIN nameserver n ON n.id = i.nameserver_id "
          "ORDER BY n.fqdn, j.id");
    if (!dbres.empty())
    {
        result.batch_id = dbres.front().get<ScanBatchId>();
        std::for_each(begin(dbres), end(dbres), [&result](auto&& row)
        {
            result.jobs[row.template get<Nameserver>()].push_back(
                    ScanJob
                    {
                        row.template get<ScanJobId>(),
                        row.template get<DomainFqdn>(),
                        row.template get<DomainUid>()
                    });
        });
    }
    return result;
}

void CdnskeyProcessor::batch_sent(const LibPg::PgRwTransaction& tx, ScanBatchId batch_id)
{
    const auto cmdres = exec(tx, LibPg::make_query(
            "INSERT INTO scan_batch_progress ("
                               "scan_batch_id, "
                               "worker_id, "
                               "sent_at) "
                        "SELECT b.id, "
                               "w.id, "
                               "NOW() "
                          "FROM worker w "
                    "CROSS JOIN scan_batch b "
                         "WHERE w.removed_at IS NULL AND "
                               "b.id = ") << LibPg::parameter<ScanBatchId>().as_big_int(),
            batch_id);
    if (cmdres.get_number_of_rows_affected() <= LibPg::RowIndex{0})
    {
        struct InvalidArgument : std::exception
        {
            const char* what() const noexcept override { return "no worker exists or given scan batch not found"; }
        };
        throw InvalidArgument{};
    }
}

void CdnskeyProcessor::batch_started(
        const LibPg::PgRwTransaction& tx,
        ScanBatchId batch_id)
{
    const auto cmdres = exec(tx, LibPg::make_query(
            "UPDATE scan_batch_progress "
               "SET started_at = NOW() "
             "WHERE sent_at IS NOT NULL AND "
                   "started_at IS NULL AND "
                   "finished_at IS NULL AND "
                   "scan_batch_id = ") << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                   "worker_id = (SELECT id "
                                  "FROM worker "
                                 "WHERE removed_at IS NULL)",
            batch_id);
    if (cmdres.get_number_of_rows_affected() <= LibPg::RowIndex{0})
    {
        struct InvalidArgument : std::exception
        {
            const char* what() const noexcept override { return "no suitable scan batch found"; }
        };
        throw InvalidArgument{};
    }
}

void CdnskeyProcessor::batch_started(
        const LibPg::PgRwTransaction& tx,
        ScanBatchId batch_id,
        const WorkerName& worker)
{
    const auto cmdres = exec(tx, LibPg::make_query(
            "UPDATE scan_batch_progress "
               "SET started_at = NOW() "
             "WHERE sent_at IS NOT NULL AND "
                   "started_at IS NULL AND "
                   "finished_at IS NULL AND "
                   "scan_batch_id = ") << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                   "worker_id = (SELECT id "
                                  "FROM worker "
                                 "WHERE name = " << LibPg::parameter<WorkerName>().as_text() << " AND "
                                       "removed_at IS NULL)",
            batch_id,
            worker);
    if (cmdres.get_number_of_rows_affected() <= LibPg::RowIndex{0})
    {
        struct InvalidArgument : std::exception
        {
            const char* what() const noexcept override { return "no suitable scan batch found"; }
        };
        throw InvalidArgument{};
    }
}

void CdnskeyProcessor::batch_result_received(const LibPg::PgRwTransaction& tx, const ScanResultCollection& set)
{
    class Savepoint
    {
    public:
        explicit Savepoint(const LibPg::PgRwTransaction& transaction)
            : transaction_{transaction}
        {
            exec(transaction_, LibPg::make_query("SAVEPOINT _sp_brr_"));
        }
        ~Savepoint()
        {
            try
            {
                exec(transaction_, LibPg::make_query("RELEASE SAVEPOINT _sp_brr_"));
            }
            catch (...) { }
        }
        void recovery() && noexcept
        {
            try
            {
                exec(transaction_, LibPg::make_query("ROLLBACK TO SAVEPOINT _sp_brr_"));
            }
            catch (...) { }
        }
    private:
        const LibPg::PgRwTransaction& transaction_;
    };
    std::for_each(begin(set.scan_results), end(set.scan_results), [&](auto&& result)
    {
        Savepoint savepoint{tx};
        try
        {
            switch (result.cdnskey.status)
            {
                case CdnskeyStatus::insecure:
                    store_insecure_result(
                            tx,
                            set.scan_batch_id,
                            set.worker_name,
                            result.scan_job_id,
                            *result.nameserver,
                            *result.nameserver_ip,
                            result.cdnskey);
                    return;
                case CdnskeyStatus::insecure_empty:
                    store_insecure_result(
                            tx,
                            set.scan_batch_id,
                            set.worker_name,
                            result.scan_job_id,
                            *result.nameserver,
                            *result.nameserver_ip,
                            "insecure-empty");
                    return;
                case CdnskeyStatus::secure:
                    store_secure_result(
                            tx,
                            set.scan_batch_id,
                            set.worker_name,
                            result.scan_job_id,
                            result.cdnskey);
                    return;
                case CdnskeyStatus::secure_empty:
                    store_secure_result(
                            tx,
                            set.scan_batch_id,
                            set.worker_name,
                            result.scan_job_id,
                            "secure-empty");
                    return;
                case CdnskeyStatus::untrustworthy:
                    store_secure_result(
                            tx,
                            set.scan_batch_id,
                            set.worker_name,
                            result.scan_job_id,
                            "untrustworthy");
                    return;
                case CdnskeyStatus::unknown:
                    store_unknown_result(
                            tx,
                            set.scan_batch_id,
                            set.worker_name,
                            result.scan_job_id);
                    return;
                case CdnskeyStatus::unresolved:
                    store_insecure_result(
                            tx,
                            set.scan_batch_id,
                            set.worker_name,
                            result.scan_job_id,
                            *result.nameserver,
                            *result.nameserver_ip,
                            "unresolved");
                    return;
                case CdnskeyStatus::unresolved_ip:
                    store_unresolved_ip_result(
                            tx,
                            set.scan_batch_id,
                            set.worker_name,
                            result.scan_job_id,
                            *result.nameserver);
                    return;
            }
        }
        catch (const LibPg::ExecFailure& e)
        {
            LIBLOG_WARNING("ExecFailure exception caught: {}", e.what());
            std::move(savepoint).recovery();
        }
        catch (const std::exception& e)
        {
            LIBLOG_WARNING("std::exception caught: {}", e.what());
        }
        catch (...)
        {
            LIBLOG_WARNING("unknown exception caught");
        }
    });
    if (set.finished_at != Timestamp::nullopt)
    {
        Savepoint savepoint{tx};
        try
        {
            finish_batch(tx, set.scan_batch_id, set.worker_name, *set.finished_at);
        }
        catch (const LibPg::ExecFailure& e)
        {
            LIBLOG_WARNING("ExecFailure exception caught: {}", e.what());
            std::move(savepoint).recovery();
        }
        catch (const std::exception& e)
        {
            LIBLOG_WARNING("std::exception caught: {}", e.what());
        }
        catch (...)
        {
            LIBLOG_WARNING("unknown exception caught");
        }
    }
}

AcceptableCdnskeys CdnskeyProcessor::get_acceptable_cdnskey_records(
        const LibPg::PgTransaction& tx,
        MaxScanPeriod max_scan_period,
        MinPeriodFixedCdnskey min_period_fixed_cdnskey)
{
    AcceptableCdnskeys result;
    const auto dbres = exec(tx, LibPg::make_query(
            "WITH insecure_candidate AS "// domains with at least one CDNSKEY record found in the last period and not imported as secure domain
            "("
                "SELECT DISTINCT s.domain_id "
                  "FROM scan_result r "
                  "JOIN scan_job j ON j.id = r.scan_job_id "
                  "JOIN scan s ON s.id = j.scan_id "
                 "WHERE r.scan_status = 'insecure'::cdnskey_status AND "
                       "(NOW() - ") << LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") << ")::TIMESTAMP <= r.scanned_at AND "
                       "NOT EXISTS(SELECT FROM scan ss WHERE ss.domain_id = s.domain_id AND "
                                                            "s.scan_queue_id < ss.scan_queue_id AND "// newer import
                                                            "NOT EXISTS(SELECT FROM insecure_scan i WHERE i.scan_id = ss.id)) AND "// as secure domain
                       "NOT EXISTS(SELECT FROM domain WHERE fqdn = (SELECT fqdn FROM domain WHERE id = s.domain_id) AND "// newer domain does not exist
                                                           "s.domain_id < id)"
            "), "
            "insecure_filtered AS "// filter out completely unsuccessful scans and insecure scans followed by a secure one
            "("
                "SELECT s.id AS scan_id "
                  "FROM insecure_candidate ic "
                  "JOIN scan s ON s.domain_id = ic.domain_id "
                  "JOIN insecure_scan i ON i.scan_id = s.id "
                  "JOIN scan_job j ON j.scan_id = s.id "
             "LEFT JOIN scan_result r ON r.scan_job_id = j.id AND "
                                        "r.nameserver_id = i.nameserver_id "
                 "WHERE NOT EXISTS(SELECT FROM scan ss WHERE ss.domain_id = s.domain_id AND "
                                                            "s.scan_queue_id < ss.scan_queue_id AND "// newer import
                                                            "NOT EXISTS(SELECT FROM insecure_scan i WHERE i.scan_id = ss.id)) "// as secure domain
              "GROUP BY 1 "
                "HAVING BOOL_OR(COALESCE(r.scan_status IN ('insecure'::cdnskey_status, 'insecure-empty'::cdnskey_status), FALSE))"
            "), "
            "insecure_by_ip AS "// aggregate CDNSKEY records over scan-nameserver-ip
            "("
                "SELECT s.scan_id, "
                       "MIN(r.scanned_at) AS scan_started_at, "
                       "MAX(r.scanned_at) AS scan_finished_at, "
                       "BOOL_AND(COALESCE(r.scan_status IN ('insecure'::cdnskey_status, 'insecure-empty'::cdnskey_status), FALSE)) AS is_ok, "
                       "ARRAY_AGG(ROW(r.cdnskey_flags, r.cdnskey_proto, r.cdnskey_alg, r.cdnskey_public_key) "
                                 "ORDER BY r.cdnskey_flags, r.cdnskey_proto, r.cdnskey_alg, r.cdnskey_public_key) AS keyset "
                 "FROM insecure_filtered s "
                 "JOIN insecure_scan i ON i.scan_id = s.scan_id "
                 "JOIN scan_job j ON j.scan_id = s.scan_id "
            "LEFT JOIN scan_result r ON r.scan_job_id = j.id AND "
                                       "r.nameserver_id = i.nameserver_id "
             "GROUP BY 1, i.nameserver_id, r.nameserver_ip"
            "), "
            "insecure_by_ip_class AS "// classify sequence consistency over scan
            "("
                "SELECT scan_id, "
                       "scan_started_at, "
                       "scan_finished_at, "
                       "is_ok AND "
                       "COALESCE(lead(keyset, 0) OVER s = lead(keyset, 1) OVER s, TRUE) AS is_ok, "// next scan has the same keyset => consistent scan
                       "keyset "
                  "FROM insecure_by_ip "
                "WINDOW s AS (PARTITION BY scan_id ORDER BY scan_started_at)"
            "), "
            "insecure_by_scan AS "// group scans by scan_id
            "("
                "SELECT s.domain_id, "
                       "iic.scan_id, "
                       "MIN(iic.scan_started_at) AS scan_started_at, "
                       "MAX(iic.scan_finished_at) AS scan_finished_at, "
                       "BOOL_AND(iic.is_ok) AS is_ok, "
                       "MAX(iic.keyset) AS keyset "
                  "FROM insecure_by_ip_class iic "
                  "JOIN scan s ON s.id = iic.scan_id "
              "GROUP BY 2, 1"
            "),"
            "insecure_by_scan_class AS "// classify domain scan sequence
            "("
                "SELECT domain_id, "
                       "scan_id, "
                       "is_ok AND "
                       "COALESCE((lead(scan_started_at, 1) OVER d - lead(scan_finished_at, 0) OVER d) < " << // two scans time difference
                                 LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") << ", TRUE) AS is_ok, "
                       "COALESCE((MAX(scan_finished_at) OVER (PARTITION BY domain_id) - lead(scan_started_at, 1) OVER d) < " << // time difference from next to the last scan
                                 LibPg::parameter<MinPeriodFixedCdnskey>().as_type("INTERVAL") << ", TRUE) AS is_seq_member, " <<
                        LibPg::parameter<MinPeriodFixedCdnskey>().as_type("INTERVAL") << " <= "
                        "(MAX(scan_finished_at) OVER (PARTITION BY domain_id) - scan_started_at) AS is_long_enough, "// time difference from current to the last scan
                       "COALESCE(lead(keyset, 0) OVER d = lead(keyset, 1) OVER d, TRUE) AS has_follower "// follower is next scan having the same keyset
                  "FROM insecure_by_scan "
                "WINDOW d AS (PARTITION BY domain_id ORDER BY scan_id)"
            "),"
            "insecure_acceptable AS "
            "("
                "SELECT MAX(scan_id) AS scan_id "
                  "FROM insecure_by_scan_class "
                 "WHERE is_seq_member "
              "GROUP BY domain_id "
                "HAVING BOOL_AND(is_ok AND has_follower) AND "// correct continuous sequence
                       "BOOL_OR(is_long_enough)"              // long enough
            "),"
            "secure_candidate AS "// domains with at least one CDNSKEY record found in the last period
            "("
                "SELECT DISTINCT s.domain_id "
                  "FROM scan_result r "
                  "JOIN scan_job j ON j.id = r.scan_job_id "
                  "JOIN scan s ON s.id = j.scan_id "
                 "WHERE r.scan_status = 'secure'::cdnskey_status AND "
                       "(NOW() - " << LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") << ")::TIMESTAMP <= r.scanned_at AND "
                       "NOT EXISTS(SELECT FROM scan si WHERE si.domain_id = s.domain_id AND "
                                                            "s.scan_queue_id < si.scan_queue_id AND "
                                                            "EXISTS(SELECT FROM insecure_scan i WHERE i.scan_id = si.id)) AND "
                       "NOT EXISTS(SELECT FROM domain WHERE fqdn = (SELECT fqdn FROM domain WHERE id = s.domain_id) AND "
                                                           "s.domain_id < id)"
            "), "
            "secure_result AS "
            "("
                "SELECT s.id AS scan_id, "
                       "sc.domain_id, "
                       "MAX(scanned_at) AS scan_finished_at "
                  "FROM secure_candidate sc "
                  "JOIN scan s ON s.domain_id = sc.domain_id "
                  "JOIN scan_job j ON j.scan_id = s.id "
                  "JOIN scan_result r ON r.scan_job_id = j.id "
              "GROUP BY 1, 2 "
                "HAVING BOOL_AND(r.scan_status = 'secure'::cdnskey_status AND "// CDNSKEY issued
                                "(NOW() - " << LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") << ")::TIMESTAMP <= r.scanned_at) OR "
                       "BOOL_AND(r.scan_status = 'secure-empty'::cdnskey_status AND "// no CDNSKEY issued
                                "(NOW() - " << LibPg::parameter<MaxScanPeriod>().as_type("INTERVAL") << ")::TIMESTAMP <= r.scanned_at)"
            "), "
            "secure_result_data AS "
            "("
                "SELECT scan_id, "
                       "row_number() OVER d = 1 AS is_last "
                  "FROM secure_result "
                "WINDOW d AS (PARTITION BY domain_id ORDER BY scan_finished_at DESC)"
            ") "
            "SELECT DISTINCT" <<
                    LibPg::item<DomainFqdn>() << "d.fqdn" <<
                    LibPg::item<DomainUid>() << "d.uid" <<
                    LibPg::item<CdnskeyFlags>() << "r.cdnskey_flags" <<
                    LibPg::item<CdnskeyProto>() << "r.cdnskey_proto" <<
                    LibPg::item<CdnskeyAlg>() << "r.cdnskey_alg" <<
                    LibPg::item<CdnskeyPublicKey>() << "r.cdnskey_public_key "
              "FROM insecure_acceptable ia "
              "JOIN scan s ON s.id = ia.scan_id "
              "JOIN domain d ON d.id = s.domain_id "
              "JOIN insecure_scan i ON i.scan_id = s.id "
              "JOIN scan_job j ON j.scan_id = s.id "
              "JOIN scan_result r ON r.scan_job_id = j.id AND "
                                    "r.nameserver_id = i.nameserver_id "
            "UNION "
            "SELECT DISTINCT "
                   "d.fqdn, "
                   "d.uid, "
                   "r.cdnskey_flags, "
                   "r.cdnskey_proto, "
                   "r.cdnskey_alg, "
                   "r.cdnskey_public_key "
              "FROM secure_result_data srd "
              "JOIN scan s ON s.id = srd.scan_id "
              "JOIN domain d ON d.id = s.domain_id "
              "JOIN scan_job j ON j.scan_id = s.id "
              "JOIN scan_result r ON r.scan_job_id = j.id "
             "WHERE srd.is_last AND "
                   "r.cdnskey_public_key IS NOT NULL "// last scan found no CDNSKEY issued
          "ORDER BY 1, 2, 3, 4, 5, 6",
            min_period_fixed_cdnskey,
            max_scan_period);
    if (!dbres.empty())
    {
        result.reserve(*dbres.size());
        std::for_each(begin(dbres), end(dbres), [&result](auto&& row)
        {
            result.push_back(
                    DomainCdnskey
                    {
                        row.template get<DomainFqdn>(),
                        row.template get<DomainUid>(),
                        row.template get<CdnskeyFlags>(),
                        row.template get<CdnskeyProto>(),
                        row.template get<CdnskeyAlg>(),
                        row.template get<CdnskeyPublicKey>()
                    });
        });
    }
    return result;
}

template <typename I, typename R>
typename I::Optional get_optional(const R& row)
{
    if (row.template is_null<I>())
    {
        return I::nullopt;
    }
    return row.template get_nullable<I>();
}

Report::RawScanResultsReply CdnskeyProcessor::get_raw_scan_results(
        const LibPg::PgTransaction& tx,
        const DomainFqdn& domain)
{
    const auto dbres = exec(tx, LibPg::make_query(
            "SELECT") << LibPg::nullable_item<Nameserver>() << "n.fqdn"
                      << LibPg::nullable_item<WorkerName>() << "w.name"
                      << LibPg::nullable_item<IpAddress>() << "r.nameserver_ip"
                      << LibPg::nullable_item<Timestamp>() << "r.scanned_at"
                      << LibPg::item<CdnskeyStatus>() << "CASE WHEN r.scan_status IS NOT NULL THEN r.scan_status "
                                                              "WHEN i.scan_id IS NULL THEN 'unknown'::cdnskey_status "
                                                                                     "ELSE 'unresolved'::cdnskey_status "
                                                         "END"
                      << LibPg::nullable_item<CdnskeyFlags>() << "r.cdnskey_flags"
                      << LibPg::nullable_item<CdnskeyProto>() << "r.cdnskey_proto"
                      << LibPg::nullable_item<CdnskeyAlg>() << "r.cdnskey_alg"
                      << LibPg::nullable_item<CdnskeyPublicKey>() << "r.cdnskey_public_key "
              "FROM domain d "
              "JOIN scan s ON s.domain_id = d.id "
         "LEFT JOIN ("
                        "insecure_scan i "
                   "JOIN nameserver n ON n.id = i.nameserver_id"
                   ") ON i.scan_id = s.id "
              "JOIN scan_job j ON j.scan_id = s.id "
         "LEFT JOIN ("
                        "scan_result r "
                   "JOIN worker w ON w.id = r.worker_id"
                   ") ON r.scan_job_id = j.id AND "
                        "(r.nameserver_id = n.id OR "
                         "n.id IS NULL) "
             "WHERE d.id = (SELECT MAX(id) "
                                "FROM domain "
                            "WHERE fqdn = LOWER(" << LibPg::parameter<DomainFqdn>().as_text() << ")) "
          "ORDER BY r.id, s.id",
            domain);
    Report::RawScanResultsReply reply;
    reply.reserve(*dbres.size());
    std::transform(begin(dbres), end(dbres), std::back_inserter(reply), [](auto&& item)
    {
        return Report::RawScanResult{
            get_optional<WorkerName>(item),
            get_optional<Timestamp>(item),
            get_optional<Nameserver>(item),
            get_optional<IpAddress>(item),
            {
                item.template get<CdnskeyStatus>(),
                get_optional<CdnskeyFlags>(item),
                get_optional<CdnskeyProto>(item),
                get_optional<CdnskeyAlg>(item),
                get_optional<CdnskeyPublicKey>(item)
            }};
    });
    return reply;
}

Report::RawScanResultsReply CdnskeyProcessor::get_raw_scan_results(
        const LibPg::PgTransaction& tx,
        const DomainUid& domain)
{
    try
    {
        const auto dbres = exec(tx, LibPg::make_query(
                "SELECT") << LibPg::nullable_item<Nameserver>() << "n.fqdn"
                          << LibPg::nullable_item<WorkerName>() << "w.name"
                          << LibPg::nullable_item<IpAddress>() << "r.nameserver_ip"
                          << LibPg::nullable_item<Timestamp>() << "r.scanned_at"
                          << LibPg::item<CdnskeyStatus>() << "CASE WHEN r.scan_status IS NOT NULL THEN r.scan_status "
                                                                  "WHEN i.scan_id IS NULL THEN 'unknown'::cdnskey_status "
                                                                                         "ELSE 'unresolved'::cdnskey_status "
                                                             "END"
                          << LibPg::nullable_item<CdnskeyFlags>() << "r.cdnskey_flags"
                          << LibPg::nullable_item<CdnskeyProto>() << "r.cdnskey_proto"
                          << LibPg::nullable_item<CdnskeyAlg>() << "r.cdnskey_alg"
                          << LibPg::nullable_item<CdnskeyPublicKey>() << "r.cdnskey_public_key "
                  "FROM domain d "
                  "JOIN scan s ON s.domain_id = d.id "
             "LEFT JOIN ("
                            "insecure_scan i "
                       "JOIN nameserver n ON n.id = i.nameserver_id"
                       ") ON i.scan_id = s.id "
                  "JOIN scan_job j ON j.scan_id = s.id "
             "LEFT JOIN ("
                            "scan_result r "
                       "JOIN worker w ON w.id = r.worker_id"
                       ") ON r.scan_job_id = j.id AND "
                            "(r.nameserver_id = n.id OR "
                             "n.id IS NULL) "
                 "WHERE d.uid::UUID = " << LibPg::parameter<DomainUid>().as_type("UUID") << " "
              "ORDER BY r.id, s.id",
                domain);
        Report::RawScanResultsReply reply;
        reply.reserve(*dbres.size());
        std::transform(begin(dbres), end(dbres), std::back_inserter(reply), [](auto&& item)
        {
            return Report::RawScanResult{
                get_optional<WorkerName>(item),
                get_optional<Timestamp>(item),
                get_optional<Nameserver>(item),
                get_optional<IpAddress>(item),
                {
                    item.template get<CdnskeyStatus>(),
                    get_optional<CdnskeyFlags>(item),
                    get_optional<CdnskeyProto>(item),
                    get_optional<CdnskeyAlg>(item),
                    get_optional<CdnskeyPublicKey>(item)
                }};
        });
        return reply;
    }
    catch (const LibPg::ExecFailure& e)
    {
        constexpr auto invalid_text_representation = LibPg::SqlState::data_exception_class.subclass<'P', '0', '2'>();
        if (e.error_code() == invalid_text_representation)
        {
            struct InvalidUuid : InvalidArgument
            {
                const char* what() const noexcept override { return "not an uuid value"; }
            };
            throw InvalidUuid{};
        }
        throw;
    }
}

void CdnskeyProcessor::clean_old_records(
        const LibPg::PgRwTransaction& tx,
        MaxScanPeriod max_scan_period,
        MinPeriodFixedCdnskey min_period_fixed_cdnskey)
{
    clean_scan_queue(tx, max_scan_period, min_period_fixed_cdnskey);
    clean_unused(tx);
}

void CdnskeyProcessor::vacuum(const LibPg::PgConnection& conn)
{
    exec(conn, "VACUUM ANALYZE domain, "
                              "insecure_scan, "
                              "nameserver, "
                              "scan, "
                              "scan_batch, "
                              "scan_batch_progress, "
                              "scan_job, "
                              "scan_queue, "
                              "scan_result, "
                              "worker");
}
