/*
 * Copyright (C) 2023-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/unwrapper/unwrapper.hh"

#include <exception>

namespace CdnskeyProcessor {
namespace Unwrapper {

DomainUid::DomainUid(const Api::DomainUid& data)
    : data_{data}
{ }

CdnskeyProcessor::DomainUid DomainUid::operator*() const
{
    return CdnskeyProcessor::DomainUid{data_.value()};
}

DomainFqdn::DomainFqdn(const Api::DomainFqdn& data)
    : data_{data}
{ }

CdnskeyProcessor::DomainFqdn DomainFqdn::operator*() const
{
    return CdnskeyProcessor::DomainFqdn{data_.value()};
}

Nameserver::Nameserver(const Api::Nameserver& data)
    : data_{data}
{ }

CdnskeyProcessor::Nameserver Nameserver::operator*() const
{
    return CdnskeyProcessor::Nameserver{data_.value()};
}

WorkerName::WorkerName(const Api::WorkerName& data)
    : data_{data}
{ }

CdnskeyProcessor::WorkerName WorkerName::operator*() const
{
    return CdnskeyProcessor::WorkerName{data_.value()};
}

ScanBatchId::ScanBatchId(const Api::Internal::ScanBatchId& data)
    : data_{data}
{ }

CdnskeyProcessor::ScanBatchId ScanBatchId::operator*() const
{
    return CdnskeyProcessor::ScanBatchId{data_.value()};
}

ScanJobId::ScanJobId(const Api::Internal::ScanJobId& data)
    : data_{data}
{ }

CdnskeyProcessor::ScanJobId ScanJobId::operator*() const
{
    return CdnskeyProcessor::ScanJobId{data_.value()};
}

Domain::Domain(const Api::Domain& data)
    : data_{data}
{ }

bool Domain::has_uid() const
{
    return data_.has_uid();
}

CdnskeyProcessor::DomainUid Domain::uid() const
{
    return *make_unwrapper(data_.uid());
}

bool Domain::has_fqdn() const
{
    return data_.has_fqdn();
}

CdnskeyProcessor::DomainFqdn Domain::fqdn() const
{
    return *make_unwrapper(data_.fqdn());
}

ScanType Domain::scan_type() const
{
    switch (data_.scan_type())
    {
        case Api::ScanType::INSECURE: return ScanType::insecure;
        case Api::ScanType::SECURE: return ScanType::secure;
        default: break;
    }
    struct UnknownScanType : std::exception
    {
        const char* what() const noexcept override { return "unknown scan type"; }
    };
    throw UnknownScanType{};
}

ScanJob::ScanJob(const Api::Internal::ScanJob& data)
    : data_{data}
{ }

CdnskeyProcessor::ScanJob ScanJob::operator*() const
{
    const auto domain_unwrapper = make_unwrapper(data_.domain());
    return CdnskeyProcessor::ScanJob{
            *make_unwrapper(data_.id()),
            domain_unwrapper.fqdn(),
            domain_unwrapper.uid()};
}

ScanTask::ScanTask(const Api::Supervisor::ScanTask& data)
    : data_{data}
{ }

bool ScanTask::has_domain() const
{
    return data_.has_domain();
}

Domain ScanTask::domain() const
{
    return make_unwrapper(data_.domain());
}

std::size_t ScanTask::nameservers_size() const
{
    return data_.nameservers_size();
}

ScanTaskCollectionRequest::ScanTaskCollectionRequest(const Api::Supervisor::ScanTaskCollectionRequest& data)
    : data_{data}
{ }

AddWorkerRequest::AddWorkerRequest(const Api::Supervisor::AddWorkerRequest& data)
    : data_{data}
{ }

bool AddWorkerRequest::has_worker_name() const
{
    return data_.has_worker_name();
}

CdnskeyProcessor::WorkerName AddWorkerRequest::worker_name() const
{
    return *make_unwrapper(data_.worker_name());
}

RemoveWorkerRequest::RemoveWorkerRequest(const Api::Supervisor::RemoveWorkerRequest& data)
    : data_{data}
{ }

bool RemoveWorkerRequest::has_worker_name() const
{
    return data_.has_worker_name();
}

CdnskeyProcessor::WorkerName RemoveWorkerRequest::worker_name() const
{
    return *make_unwrapper(data_.worker_name());
}

InsecureJobs::InsecureJobs(const Api::Internal::InsecureScanTaskCollection& data)
    : data_{data}
{ }

CdnskeyProcessor::InsecureJobs InsecureJobs::operator*() const
{
    CdnskeyProcessor::InsecureJobs result;
    std::for_each(data_.items().begin(), data_.items().end(), [&result](auto&& scan_task)
    {
        auto& jobs = result[*make_unwrapper(scan_task.nameserver())];
        std::for_each(scan_task.scan_jobs().begin(), scan_task.scan_jobs().end(), [&jobs](auto&& scan_job)
        {
            jobs.push_back(*make_unwrapper(scan_job));
        });
    });
    return result;
}

SecureJobs::SecureJobs(const Api::Internal::SecureScanTaskCollection& data)
    : data_{data}
{ }

CdnskeyProcessor::SecureJobs SecureJobs::operator*() const
{
    CdnskeyProcessor::SecureJobs result;
    std::for_each(data_.items().begin(), data_.items().end(), [&result](auto&& task)
    {
        result.push_back(*make_unwrapper(task.scan_job()));
    });
    return result;
}

InsecureScanBatch::InsecureScanBatch(const Api::Internal::ScanBatchInsecure& data)
    : data_{data}
{ }

bool InsecureScanBatch::has_id() const
{
    return data_.has_id();
}

CdnskeyProcessor::ScanBatchId InsecureScanBatch::id() const
{
    return *make_unwrapper(data_.id());
}

bool InsecureScanBatch::has_scan_tasks() const
{
    return data_.has_scan_tasks();
}

CdnskeyProcessor::InsecureJobs InsecureScanBatch::scan_tasks() const
{
    return *make_unwrapper(data_.scan_tasks());
}

SecureScanBatch::SecureScanBatch(const Api::Internal::ScanBatchSecure& data)
    : data_{data}
{ }

bool SecureScanBatch::has_id() const
{
    return data_.has_id();
}

CdnskeyProcessor::ScanBatchId SecureScanBatch::id() const
{
    return *make_unwrapper(data_.id());
}

bool SecureScanBatch::has_scan_tasks() const
{
    return data_.has_scan_tasks();
}

CdnskeyProcessor::SecureJobs SecureScanBatch::scan_tasks() const
{
    return *make_unwrapper(data_.scan_tasks());
}

DnskeyFlags::DnskeyFlags(const Api::DnskeyFlags& data)
    : data_{data}
{ }

CdnskeyProcessor::CdnskeyFlags DnskeyFlags::operator*() const
{
    return CdnskeyProcessor::CdnskeyFlags{data_.value()};
}

DnskeyProto::DnskeyProto(const Api::DnskeyProto& data)
    : data_{data}
{ }

CdnskeyProcessor::CdnskeyProto DnskeyProto::operator*() const
{
    return CdnskeyProcessor::CdnskeyProto{data_.value()};
}

DnskeyAlg::DnskeyAlg(const Api::DnskeyAlg& data)
    : data_{data}
{ }

CdnskeyProcessor::CdnskeyAlg DnskeyAlg::operator*() const
{
    return CdnskeyProcessor::CdnskeyAlg{data_.value()};
}

DnskeyPublicKey::DnskeyPublicKey(const Api::DnskeyPublicKey& data)
    : data_{data}
{ }

CdnskeyProcessor::CdnskeyPublicKey DnskeyPublicKey::operator*() const
{
    return CdnskeyProcessor::CdnskeyPublicKey{data_.value()};
}

IpAddress::IpAddress(const Api::IpAddress& data)
    : data_{data}
{ }

CdnskeyProcessor::IpAddress IpAddress::operator*() const
{
    return CdnskeyProcessor::IpAddress{boost::asio::ip::make_address(data_.value())};
}

Timestamp::Timestamp(const google::protobuf::Timestamp& data)
    : data_{data}
{ }

CdnskeyProcessor::Timestamp Timestamp::operator*() const
{
    const auto seconds = std::chrono::seconds{data_.seconds()};
    const auto nanos = std::chrono::nanoseconds{data_.nanos()};
    const auto from_epoch = std::chrono::duration_cast<std::chrono::nanoseconds>(seconds) + nanos;
    return CdnskeyProcessor::Timestamp{std::chrono::system_clock::time_point{from_epoch}};
}

Cdnskey::Cdnskey(const Api::Cdnskey& data)
    : data_{data}
{ }

CdnskeyProcessor::CdnskeyStatus Cdnskey::status() const
{
    switch (data_.status())
    {
        case Api::CdnskeyStatus::INSECURE_EMPTY:
            return CdnskeyStatus::insecure_empty;
        case Api::CdnskeyStatus::INSECURE_KEY:
            return CdnskeyStatus::insecure;
        case Api::CdnskeyStatus::SECURE_EMPTY:
            return CdnskeyStatus::secure_empty;
        case Api::CdnskeyStatus::SECURE_KEY:
            return CdnskeyStatus::secure;
        case Api::CdnskeyStatus::UNTRUSTWORTHY:
            return CdnskeyStatus::untrustworthy;
        case Api::CdnskeyStatus::UNKNOWN:
            return CdnskeyStatus::unknown;
        case Api::CdnskeyStatus::UNRESOLVED:
            return CdnskeyStatus::unresolved;
        case Api::CdnskeyStatus::UNRESOLVED_IP:
            return CdnskeyStatus::unresolved_ip;
        default:
            throw std::runtime_error{"unknown cdnskey status"};
    }
}

CdnskeyProcessor::CdnskeyFlags Cdnskey::flags() const
{
    return *make_unwrapper(data_.flags());
}

bool Cdnskey::has_flags() const
{
    return data_.has_flags();
}

CdnskeyProcessor::CdnskeyProto Cdnskey::proto() const
{
    return *make_unwrapper(data_.proto());
}

bool Cdnskey::has_proto() const
{
    return data_.has_proto();
}

CdnskeyProcessor::CdnskeyAlg Cdnskey::alg() const
{
    return *make_unwrapper(data_.alg());
}

bool Cdnskey::has_alg() const
{
    return data_.has_alg();
}

CdnskeyProcessor::CdnskeyPublicKey Cdnskey::public_key() const
{
    return *make_unwrapper(data_.public_key());
}

bool Cdnskey::has_public_key() const
{
    return data_.has_public_key();
}

ScanResult::ScanResult(const Api::Internal::ScanResult& data)
    : data_{data}
{ }

CdnskeyProcessor::ScanJobId ScanResult::scan_job_id() const
{
    return *make_unwrapper(data_.scan_job_id());
}

bool ScanResult::has_scan_job_id() const
{
    return data_.has_scan_job_id();
}

CdnskeyProcessor::Timestamp ScanResult::scan_at() const
{
    return *make_unwrapper(data_.scan_at());
}

bool ScanResult::has_scan_at() const
{
    return data_.has_scan_at();
}

CdnskeyProcessor::IpAddress ScanResult::nameserver_ip() const
{
    return *make_unwrapper(data_.nameserver_ip());
}

bool ScanResult::has_nameserver_ip() const
{
    return data_.has_nameserver_ip();
}

CdnskeyProcessor::Cdnskey ScanResult::cdnskey() const
{
    const auto unwrapper = make_unwrapper(data_.cdnskey());
    return CdnskeyProcessor::Cdnskey{
            unwrapper.status(),
            unwrapper.has_flags() ? CdnskeyProcessor::CdnskeyFlags::Optional(unwrapper.flags())
                                  : CdnskeyProcessor::CdnskeyFlags::nullopt,
            unwrapper.has_proto() ? CdnskeyProcessor::CdnskeyProto::Optional(unwrapper.proto())
                                  : CdnskeyProcessor::CdnskeyProto::nullopt,
            unwrapper.has_alg() ? CdnskeyProcessor::CdnskeyAlg::Optional(unwrapper.alg())
                                : CdnskeyProcessor::CdnskeyAlg::nullopt,
            unwrapper.has_public_key() ? CdnskeyProcessor::CdnskeyPublicKey::Optional(unwrapper.public_key())
                                       : CdnskeyProcessor::CdnskeyPublicKey::nullopt
    };
}

bool ScanResult::has_cdnskey() const
{
    return data_.has_cdnskey();
}

CdnskeyProcessor::Nameserver ScanResult::nameserver() const
{
    return *make_unwrapper(data_.nameserver());
}

bool ScanResult::has_nameserver() const
{
    return data_.has_nameserver();
}

ScanResultCollection::ScanResultCollection(const Api::Internal::ScanResultCollection& data)
    : data_{data}
{ }

CdnskeyProcessor::ScanBatchId ScanResultCollection::scan_batch_id() const
{
    return *make_unwrapper(data_.scan_batch_id());
}

bool ScanResultCollection::has_scan_batch_id() const
{
    return data_.has_scan_batch_id();
}

CdnskeyProcessor::WorkerName ScanResultCollection::worker_name() const
{
    return *make_unwrapper(data_.worker_name());
}

bool ScanResultCollection::has_worker_name() const
{
    return data_.has_worker_name();
}

CdnskeyProcessor::Timestamp ScanResultCollection::sent_at() const
{
    return *make_unwrapper(data_.sent_at());
}

bool ScanResultCollection::has_sent_at() const
{
    return data_.has_sent_at();
}

CdnskeyProcessor::Timestamp ScanResultCollection::started_at() const
{
    return *make_unwrapper(data_.start_at());
}

bool ScanResultCollection::has_started_at() const
{
    return data_.has_start_at();
}

CdnskeyProcessor::Timestamp ScanResultCollection::finished_at() const
{
    return *make_unwrapper(data_.finished_at());
}

bool ScanResultCollection::has_finished_at() const
{
    return data_.has_finished_at();
}

std::size_t ScanResultCollection::items_size() const
{
    return data_.items_size();
}

RawScanResultsRequest::RawScanResultsRequest(const Api::Report::RawScanResultsRequest& data)
    : data_{data}
{ }

CdnskeyProcessor::DomainUid RawScanResultsRequest::domain_uid() const
{
    return *make_unwrapper(data_.domain_uid());
}

bool RawScanResultsRequest::has_domain_uid() const
{
    return data_.has_domain_uid();
}

CdnskeyProcessor::DomainFqdn RawScanResultsRequest::domain_fqdn() const
{
    return *make_unwrapper(data_.domain_fqdn());
}

bool RawScanResultsRequest::has_domain_fqdn() const
{
    return data_.has_domain_fqdn();
}

}//namespace CdnskeyProcessor::Unwrapper
}//namespace CdnskeyProcessor

using namespace CdnskeyProcessor::Unwrapper;

DomainUid CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::DomainUid& data)
{
    return DomainUid{data};
}

DomainFqdn CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::DomainFqdn& data)
{
    return DomainFqdn{data};
}

Nameserver CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Nameserver& data)
{
    return Nameserver{data};
}

WorkerName CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::WorkerName& data)
{
    return WorkerName{data};
}

ScanBatchId CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Internal::ScanBatchId& data)
{
    return ScanBatchId{data};
}

ScanJobId CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Internal::ScanJobId& data)
{
    return ScanJobId{data};
}

ScanJob CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Internal::ScanJob& data)
{
    return ScanJob{data};
}

Domain CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Domain& data)
{
    return Domain{data};
}

ScanTask CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Supervisor::ScanTask& data)
{
    return ScanTask{data};
}

ScanTaskCollectionRequest CdnskeyProcessor::Unwrapper::make_unwrapper(
        const Api::Supervisor::ScanTaskCollectionRequest& data)
{
    return ScanTaskCollectionRequest{data};
}

AddWorkerRequest CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Supervisor::AddWorkerRequest& data)
{
    return AddWorkerRequest{data};
}

RemoveWorkerRequest CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Supervisor::RemoveWorkerRequest& data)
{
    return RemoveWorkerRequest{data};
}

InsecureJobs CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Internal::InsecureScanTaskCollection& data)
{
    return InsecureJobs{data};
}

SecureJobs CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Internal::SecureScanTaskCollection& data)
{
    return SecureJobs{data};
}

InsecureScanBatch CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Internal::ScanBatchInsecure& data)
{
    return InsecureScanBatch{data};
}

SecureScanBatch CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Internal::ScanBatchSecure& data)
{
    return SecureScanBatch{data};
}

DnskeyFlags CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::DnskeyFlags& data)
{
    return DnskeyFlags{data};
}

DnskeyProto CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::DnskeyProto& data)
{
    return DnskeyProto{data};
}

DnskeyAlg CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::DnskeyAlg& data)
{
    return DnskeyAlg{data};
}

DnskeyPublicKey CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::DnskeyPublicKey& data)
{
    return DnskeyPublicKey{data};
}

IpAddress CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::IpAddress& data)
{
    return IpAddress{data};
}

Timestamp CdnskeyProcessor::Unwrapper::make_unwrapper(const google::protobuf::Timestamp& data)
{
    return Timestamp{data};
}

Cdnskey CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Cdnskey& data)
{
    return Cdnskey{data};
}

ScanResult CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Internal::ScanResult& data)
{
    return ScanResult{data};
}

ScanResultCollection CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Internal::ScanResultCollection& data)
{
    return ScanResultCollection{data};
}

RawScanResultsRequest CdnskeyProcessor::Unwrapper::make_unwrapper(const Api::Report::RawScanResultsRequest& data)
{
    return RawScanResultsRequest{data};
}
