/*
 * Copyright (C) 2020-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/cfg/worker.hh"
#include "src/internal/amqp/bind_queue.hh"
#include "src/internal/amqp/boost_asio_handler.hh"
#include "src/internal/amqp/consume.hh"
#include "src/internal/amqp/exchange.hh"
#include "src/internal/amqp/publish.hh"
#include "src/internal/amqp/queue.hh"
#include "src/internal/amqp/reliable_tcp_channel.hh"
#include "src/internal/amqp/tcp_connection.hh"
#include "src/internal/scan_result.hh"
#include "src/internal/subprocess.hh"
#include "src/internal/timer.hh"
#include "src/scan_task.hh"

#include <liblog/liblog.hh>

#include <boost/asio/signal_set.hpp>
#include <boost/variant.hpp>

#include <amqpcpp.h>
#include <amqpcpp/libboostasio.h>

#include <unistd.h>

#include <algorithm>
#include <atomic>
#include <array>
#include <csignal>
#include <cstdlib>
#include <functional>
#include <iostream>
#include <list>
#include <utility>
#include <vector>

#define NOTHROW_LOG_TRACE(msg, ...) do { try { LIBLOG_TRACE(msg, ## __VA_ARGS__); } catch (...) {} } while (false)
#define NOTHROW_LOG_DEBUG(msg, ...) do { try { LIBLOG_DEBUG(msg, ## __VA_ARGS__); } catch (...) {} } while (false)
#define NOTHROW_LOG_INFO(msg, ...) do { try { LIBLOG_INFO(msg, ## __VA_ARGS__); } catch (...) {} } while (false)
#define NOTHROW_LOG_WARNING(msg, ...) do { try { LIBLOG_WARNING(msg, ## __VA_ARGS__); } catch (...) {} } while (false)
#define NOTHROW_LOG_ERROR(msg, ...) do { try { LIBLOG_ERROR(msg, ## __VA_ARGS__); } catch (...) {} } while (false)
#define NOTHROW_LOG_CRITICAL(msg, ...) do { try { LIBLOG_CRITICAL(msg, ## __VA_ARGS__); } catch (...) {} } while (false)

namespace CdnskeyProcessor {

namespace {

class Worker
{
public:
    explicit Worker(
            Internal::Timer&,
            const Cfg::Worker&);
    ~Worker();
    void start();
    void stop_requested();
private:
    void on_insecure_scan_queue_declare(std::uint32_t, std::uint32_t);
    void on_insecure_scan_queue_bind();
    void on_insecure_scan_queue_consume(const AMQP::Message& message, std::uint64_t delivery_tag, bool redelivered);
    void on_secure_scan_queue_consume(const AMQP::Message& message, std::uint64_t delivery_tag, bool redelivered);
    void send(const ScanResultCollection&);
    void manage_scan(Internal::Timer::TimePoint);
    struct InsecureScanData
    {
        InsecureScanBatch batch;
        std::uint64_t delivery_tag;
        std::map<DomainFqdn, ScanJobId> domain_to_job_id;
        std::map<Nameserver, std::vector<ScanJobId>> nameserver_to_job_ids;
        std::unique_ptr<Subprocess> scanner;
        std::size_t domains;
        std::string buffer;
        ScanResultCollection result_collection;
    };
    struct SecureScanData
    {
        SecureScanBatch batch;
        std::uint64_t delivery_tag;
        std::map<DomainFqdn, ScanJobId> domain_to_job_id;
        std::unique_ptr<Subprocess> scanner;
        std::size_t domains;
        std::string buffer;
        ScanResultCollection result_collection;
    };
    using ScanData = boost::variant<InsecureScanData, SecureScanData>;
    class ScanManager;
    std::list<ScanData> to_scan_;
    Internal::Timer& timer_;
    WorkerName name_;
    std::string scanner_;
    std::vector<std::string> scanner_args_;
    bool insecure_scan_queue_declared_ = false;
    bool secure_scan_queue_declared_ = false;
    std::unique_ptr<Internal::Amqp::Queue> insecure_scan_queue_;
    std::unique_ptr<Internal::Amqp::Queue> secure_scan_queue_;
    std::unique_ptr<AMQP::TcpConnection> connection_;
    std::unique_ptr<AMQP::TcpChannel> channel_;
//    Internal::Amqp::FanoutExchange exchange_diag_;
//    using Diag = Internal::Amqp::RoutingKey<struct Diag_>;
    using Results = Internal::Amqp::RoutingKey<struct Results_>;
    Internal::Amqp::DirectExchange</*Diag,*/ Results> exchange_response_;
    Internal::Amqp::FanoutExchange exchange_insecure_fanout_;
//    Internal::Amqp::DirectExchange<> exchange_insecure_direct_;
};

volatile std::sig_atomic_t occuring_signal = -1;

}//namespace CdnskeyProcessor::{anonymous}

}//namespace CdnskeyProcessor

int main(int argc, char* argv[])
{
    try
    {
        const auto cfg = [argc, argv]()
        {
            try
            {
                return CdnskeyProcessor::Cfg::get_cfg(argc, argv, CdnskeyProcessor::Cfg::set_up_logging);
            }
            catch (const CdnskeyProcessor::Cfg::Worker::Help&)
            {
                std::exit(EXIT_SUCCESS);
            }
            catch (const std::exception& e)
            {
                std::cerr << "std::exception caught: " << e.what() << std::endl;
            }
            catch (...)
            {
                std::cerr << "unknown exception caught" << std::endl;
            }
            std::exit(EXIT_FAILURE);
        }();
        LIBLOG_SET_CONTEXT(WorkerLogContext, log_ctx, cfg.name);
        CdnskeyProcessor::Internal::Timer timer{1};
        boost::asio::signal_set signals{timer.service(), SIGINT, SIGHUP, SIGQUIT};
        signals.add(SIGTERM);
        CdnskeyProcessor::Worker worker{timer, cfg};
        signals.async_wait([&worker](const boost::system::error_code& error, int signal_number)
        {
            if (!error)
            {
                CdnskeyProcessor::occuring_signal = signal_number;
                worker.stop_requested();
            }
        });
        timer.loop();
        return EXIT_SUCCESS;
    }
    catch (const std::exception& e)
    {
        LIBLOG_ERROR("std::exception caught: {}", e.what());
        return EXIT_FAILURE;
    }
    catch (...)
    {
        LIBLOG_ERROR("unknown exception caught");
        return EXIT_FAILURE;
    }
}

namespace CdnskeyProcessor {

namespace {

std::string get_signal()
{
    if (occuring_signal == SIGINT)
    {
        return "SIGINT";
    }
    if (occuring_signal == SIGHUP)
    {
        return "SIGHUP";
    }
    if (occuring_signal == SIGQUIT)
    {
        return "SIGQUIT";
    }
    if (occuring_signal == SIGTERM)
    {
        return "SIGTERM";
    }
    return "unknown " + std::to_string(occuring_signal);
}

AMQP::Address make_address(const Cfg::Worker::Amqp::Broker& cfg)
{
    return AMQP::Address{cfg.hostname, cfg.port, {cfg.username, cfg.password}, cfg.vhost};
}

template <typename Rep, typename Period>
void set_expiration(AMQP::Envelope& envelope, std::chrono::duration<Rep, Period> ttl)
{
    const auto ttl_ms = std::chrono::duration_cast<std::chrono::milliseconds>(ttl);
    envelope.setExpiration(std::to_string(ttl_ms.count()));
}

std::size_t send_to_scanner(const InsecureScanBatch& batch, std::function<void(const std::string&)> writer)
{
    std::size_t domains_counter = 0;
    bool marker_written = false;
    std::for_each(begin(batch.jobs), end(batch.jobs), [&](auto&& ns_jobs_pair)
    {
        if (ns_jobs_pair.second.empty())
        {
            return;
        }
        // [insecure]
        // nameserver domain domain ...
        // ...
        auto line = *ns_jobs_pair.first;
        std::for_each(begin(ns_jobs_pair.second), end(ns_jobs_pair.second), [&](auto&& job)
        {
            static constexpr char delimiter[] = " ";
            line.append(delimiter + *job.domain_fqdn);
        });
        domains_counter += ns_jobs_pair.second.size();
        if (!marker_written)
        {
            static constexpr char marker[] = "[insecure]\n";
            writer(marker);
            marker_written = true;
        }
        writer(line + "\n");
    });
    return domains_counter;
}

std::size_t send_to_scanner(const SecureScanBatch& batch, std::function<void(const std::string&)> writer)
{
    if (batch.jobs.empty())
    {
        return 0;
    }
    bool marker_written = false;
    // [secure]
    // domain domain ...
    std::for_each(begin(batch.jobs), end(batch.jobs), [&](auto&& job)
    {
        if (!marker_written)
        {
            static constexpr char marker[] = "[secure]\n";
            writer(marker + *job.domain_fqdn);
            marker_written = true;
        }
        else
        {
            static constexpr char delimiter[] = " ";
            writer(delimiter + *job.domain_fqdn);
        }
    });
    writer("\n");
    return batch.jobs.size();
}

constexpr char tag_unresolved_ip[] = "unresolved-ip";
constexpr char tag_insecure[] = "insecure";
constexpr char tag_insecure_empty[] = "insecure-empty";
constexpr char tag_unresolved[] = "unresolved";
constexpr char tag_secure[] = "secure";
constexpr char tag_secure_empty[] = "secure-empty";
constexpr char tag_untrustworthy[] = "untrustworthy";
constexpr char tag_unknown[] = "unknown";

struct ResultLine
{
    explicit ResultLine(
            const std::string& line,
            std::string::const_iterator type_end)
        : line{line},
          type_end{type_end}
    { }
    const std::string& line;
    std::string::const_iterator type_end;
};

template <std::size_t rhs_size>
bool operator==(const ResultLine& lhs, const char (&rhs)[rhs_size])
{
    static_assert(0 < rhs_size, "rhs_size must be a positive number");
    return lhs.line.compare(0, lhs.type_end - begin(lhs.line), rhs, rhs_size - 1) == 0;
}

template <std::size_t tokens>
std::array<std::string, tokens> split_into_tokens(const ResultLine& in, char delimiter = ' ')
{
    std::array<std::string, tokens> result;
    const auto end_ptr = end(in.line);
    if (in.type_end == end_ptr)
    {
        if (tokens == 0)
        {
            return result;
        }
        throw std::runtime_error{"no token found, " + std::to_string(tokens) + " expected"};
    }
    auto token_begin = in.type_end + 1;
    for (std::size_t idx = 0; idx < tokens; ++idx)
    {
        const auto token_end = std::find(token_begin, end_ptr, delimiter);
        result[idx] = std::string{token_begin, token_end};
        if (token_end == end_ptr)
        {
            if ((idx + 1) == tokens)
            {
                return result;
            }
            throw std::runtime_error{std::to_string(idx + 1) + " token(s) found, " + std::to_string(tokens) + " expected"};
        }
        token_begin = token_end + 1;
    }
    throw std::runtime_error{"more then " + std::to_string(tokens) + " expected tokens found"};
}

ScanResult parse_type_insecure(
        const ResultLine& result,
        const std::map<DomainFqdn, ScanJobId>& domain_to_job_id)
{
    auto tokens = split_into_tokens<7>(result);
    return ScanResult
    {
        domain_to_job_id.at(DomainFqdn{std::move(tokens[2])}),
        Timestamp{Timestamp::ValueType::clock::now()},
        IpAddress{boost::asio::ip::make_address(tokens[1])},
        Cdnskey
        {
            CdnskeyStatus::insecure,
            CdnskeyFlags{boost::lexical_cast<uint32_t>(tokens[3])},
            CdnskeyProto{boost::lexical_cast<uint32_t>(tokens[4])},
            CdnskeyAlg{boost::lexical_cast<uint32_t>(tokens[5])},
            CdnskeyPublicKey{std::move(tokens[6])}
        },
        Nameserver{std::move(tokens[0])}
    };
}

ScanResult parse_type_insecure_empty(
        const ResultLine& result,
        const std::map<DomainFqdn, ScanJobId>& domain_to_job_id)
{
    auto tokens = split_into_tokens<3>(result);
    return ScanResult
    {
        domain_to_job_id.at(DomainFqdn{std::move(tokens[2])}),
        Timestamp{Timestamp::ValueType::clock::now()},
        IpAddress{boost::asio::ip::make_address(tokens[1])},
        Cdnskey{CdnskeyStatus::insecure_empty},
        Nameserver{std::move(tokens[0])}
    };
}

ScanResult parse_type_unresolved(
        const ResultLine& result,
        const std::map<DomainFqdn, ScanJobId>& domain_to_job_id)
{
    auto tokens = split_into_tokens<3>(result);
    return ScanResult
    {
        domain_to_job_id.at(DomainFqdn{std::move(tokens[2])}),
        Timestamp{Timestamp::ValueType::clock::now()},
        IpAddress{boost::asio::ip::make_address(tokens[1])},
        Cdnskey{CdnskeyStatus::unresolved},
        Nameserver{std::move(tokens[0])}
    };
}

ScanResult parse_type_secure(
        const ResultLine& result,
        const std::map<DomainFqdn, ScanJobId>& domain_to_job_id)
{
    auto tokens = split_into_tokens<5>(result);
    return ScanResult
    {
        domain_to_job_id.at(DomainFqdn{std::move(tokens[0])}),
        Timestamp{Timestamp::ValueType::clock::now()},
        IpAddress::nullopt,
        Cdnskey
        {
            CdnskeyStatus::secure,
            CdnskeyFlags{boost::lexical_cast<uint32_t>(tokens[1])},
            CdnskeyProto{boost::lexical_cast<uint32_t>(tokens[2])},
            CdnskeyAlg{boost::lexical_cast<uint32_t>(tokens[3])},
            CdnskeyPublicKey{std::move(tokens[4])}
        },
        Nameserver::nullopt
    };
}

ScanResult parse_type_untrustworthy(
        const ResultLine& result,
        const std::map<DomainFqdn, ScanJobId>& domain_to_job_id)
{
    auto tokens = split_into_tokens<1>(result);
    return ScanResult
    {
        domain_to_job_id.at(DomainFqdn{std::move(tokens[0])}),
        Timestamp{Timestamp::ValueType::clock::now()},
        IpAddress::nullopt,
        Cdnskey{CdnskeyStatus::untrustworthy},
        Nameserver::nullopt
    };
}

ScanResult parse_type_unknown(
        const ResultLine& result,
        const std::map<DomainFqdn, ScanJobId>& domain_to_job_id)
{
    auto tokens = split_into_tokens<1>(result);
    return ScanResult
    {
        domain_to_job_id.at(DomainFqdn{std::move(tokens[0])}),
        Timestamp{Timestamp::ValueType::clock::now()},
        IpAddress::nullopt,
        Cdnskey{CdnskeyStatus::unknown},
        Nameserver::nullopt
    };
}

ScanResult parse_type_secure_empty(
        const ResultLine& result,
        const std::map<DomainFqdn, ScanJobId>& domain_to_job_id)
{
    auto tokens = split_into_tokens<1>(result);
    return ScanResult
    {
        domain_to_job_id.at(DomainFqdn{std::move(tokens[0])}),
        Timestamp{Timestamp::ValueType::clock::now()},
        IpAddress::nullopt,
        Cdnskey{CdnskeyStatus::secure_empty},
        Nameserver::nullopt
    };
}

std::vector<ScanResult> parse_type_unresolved_ip(
        const ResultLine& result,
        const std::map<Nameserver, std::vector<ScanJobId>>& nameserver_to_job_ids)
{
    auto tokens = split_into_tokens<1>(result);
    const auto nameserver = Nameserver{std::move(tokens[0])};
    const auto& jobs = nameserver_to_job_ids.at(nameserver);
    const auto make_result = [now = Timestamp{Timestamp::ValueType::clock::now()}, &nameserver](ScanJobId job_id)
    {
        return ScanResult
        {
            job_id,
            now,
            IpAddress::nullopt,
            Cdnskey{CdnskeyStatus::unresolved_ip},
            nameserver
        };
    };
    std::vector<ScanResult> affected_results;
    affected_results.reserve(jobs.size());
    std::for_each(begin(jobs), end(jobs), [&](auto&& job_id)
    {
        affected_results.push_back(make_result(job_id));
    });
    return affected_results;
}

std::vector<ScanResult> parse_scanner_output(
        const std::string& line,
        const std::map<DomainFqdn, ScanJobId>& domain_to_job_id,
        const std::map<Nameserver, std::vector<ScanJobId>>& nameserver_to_job_ids)
{
    const ResultLine result_line{line, std::find(begin(line), end(line), ' ')};
    try
    {
        if (result_line == tag_insecure_empty)
        {
            return {parse_type_insecure_empty(result_line, domain_to_job_id)};
        }
        if (result_line == tag_secure_empty)
        {
            return {parse_type_secure_empty(result_line, domain_to_job_id)};
        }
        if (result_line == tag_unresolved)
        {
            return {parse_type_unresolved(result_line, domain_to_job_id)};
        }
        if (result_line == tag_unknown)
        {
            return {parse_type_unknown(result_line, domain_to_job_id)};
        }
        if (result_line == tag_insecure)
        {
            return {parse_type_insecure(result_line, domain_to_job_id)};
        }
        if (result_line == tag_secure)
        {
            return {parse_type_secure(result_line, domain_to_job_id)};
        }
        if (result_line == tag_unresolved_ip)
        {
            return parse_type_unresolved_ip(result_line, nameserver_to_job_ids);
        }
        if (result_line == tag_untrustworthy)
        {
            return {parse_type_untrustworthy(result_line, domain_to_job_id)};
        }
    }
    catch (const std::exception& e)
    {
        throw std::runtime_error{std::string("subparser failed(") + e.what() + ")"};
    }
    throw std::runtime_error{"unrecognized/unimplemented scan result type '" + std::string{begin(result_line.line), result_line.type_end} + "'"};
}

void append(std::vector<ScanResult>& dst, const std::vector<ScanResult>& data)
{
    dst.insert(end(dst), begin(data), end(data));
}

std::vector<ScanResult> parse(
        std::string& buffer,
        const std::map<DomainFqdn, ScanJobId>& domain_to_job_id,
        const std::map<Nameserver, std::vector<ScanJobId>>& nameserver_to_job_ids)
{
    static constexpr char delimiter = '\n';
    std::vector<ScanResult> results;
    auto data_begin_ptr = begin(buffer);
    auto data_end_ptr = std::find(data_begin_ptr, end(buffer), delimiter);
    while (data_end_ptr != end(buffer))
    {
        const auto data = std::string{data_begin_ptr, data_end_ptr};
        try
        {
            LIBLOG_TRACE("{}", data);
            append(results, parse_scanner_output(data, domain_to_job_id, nameserver_to_job_ids));
        }
        catch (const std::exception& e)
        {
            LIBLOG_ERROR("{}: {}", e.what(), data);
        }
        data_begin_ptr = data_end_ptr + 1;
        data_end_ptr = std::find(data_begin_ptr, end(buffer), delimiter);
    }
    buffer.erase(begin(buffer), data_begin_ptr);
    return results;
}

std::map<DomainFqdn, ScanJobId> make_domain_to_job_id_table(const InsecureScanBatch& batch)
{
    std::map<DomainFqdn, ScanJobId> result;
    std::for_each(begin(batch.jobs), end(batch.jobs), [&result](auto&& ns_jobs_pair)
    {
        std::for_each(begin(ns_jobs_pair.second), end(ns_jobs_pair.second), [&result](auto&& job)
        {
            result[job.domain_fqdn] = job.scan_job_id;
        });
    });
    return result;
}

std::map<DomainFqdn, ScanJobId> make_domain_to_job_id_table(const SecureScanBatch& batch)
{
    std::map<DomainFqdn, ScanJobId> result;
    std::for_each(begin(batch.jobs), end(batch.jobs), [&result](auto&& job)
    {
        result[job.domain_fqdn] = job.scan_job_id;
    });
    return result;
}

std::map<Nameserver, std::vector<ScanJobId>> make_nameserver_to_job_ids_table(const InsecureScanBatch& batch)
{
    std::map<Nameserver, std::vector<ScanJobId>> result;
    std::for_each(begin(batch.jobs), end(batch.jobs), [&result](auto&& ns_jobs_pair)
    {
        auto& ids = result[ns_jobs_pair.first];
        ids.reserve(ns_jobs_pair.second.size());
        std::for_each(begin(ns_jobs_pair.second), end(ns_jobs_pair.second), [&ids](auto&& job)
        {
            ids.push_back(job.scan_job_id);
        });
    });
    return result;
}

std::chrono::seconds compute_scanner_runtime(const InsecureScanBatch& batch)
{
    const auto queries = *estimated_size(batch);
    static constexpr auto min_runtime = std::chrono::seconds{30};
    static const auto queries_per_second = 50;
    const auto runtime = std::chrono::seconds{queries / queries_per_second};
    return std::max(min_runtime, runtime);
}

std::chrono::seconds compute_scanner_runtime(const SecureScanBatch& batch)
{
    const auto queries = *estimated_size(batch);
    static constexpr auto min_runtime = std::chrono::seconds{30};
    static const auto queries_per_second = 50;
    const auto runtime = std::chrono::seconds{queries / queries_per_second};
    return std::max(min_runtime, runtime);
}

std::vector<const char*> make_scanner_args(
        const std::string* scanner,
        const std::vector<std::string>* args,
        const std::string* runtime)
{
    std::vector<const char*> result;
    // cmd args... runtime NULL
    result.reserve(1 + args->size() + 1 + 1);
    result.push_back(scanner->c_str());
    std::transform(begin(*args), end(*args), std::back_inserter(result), [](auto&& value) { return value.c_str(); });
    result.push_back(runtime->c_str());
    result.push_back(nullptr);
    return result;
}

Worker::Worker(Internal::Timer& timer, const Cfg::Worker& cfg)
    : timer_{timer},
      name_{cfg.name},
      scanner_{cfg.scanner},
      scanner_args_{cfg.scanner_args},
      insecure_scan_queue_{cfg.amqp.queue.insecure_prefix.empty()
            ? nullptr
            : std::make_unique<Internal::Amqp::Queue>(cfg.amqp.queue.insecure_prefix + cfg.name)},
      secure_scan_queue_{cfg.amqp.queue.secure.empty()
            ? nullptr
            : std::make_unique<Internal::Amqp::Queue>(cfg.amqp.queue.secure)},
      connection_{Internal::Amqp::make_tcp_connection(
            Internal::Amqp::make_boost_asio_handler(
                    timer.service(),
                    [this](AMQP::TcpConnection*)
                    {
                        NOTHROW_LOG_INFO("{} established connection to the RabbitMQ", *name_);
                        this->start();
                    },
                    [this](AMQP::TcpConnection*, const char* message)
                    {
                        NOTHROW_LOG_ERROR("{} connection to the RabbitMQ failed: {}", *name_, message);
                    },
                    [this](AMQP::TcpConnection*)
                    {
                        NOTHROW_LOG_INFO("{} lost connection to the RabbitMQ", *name_);
                    }),
            make_address(cfg.amqp.broker))},
      channel_{std::make_unique<AMQP::TcpChannel>(connection_.get())},
//      exchange_diag_{Internal::Amqp::ExchangeName{amqp_conf.exchange.diag.get()}},
      exchange_response_{
            Internal::Amqp::ExchangeName{cfg.amqp.exchange.response},
//            Diag{amqp_conf.queue.diag.get()},
            Results{cfg.amqp.queue.results}},
      exchange_insecure_fanout_{Internal::Amqp::ExchangeName{cfg.amqp.exchange.insecure_fanout}}
//      exchange_insecure_direct_{Internal::Amqp::ExchangeName{cfg.amqp.exchange.insecure_direct}},
{
    if ((insecure_scan_queue_ == nullptr) &&
        (secure_scan_queue_ == nullptr))
    {
        throw std::runtime_error{*name_ + " misconfigured; at least one of the scan queues must be present"};
    }
    channel_->setQos(cfg.amqp.prefetch_count)
        .onError([](const char* msg)
        {
            NOTHROW_LOG_INFO("error on channel: {}", msg);
        })
        .onFinalize([]()
        {
            NOTHROW_LOG_INFO("channel finalized");
        })
        .onSuccess([]()
        {
            NOTHROW_LOG_INFO("channel success");
        });
#if 0
    channel_->declareQueue(
            ping_request_queue_,
            AMQP::durable,
            AMQP::Table{}.set("x-queue-type", "quorum"))
        .onSuccess([this](const std::string&, uint32_t messagecount, uint32_t consumercount)
        {
            try
            {
                this->on_ping_request_queue_declare(messagecount, consumercount);
            }
            catch(...) { }
        })
        .onError([this](const char* message)
        {
            try
            {
                LIBLOG_ERROR("declare ping request queue failure: {}", message);
            }
            catch(...) { }
        });
#endif
    if (insecure_scan_queue_ != nullptr)
    {
        channel_->declareQueue(
                *insecure_scan_queue_,
                AMQP::durable,
                AMQP::Table{}.set("x-queue-type", "quorum"))
            .onSuccess([this](const std::string& name, uint32_t messagecount, uint32_t consumercount)
            {
                NOTHROW_LOG_INFO("{} queue declared, messagecount: {}, consumercount: {}", name, messagecount, consumercount);
                try
                {
                    this->on_insecure_scan_queue_declare(messagecount, consumercount);
                }
                catch(...) { }
            })
            .onError([](const char* msg)
            {
                NOTHROW_LOG_ERROR("insecure scan queue declaration failure: {}", msg);
            })
            .onFinalize([]()
            {
                NOTHROW_LOG_INFO("insecure scan queue declaration finalized");
            });
    }
    if (secure_scan_queue_ != nullptr)
    {
        Internal::Amqp::consume(*channel_, *secure_scan_queue_, std::string{}, AMQP::exclusive, AMQP::Table{})
            .onReceived([this](const AMQP::Message& message, std::uint64_t delivery_tag, bool redelivered)
            {
                this->on_secure_scan_queue_consume(message, delivery_tag, redelivered);
            })
            .onError([](const char* message)
            {
                NOTHROW_LOG_ERROR("consume secure scan queue failure: {}", message);
            })
            .onFinalize([]()
            {
                NOTHROW_LOG_INFO("consume secure scan queue: finalized");
            });
    }
    channel_->onError([](const char* msg)
        {
            NOTHROW_LOG_INFO("error on channel: {}", msg);
        });
    manage_scan(Internal::Timer::TimePoint::clock::now());
    LIBLOG_INFO("run worker \"{}\"", *name_);
}

Worker::~Worker()
{
    NOTHROW_LOG_INFO("worker done by signal {}", get_signal());
}

void Worker::start()
{
}

void Worker::stop_requested()
{
    timer_.service().stop();
}

void Worker::on_insecure_scan_queue_declare(std::uint32_t messagecount, std::uint32_t consumercount)
{
    insecure_scan_queue_declared_ = true;
    bind_queue(*channel_, *insecure_scan_queue_, exchange_insecure_fanout_, AMQP::Table{})
        .onSuccess([&]()
        {
            NOTHROW_LOG_INFO("insecure scan queue bound (messagecount: {}, consumercount: {})", messagecount, consumercount);
            try
            {
                this->on_insecure_scan_queue_bind();
            }
            catch (...) { }
        })
        .onError([](const char* message)
        {
            NOTHROW_LOG_ERROR("bind insecure scan queue failure: {}", message);
        })
        .onFinalize([]()
        {
            NOTHROW_LOG_INFO("bind insecure scan queue: finalized");
        });
#if 0
    bind_queue(*channel_, insecure_scan_queue_, exchange_insecure_direct_.recipient(name_), AMQP::Table{})
        .onSuccess([this]()
        {
            try
            {
                this->on_insecure_scan_queue_bind();
            }
            catch (...) { }
        })
        .onError([](const char* message)
        {
            NOTHROW_LOG_ERROR("bind insecure scan queue failure: {}", message);
        });
#endif
}

void Worker::on_insecure_scan_queue_bind()
{
    consume(*channel_, *insecure_scan_queue_, std::string{}, AMQP::exclusive, AMQP::Table{})
        .onReceived([this](const AMQP::Message& message, std::uint64_t delivery_tag, bool redelivered)
        {
            this->on_insecure_scan_queue_consume(message, delivery_tag, redelivered);
        })
        .onError([](const char* msg)
        {
            NOTHROW_LOG_ERROR("consume insecure scan queue failure: {}", msg);
        })
        .onSuccess([]()
        {
            NOTHROW_LOG_INFO("consume insecure scan queue: success");
        });
}

void Worker::on_insecure_scan_queue_consume(const AMQP::Message& message, std::uint64_t delivery_tag, bool redelivered)
{
    try
    {
        auto batch = deserialize<InsecureScanBatch>(message.body(), message.bodySize());
        const auto batch_id = batch.batch_id;
        if (redelivered)
        {
            LIBLOG_INFO("insecure scan batch {} redelivered", *batch_id);
        }
        else
        {
            LIBLOG_INFO("insecure scan batch {} delivered", *batch_id);
        }
        to_scan_.emplace_back(InsecureScanData{
            std::move(batch),
            delivery_tag,
            {},
            {},
            nullptr,
            0,
            {},
            {
                batch_id,
                name_,
                Timestamp{Timestamp::ValueType::clock::now()},
                Timestamp::nullopt,
                Timestamp::nullopt,
                {}
            }
        });
    }
    catch (const std::exception& e)
    {
        NOTHROW_LOG_INFO("insecure scan batch handling failure: {}", e.what());
    }
    catch (...)
    {
        NOTHROW_LOG_INFO("insecure scan batch handling failed");
    }
}

void Worker::on_secure_scan_queue_consume(const AMQP::Message& message, std::uint64_t delivery_tag, bool redelivered)
{
    try
    {
        auto batch = deserialize<SecureScanBatch>(message.body(), message.bodySize());
        const auto batch_id = batch.batch_id;
        if (redelivered)
        {
            LIBLOG_INFO("secure scan batch {} redelivered", *batch_id);
        }
        else
        {
            LIBLOG_INFO("secure scan batch {} delivered", *batch_id);
        }
        to_scan_.emplace_back(SecureScanData{
            std::move(batch),
            delivery_tag,
            {},
            nullptr,
            0,
            {},
            {
                batch_id,
                name_,
                Timestamp{Timestamp::ValueType::clock::now()},
                Timestamp::nullopt,
                Timestamp::nullopt,
                {}
            }
        });
    }
    catch (const std::exception& e)
    {
        NOTHROW_LOG_INFO("secure scan batch handling failure: {}", e.what());
    }
    catch (...)
    {
        NOTHROW_LOG_INFO("secure scan batch handling failed");
    }
}

void Worker::send(const ScanResultCollection& result)
{
    const auto raw_data = serialize(result);
    const AMQP::Envelope envelope{raw_data.c_str(), raw_data.length()};
    const bool published = publish(
            *channel_,
            exchange_response_.recipient<Results>(),
            envelope,
            0);
    if (published)
    {
        LIBLOG_DEBUG("{} published scan result {} of {} bytes into exchange_response.queue_results",
                     *name_,
                     *result.scan_batch_id,
                     raw_data.size());
    }
    else
    {
        LIBLOG_INFO("{} did not publish scan result {} into exchange_response.queue_results",
                    *name_,
                    *result.scan_batch_id);
    }
}

class Worker::ScanManager : boost::static_visitor<void>
{
public:
    explicit ScanManager(Worker& worker)
        : worker_{worker}
    { }
    void operator()(InsecureScanData& data) const
    {
        const auto scanner_is_running = data.scanner != nullptr;
        if (!scanner_is_running)
        {
            data.domain_to_job_id = make_domain_to_job_id_table(data.batch);
            data.nameserver_to_job_ids = make_nameserver_to_job_ids_table(data.batch);
            const auto runtime = std::to_string(compute_scanner_runtime(data.batch).count());
            LIBLOG_DEBUG("start insecure scan with runtime limited to {} seconds", runtime);
            data.scanner = std::make_unique<Subprocess>(make_scanner_args(
                    &worker_.scanner_,
                    &worker_.scanner_args_,
                    &runtime));
            data.result_collection.started_at = Timestamp{Timestamp::ValueType::clock::now()};
            data.domains = send_to_scanner(data.batch, [&data](const std::string& bytes)
            {
                data.scanner->write(bytes);
            });
            data.scanner->write_eof();
            LIBLOG_DEBUG("{} insecure domains to scan", data.domains);
            return;
        }
        Subprocess::ReadBuffer chunk;
        const auto bytes = data.scanner->read(chunk);
        if (bytes < 0)
        {
            struct ReadFailure : std::exception
            {
                const char* what() const noexcept override { return "scanner output read failed"; }
            };
            throw ReadFailure{};
        }
        if (bytes == 0)
        {
            data.result_collection.finished_at = Timestamp{Timestamp::ValueType::clock::now()};
            worker_.send(data.result_collection);
            data.scanner->wait();
            LIBLOG_DEBUG("{} insecure domains scanned", data.domains);
            const bool acknowledged = worker_.channel_->ack(data.delivery_tag);
            if (!acknowledged)
            {
                LIBLOG_INFO("receiving of an insecure scan batch not acknowledged");
            }
            worker_.to_scan_.pop_front();
            return;
        }
        data.buffer.append(chunk.data(), bytes);
        append(data.result_collection.scan_results, parse(
                data.buffer,
                data.domain_to_job_id,
                data.nameserver_to_job_ids));
        static constexpr auto scan_results_treshold = 1024;
        if (scan_results_treshold <= data.result_collection.scan_results.size())
        {
            worker_.send(data.result_collection);
            data.result_collection.scan_results.clear();
        }
    }
    void operator()(SecureScanData& data) const
    {
        const auto scanner_is_running = data.scanner != nullptr;
        if (!scanner_is_running)
        {
            data.domain_to_job_id = make_domain_to_job_id_table(data.batch);
            const auto runtime = std::to_string(compute_scanner_runtime(data.batch).count());
            LIBLOG_DEBUG("start secure scan with runtime limited to {} seconds", runtime);
            data.scanner = std::make_unique<Subprocess>(make_scanner_args(
                    &worker_.scanner_,
                    &worker_.scanner_args_,
                    &runtime));
            data.result_collection.started_at = Timestamp{Timestamp::ValueType::clock::now()};
            data.domains = send_to_scanner(data.batch, [&](const std::string& bytes)
            {
                data.scanner->write(bytes);
            });
            data.scanner->write_eof();
            LIBLOG_DEBUG("{} secure domains to scan", data.domains);
            return;
        }
        Subprocess::ReadBuffer chunk;
        const auto bytes = data.scanner->read(chunk);
        if (bytes < 0)
        {
            struct ReadFailure : std::exception
            {
                const char* what() const noexcept override { return "scanner output read failed"; }
            };
            throw ReadFailure{};
        }
        if (bytes == 0)
        {
            data.result_collection.finished_at = Timestamp{Timestamp::ValueType::clock::now()};
            worker_.send(data.result_collection);
            data.scanner->wait();
            LIBLOG_DEBUG("{} secure domains scanned", data.domains);
            const bool acknowledged = worker_.channel_->ack(data.delivery_tag);
            if (!acknowledged)
            {
                LIBLOG_INFO("receiving of an secure scan batch not acknowledged");
            }
            worker_.to_scan_.pop_front();
            return;
        }
        data.buffer.append(chunk.data(), bytes);
        append(data.result_collection.scan_results, parse(
                data.buffer,
                data.domain_to_job_id,
                {}));
        static constexpr auto scan_results_treshold = 1024;
        if (scan_results_treshold <= data.result_collection.scan_results.size())
        {
            worker_.send(data.result_collection);
            data.result_collection.scan_results.clear();
        }
    }
private:
    Worker& worker_;
};

void Worker::manage_scan(Internal::Timer::TimePoint)
{
    if (!to_scan_.empty())
    {
        try
        {
            boost::apply_visitor(ScanManager{*this}, to_scan_.front());
        }
        catch (const SubprocessException& e)
        {
            NOTHROW_LOG_INFO("scanner failure: {}", e.what());
        }
        catch (const SubprocessChildException& e)
        {
            NOTHROW_LOG_INFO("scanner run failure: {}", e.what());
            _exit(EXIT_FAILURE);
        }
        catch (const std::exception& e)
        {
            NOTHROW_LOG_INFO("scan batch handling failure: {}", e.what());
        }
        catch (...)
        {
            NOTHROW_LOG_INFO("scan batch handling failed");
        }
    }
    static constexpr auto delay = std::chrono::milliseconds{100};
    timer_.register_callback(Internal::Timer::TimePoint::clock::now() + delay, [this](Internal::Timer::TimePoint t)
    {
        manage_scan(t);
    });
}

}//namespace CdnskeyProcessor::{anonymous}

}//namespace CdnskeyProcessor
