/*
 * Copyright (C) 2023-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/wrapper/wrapper.hh"

#include <exception>
#include <utility>

namespace CdnskeyProcessor {
namespace Wrapper {

ScanBatchId::ScanBatchId(Api::Internal::ScanBatchId& data)
    : data_{data}
{
    data_.Clear();
}

void ScanBatchId::set(CdnskeyProcessor::ScanBatchId id)
{
    data_.set_value(*id);
}

ScanJobId::ScanJobId(Api::Internal::ScanJobId& data)
    : data_{data}
{
    data_.Clear();
}

void ScanJobId::set(CdnskeyProcessor::ScanJobId id)
{
    data_.set_value(*id);
}

DomainFqdn::DomainFqdn(Api::DomainFqdn& data)
    : data_{data}
{
    data_.Clear();
}

void DomainFqdn::set(CdnskeyProcessor::DomainFqdn fqdn)
{
    data_.set_value(*std::move(fqdn));
}

DomainUid::DomainUid(Api::DomainUid& data)
    : data_{data}
{
    data_.Clear();
}

void DomainUid::set(CdnskeyProcessor::DomainUid uid)
{
    data_.set_value(*std::move(uid));
}

Domain::Domain(Api::Domain& data)
    : data_{data}
{
    data_.Clear();
}

void Domain::set_uid(CdnskeyProcessor::DomainUid uid)
{
    make_wrapper(*data_.mutable_uid()).set(std::move(uid));
}

void Domain::set_fqdn(CdnskeyProcessor::DomainFqdn fqdn)
{
    make_wrapper(*data_.mutable_fqdn()).set(std::move(fqdn));
}

void Domain::set_scan_type(CdnskeyProcessor::ScanType type)
{
    data_.set_scan_type([type]()
    {
        switch (type)
        {
            case ScanType::insecure: return Api::ScanType::INSECURE;
            case ScanType::secure: return Api::ScanType::SECURE;
        }
        struct OutOfRange : std::exception
        {
            const char* what() const noexcept override { return "unknown scan type"; }
        };
        throw OutOfRange{};
    }());
}

ScanJob::ScanJob(Api::Internal::ScanJob& data)
    : data_{data}
{
    data_.Clear();
}

void ScanJob::set_domain(
        CdnskeyProcessor::DomainUid uid,
        CdnskeyProcessor::DomainFqdn fqdn,
        CdnskeyProcessor::ScanType type)
{
    auto wrapper = make_wrapper(*data_.mutable_domain());
    wrapper.set_uid(std::move(uid));
    wrapper.set_fqdn(std::move(fqdn));
    wrapper.set_scan_type(type);
}

void ScanJob::set_id(CdnskeyProcessor::ScanJobId id)
{
    make_wrapper(*data_.mutable_id()).set(id);
}

Nameserver::Nameserver(Api::Nameserver& data)
    : data_{data}
{
    data_.Clear();
}

void Nameserver::set(CdnskeyProcessor::Nameserver nameserver)
{
    data_.set_value(*std::move(nameserver));
}

WorkerName::WorkerName(Api::WorkerName& data)
    : data_{data}
{
    data_.Clear();
}

void WorkerName::set(CdnskeyProcessor::WorkerName worker)
{
    data_.set_value(*std::move(worker));
}

InsecureScanTaskCollection::InsecureScanTaskCollection(Api::Internal::InsecureScanTaskCollection& data)
    : data_{data}
{
    data_.Clear();
}

void InsecureScanTaskCollection::set_items(Iterator begin, Iterator end)
{
    std::for_each(begin, end, [this](auto&& ns_jobs_pair)
    {
        auto* const dst_task = data_.add_items();
        make_wrapper(*(dst_task->mutable_nameserver())).set(ns_jobs_pair.first);
        std::for_each(ns_jobs_pair.second.begin(), ns_jobs_pair.second.end(), [dst_task](auto&& job)
        {
            auto job_wrapper = make_wrapper(*(dst_task->add_scan_jobs()));
            job_wrapper.set_domain(job.domain_uid, job.domain_fqdn, CdnskeyProcessor::ScanType::insecure);
            job_wrapper.set_id(job.scan_job_id);
        });
    });
}

SecureScanTaskCollection::SecureScanTaskCollection(Api::Internal::SecureScanTaskCollection& data)
    : data_{data}
{
    data_.Clear();
}

void SecureScanTaskCollection::set_items(Iterator begin, Iterator end)
{
    std::for_each(begin, end, [this](auto&& job)
    {
        auto job_wrapper = make_wrapper(*(data_.add_items()->mutable_scan_job()));
        job_wrapper.set_domain(job.domain_uid, job.domain_fqdn, CdnskeyProcessor::ScanType::secure);
        job_wrapper.set_id(job.scan_job_id);
    });
}

ScanBatchInsecure::ScanBatchInsecure(Api::Internal::ScanBatchInsecure& data)
    : data_{data}
{
    data_.Clear();
}

void ScanBatchInsecure::set(const CdnskeyProcessor::InsecureScanBatch& batch)
{
    make_wrapper(*data_.mutable_id()).set(batch.batch_id);
    make_wrapper(*data_.mutable_scan_tasks()).set_items(begin(batch.jobs), end(batch.jobs));
}

ScanBatchSecure::ScanBatchSecure(Api::Internal::ScanBatchSecure& data)
    : data_{data}
{
    data_.Clear();
}

void ScanBatchSecure::set(const CdnskeyProcessor::SecureScanBatch& batch)
{
    make_wrapper(*data_.mutable_id()).set(batch.batch_id);
    make_wrapper(*data_.mutable_scan_tasks()).set_items(begin(batch.jobs), end(batch.jobs));
}

Timestamp::Timestamp(google::protobuf::Timestamp& data)
    : data_{data}
{
    data_.Clear();
}

void Timestamp::set(const CdnskeyProcessor::Timestamp& timestamp)
{
    data_.set_seconds(std::chrono::duration_cast<std::chrono::seconds>(timestamp->time_since_epoch()).count());
    data_.set_nanos(std::chrono::duration_cast<std::chrono::nanoseconds>(timestamp->time_since_epoch()).count() %
                    1'000'000'000);
}

IpAddress::IpAddress(Api::IpAddress& data)
    : data_{data}
{
    data_.Clear();
}

void IpAddress::set(const CdnskeyProcessor::IpAddress& ip_address)
{
    data_.set_value(ip_address->to_string());
}

CdnskeyFlags::CdnskeyFlags(Api::DnskeyFlags& data)
    : data_{data}
{
    data_.Clear();
}

void CdnskeyFlags::set(const CdnskeyProcessor::CdnskeyFlags& flags)
{
    data_.set_value(*flags);
}

CdnskeyProto::CdnskeyProto(Api::DnskeyProto& data)
    : data_{data}
{
    data_.Clear();
}

void CdnskeyProto::set(const CdnskeyProcessor::CdnskeyProto& flags)
{
    data_.set_value(*flags);
}

CdnskeyAlg::CdnskeyAlg(Api::DnskeyAlg& data)
    : data_{data}
{
    data_.Clear();
}

void CdnskeyAlg::set(const CdnskeyProcessor::CdnskeyAlg& flags)
{
    data_.set_value(*flags);
}

CdnskeyPublicKey::CdnskeyPublicKey(Api::DnskeyPublicKey& data)
    : data_{data}
{
    data_.Clear();
}

void CdnskeyPublicKey::set(const CdnskeyProcessor::CdnskeyPublicKey& key)
{
    data_.set_value(*key);
}

Cdnskey::Cdnskey(Api::Cdnskey& data)
    : data_{data}
{
    data_.Clear();
}

void Cdnskey::set(const CdnskeyProcessor::Cdnskey& cdnskey)
{
    data_.set_status([&cdnskey]()
    {
        switch (cdnskey.status)
        {
            case CdnskeyStatus::insecure:
                return Api::CdnskeyStatus::INSECURE_KEY;
            case CdnskeyStatus::insecure_empty:
                return Api::CdnskeyStatus::INSECURE_EMPTY;
            case CdnskeyStatus::secure:
                return Api::CdnskeyStatus::SECURE_KEY;
            case CdnskeyStatus::secure_empty:
                return Api::CdnskeyStatus::SECURE_EMPTY;
            case CdnskeyStatus::untrustworthy:
                return Api::CdnskeyStatus::UNTRUSTWORTHY;
            case CdnskeyStatus::unknown:
                return Api::CdnskeyStatus::UNKNOWN;
            case CdnskeyStatus::unresolved:
                return Api::CdnskeyStatus::UNRESOLVED;
            case CdnskeyStatus::unresolved_ip:
                return Api::CdnskeyStatus::UNRESOLVED_IP;
        }
        throw std::runtime_error{"unknown cdnskey status"};
    }());
    if (cdnskey.flags != CdnskeyProcessor::CdnskeyFlags::nullopt)
    {
        make_wrapper(*data_.mutable_flags()).set(*cdnskey.flags);
    }
    if (cdnskey.proto != CdnskeyProcessor::CdnskeyProto::nullopt)
    {
        make_wrapper(*data_.mutable_proto()).set(*cdnskey.proto);
    }
    if (cdnskey.alg != CdnskeyProcessor::CdnskeyAlg::nullopt)
    {
        make_wrapper(*data_.mutable_alg()).set(*cdnskey.alg);
    }
    if (cdnskey.public_key != CdnskeyProcessor::CdnskeyPublicKey::nullopt)
    {
        make_wrapper(*data_.mutable_public_key()).set(*cdnskey.public_key);
    }
}

ScanResult::ScanResult(Api::Internal::ScanResult& data)
    : data_{data}
{
    data_.Clear();
}

void ScanResult::set(const CdnskeyProcessor::ScanResult& result)
{
    make_wrapper(*data_.mutable_scan_job_id()).set(result.scan_job_id);
    make_wrapper(*data_.mutable_scan_at()).set(result.scan_at);
    if (result.nameserver_ip != CdnskeyProcessor::IpAddress::nullopt)
    {
        make_wrapper(*data_.mutable_nameserver_ip()).set(*result.nameserver_ip);
    }
    make_wrapper(*data_.mutable_cdnskey()).set(result.cdnskey);
    if (result.nameserver != CdnskeyProcessor::Nameserver::nullopt)
    {
        make_wrapper(*data_.mutable_nameserver()).set(*result.nameserver);
    }
}

ScanResultCollection::ScanResultCollection(Api::Internal::ScanResultCollection& data)
    : data_{data}
{
    data_.Clear();
}

void ScanResultCollection::set(const CdnskeyProcessor::ScanResultCollection& collection)
{
    make_wrapper(*data_.mutable_scan_batch_id()).set(collection.scan_batch_id);
    make_wrapper(*data_.mutable_worker_name()).set(collection.worker_name);
    make_wrapper(*data_.mutable_sent_at()).set(collection.sent_at);
    if (collection.started_at != CdnskeyProcessor::Timestamp::nullopt)
    {
        make_wrapper(*data_.mutable_start_at()).set(*collection.started_at);
    }
    if (collection.finished_at != CdnskeyProcessor::Timestamp::nullopt)
    {
        make_wrapper(*data_.mutable_finished_at()).set(*collection.finished_at);
    }
    data_.mutable_items()->Reserve(collection.scan_results.size());
    std::for_each(begin(collection.scan_results), end(collection.scan_results), [items = data_.mutable_items()](auto&& result)
    {
        make_wrapper(*items->Add()).set(result);
    });
}

AcceptableDnskeysReply::AcceptableDnskeysReply(Api::Supervisor::AcceptableDnskeysReply& data)
    : data_{data}
{
    data_.Clear();
}

void AcceptableDnskeysReply::set_items(Iterator begin, Iterator end)
{
    data_.mutable_data()->clear_items();
    auto& dst = *data_.mutable_data();
    dst.mutable_items()->Reserve(end - begin);
    std::for_each(begin, end, [this, &dst](auto&& domain_cdnskey)
    {
        auto* const item = dst.add_items();
        make_wrapper(*(item->mutable_domain_uid())).set(domain_cdnskey.domain_uid);
        make_wrapper(*(item->mutable_domain_fqdn())).set(domain_cdnskey.domain_fqdn);
        auto& dst_dnskey = *(item->mutable_dnskey());
        make_wrapper(*dst_dnskey.mutable_flags()).set(domain_cdnskey.flags);
        make_wrapper(*dst_dnskey.mutable_proto()).set(domain_cdnskey.proto);
        make_wrapper(*dst_dnskey.mutable_alg()).set(domain_cdnskey.alg);
        make_wrapper(*dst_dnskey.mutable_public_key()).set(domain_cdnskey.public_key);
    });
}

RawScanResultsReply::RawScanResultsReply(Api::Report::RawScanResultsReply& data)
    : data_{data}
{
    data_.Clear();
}

void RawScanResultsReply::set_items(Iterator begin, Iterator end)
{
    data_.mutable_data()->clear_items();
    auto& dst = *data_.mutable_data();
    dst.mutable_items()->Reserve(end - begin);
    std::for_each(begin, end, [this, &dst](auto&& scan_result)
    {
        auto* const item = dst.add_items();
        if (scan_result.worker_name != CdnskeyProcessor::WorkerName::nullopt)
        {
            make_wrapper(*(item->mutable_worker_name())).set(*scan_result.worker_name);
        }
        if (scan_result.scan_at != CdnskeyProcessor::Timestamp::nullopt)
        {
            make_wrapper(*(item->mutable_scan_at())).set(*scan_result.scan_at);
        }
        if (scan_result.nameserver != CdnskeyProcessor::Nameserver::nullopt)
        {
            make_wrapper(*(item->mutable_nameserver())).set(*scan_result.nameserver);
        }
        if (scan_result.nameserver_ip != CdnskeyProcessor::IpAddress::nullopt)
        {
            make_wrapper(*(item->mutable_nameserver_ip())).set(*scan_result.nameserver_ip);
        }
        make_wrapper(*(item->mutable_cdnskey())).set(scan_result.cdnskey);
    });
}

}//namespace CdnskeyProcessor::Wrapper
}//namespace CdnskeyProcessor

using namespace CdnskeyProcessor::Wrapper;

ScanBatchId CdnskeyProcessor::Wrapper::make_wrapper(Api::Internal::ScanBatchId& data)
{
    return ScanBatchId{data};
}

ScanJobId CdnskeyProcessor::Wrapper::make_wrapper(Api::Internal::ScanJobId& data)
{
    return ScanJobId{data};
}

DomainFqdn CdnskeyProcessor::Wrapper::make_wrapper(Api::DomainFqdn& data)
{
    return DomainFqdn{data};
}

DomainUid CdnskeyProcessor::Wrapper::make_wrapper(Api::DomainUid& data)
{
    return DomainUid{data};
}

Domain CdnskeyProcessor::Wrapper::make_wrapper(Api::Domain& data)
{
    return Domain{data};
}

ScanJob CdnskeyProcessor::Wrapper::make_wrapper(Api::Internal::ScanJob& data)
{
    return ScanJob{data};
}

Nameserver CdnskeyProcessor::Wrapper::make_wrapper(Api::Nameserver& data)
{
    return Nameserver{data};
}

WorkerName CdnskeyProcessor::Wrapper::make_wrapper(Api::WorkerName& data)
{
    return WorkerName{data};
}

InsecureScanTaskCollection CdnskeyProcessor::Wrapper::make_wrapper(Api::Internal::InsecureScanTaskCollection& data)
{
    return InsecureScanTaskCollection{data};
}

SecureScanTaskCollection CdnskeyProcessor::Wrapper::make_wrapper(Api::Internal::SecureScanTaskCollection& data)
{
    return SecureScanTaskCollection{data};
}

ScanBatchInsecure CdnskeyProcessor::Wrapper::make_wrapper(Api::Internal::ScanBatchInsecure& data)
{
    return ScanBatchInsecure{data};
}

ScanBatchSecure CdnskeyProcessor::Wrapper::make_wrapper(Api::Internal::ScanBatchSecure& data)
{
    return ScanBatchSecure{data};
}

Timestamp CdnskeyProcessor::Wrapper::make_wrapper(google::protobuf::Timestamp& data)
{
    return Timestamp{data};
}

IpAddress CdnskeyProcessor::Wrapper::make_wrapper(Api::IpAddress& data)
{
    return IpAddress{data};
}

CdnskeyFlags CdnskeyProcessor::Wrapper::make_wrapper(Api::DnskeyFlags& data)
{
    return CdnskeyFlags{data};
}

CdnskeyProto CdnskeyProcessor::Wrapper::make_wrapper(Api::DnskeyProto& data)
{
    return CdnskeyProto{data};
}

CdnskeyAlg CdnskeyProcessor::Wrapper::make_wrapper(Api::DnskeyAlg& data)
{
    return CdnskeyAlg{data};
}

CdnskeyPublicKey CdnskeyProcessor::Wrapper::make_wrapper(Api::DnskeyPublicKey& data)
{
    return CdnskeyPublicKey{data};
}

Cdnskey CdnskeyProcessor::Wrapper::make_wrapper(Api::Cdnskey& data)
{
    return Cdnskey{data};
}

ScanResult CdnskeyProcessor::Wrapper::make_wrapper(Api::Internal::ScanResult& data)
{
    return ScanResult{data};
}

ScanResultCollection CdnskeyProcessor::Wrapper::make_wrapper(Api::Internal::ScanResultCollection& data)
{
    return ScanResultCollection{data};
}

AcceptableDnskeysReply CdnskeyProcessor::Wrapper::make_wrapper(Api::Supervisor::AcceptableDnskeysReply& data)
{
    return AcceptableDnskeysReply{data};
}

RawScanResultsReply CdnskeyProcessor::Wrapper::make_wrapper(Api::Report::RawScanResultsReply& data)
{
    return RawScanResultsReply{data};
}
