#!/usr/bin/env bash

check_db_healthy()
{
    docker exec -i "$CONTAINER_NAME" /usr/bin/env sh -c "pg_isready -qt 10 -U postgres -d postgres && \
    pg_isready -qt 10 -h 127.0.0.1 -U postgres -d postgres"
}

wait_for_db()
{
    local cntdn=50
    while [ 0 -lt $cntdn ] && ! check_db_healthy; do
        sleep .5
        ((--cntdn))
    done
}

stop_db()
{
    if docker container ls --filter=status=running --format "{{.Names}}"|grep -q "^${CONTAINER_NAME}$"; then
        echo -ne "Stopping container $CONTAINER_NAME ..."
        docker container stop "$CONTAINER_NAME" > /dev/null
        echo -ne "\r                                                           \r"
    fi
}

start_db()
{
    echo -ne "Starting Postgres $POSTGRES_VERSION server ..."
    docker run \
        -d --rm \
        -p ${DB_HOST}:${DB_PORT}:5432 \
        -e "POSTGRES_PASSWORD=$DB_ADMIN_PASSWORD" \
        --name "$CONTAINER_NAME" \
        postgres:$POSTGRES_VERSION > /dev/null
    wait_for_db
    echo -e "\rIn container $CONTAINER_NAME on $DB_HOST:$DB_PORT is listening Postgres $POSTGRES_VERSION server"
}

exec_sql()
{
    docker exec -i "$CONTAINER_NAME" psql -wXqbf - -o /dev/null "$@"
}

init_db()
{
    local root_dir=$(dirname "$0")
    cat "$root_dir/schema/0000_init_database.sql" |\
     sed "s,CREATE USER cdnskey_processor_view WITH PASSWORD '[^']*',CREATE USER cdnskey_processor_view WITH PASSWORD '$DB_VIEW_PASSWORD',g" |\
     sed "s,CREATE USER cdnskey_processor WITH PASSWORD '[^']*',CREATE USER cdnskey_processor WITH PASSWORD '$DB_USER_PASSWORD',g" |\
     exec_sql -U postgres -d postgres
    cat "$root_dir/schema/0001_schema.sql" | exec_sql -U cdnskey_processor -d cdnskey_processor
#    cat "$root_dir/schema/0002_delete_cascade_scan_queue.sql" | exec_sql -U cdnskey_processor -d cdnskey_processor
#    cat "$root_dir/schema/0003_enum_tables_to_enum_type.sql" | exec_sql -U cdnskey_processor -d cdnskey_processor
#    cat "$root_dir/schema/0005_worker_name_not_null_constraint.sql" | exec_sql -U cdnskey_processor -d cdnskey_processor
#    cat "$root_dir/schema/0006_scan_results_indexes.sql" | exec_sql -U cdnskey_processor -d cdnskey_processor
#    cat "$root_dir/schema/0007_get_raw_scan_results.sql" | exec_sql -U cdnskey_processor -d cdnskey_processor
#    cat "$root_dir/schema/0008_split_scan_batch_table.sql" | exec_sql -U cdnskey_processor -d cdnskey_processor
#    cat "$root_dir/schema/0009_scan_result_evaluation.sql" | exec_sql -U cdnskey_processor -d cdnskey_processor
}

set_default()
{
    local varname="$1"
    if [ -z "${!varname}" ]; then
        local default_value="$2"
        export $varname="$default_value"
    fi
}

print_usage()
{
    echo -e "Usage:\n\n" \
        $(basename "$0") "[--container-name <NAME>] [--db-host <IP>] [--db-port <PORT>] [--db-admin-password <ADMIN>] [--db-view-password <VIEW>] [--db-user-password <USER>] [--postgres-version <VERSION>]\n\n" \
        "\tStarts an empty akm database in container <NAME> listening on <IP>:<PORT> initialized by init scripts invocation and running on PostgreSQL server of specified <VERSION>.\n" \
        "\tPassword for database user postgres on database postgres is <ADMIN>, password for database user cdnskey_processor_view on database cdnskey_processor is <VIEW> and password for database user cdnskey_processor\n" \
        "\ton database cdnskey_processor is <USER>.\n\n" \
        "\tDefault values:\n" \
        "\t\tNAME - CONTAINER_NAME environment variable or 'fred-database'\n" \
        "\t\tIP - DB_HOST environment variable or '127.0.0.1'\n" \
        "\t\tPORT - DB_PORT environment variable or 25432\n" \
        "\t\tADMIN - DB_ADMIN_PASSWORD environment variable or 'eyeing4eeTha'\n" \
        "\t\tVIEW - DB_VIEW_PASSWORD environment variable or 'iWae2pho5noh'\n" \
        "\t\tUSER - DB_USER_PASSWORD environment variable or 'phupoo4Rei8E'\n" \
         "\t\tVERSION - POSTGRES_VERSION environment variable or 13\n\n" \
        "or\n\n" \
        $(basename "$0") "-h|--help - print this help"
}

parse_args()
{
    while [ 0 -lt $# ]; do
        case "$1" in
            -h|--help)
                print_usage
                exit 0
                ;;
            --container-name)
                CONTAINER_NAME="$2"
                shift 2
                ;;
            --db-host)
                DB_HOST="$2"
                shift 2
                ;;
            --db-port)
                DB_PORT="$2"
                shift 2
                ;;
            --db-admin-password)
                DB_ADMIN_PASSWORD="$2"
                shift 2
                ;;
            --db-view-password)
                DB_VIEW_PASSWORD="$2"
                shift 2
                ;;
            --db-user-password)
                DB_USER_PASSWORD="$2"
                shift 2
                ;;
            --postgres-version)
                POSTGRES_VERSION="$2"
                shift 2
                ;;
            *)
                echo -e "Invalid option $1.\n" >&2
                print_usage
                exit 1
                ;;
        esac
    done
    set_default CONTAINER_NAME "akm-database"
    set_default DB_HOST "127.0.0.1"
    set_default DB_PORT 25432
    set_default DB_ADMIN_PASSWORD "eyeing4eeTha"
    set_default DB_VIEW_PASSWORD "iWae2pho5noh"
    set_default DB_USER_PASSWORD "phupoo4Rei8E"
    set_default POSTGRES_VERSION "13"
}

main()
{
    parse_args "$@"
    stop_db
    start_db
    init_db
}

main "$@"