/*
 * Copyright (C) 2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include "db_loader/cfg/options.hh"

#include <boost/program_options.hpp>

#include <syslog.h>

#include <algorithm>
#include <initializer_list>
#include <iostream>
#include <fstream>
#include <functional>
#include <sstream>
#include <type_traits>
#include <utility>

#ifndef DEFAULT_CONFIG_FILE
#define DEFAULT_CONFIG_FILE "cdnskey-processor-db-loader.conf"
#endif

namespace CdnskeyProcessor {
namespace DbLoader {
namespace Cfg {

namespace {

template <typename Exception, typename MsgGen>
[[noreturn]] void raise(MsgGen msg_gen)
{
    struct Error : Exception
    {
        explicit Error(MsgGen msg_gen)
            : Exception{},
              msg_gen{std::move(msg_gen)}
        { }
        const char* what() const noexcept override { return msg_gen(); }
        MsgGen msg_gen;
    };
    throw Error{std::move(msg_gen)};
}

LogSeverity make_severity(const std::string& value)
{
    if (value == "trace")
    {
        return LogSeverity::trace;
    }
    if (value == "debug")
    {
        return LogSeverity::debug;
    }
    if (value == "info")
    {
        return LogSeverity::info;
    }
    if (value == "warning")
    {
        return LogSeverity::warning;
    }
    if (value == "error")
    {
        return LogSeverity::error;
    }
    if (value == "critical")
    {
        return LogSeverity::critical;
    }
    raise<Exception>([msg = "not a valid severity \"" + value + "\""]() { return msg.c_str(); });
}

void required_one_of(
        const boost::program_options::variables_map& variables_map,
        std::initializer_list<const char*> variables)
{
    if (std::all_of(
                begin(variables),
                end(variables),
                [&variables_map](auto&& variable) { return variables_map.count(variable) == 0; }))
    {
        std::string msg;
        std::for_each(begin(variables), end(variables), [&msg](auto&& variable)
        {
            if (msg.empty())
            {
                msg = variable;
            }
            else
            {
                msg += std::string{", "} + variable;
            }
        });
        raise<MissingOption>([msg = "missing one of options: '" + msg + "'"]() { return msg.c_str(); });
    }
}

void only_from(
        const boost::program_options::variables_map& variables_map,
        std::initializer_list<const char*> variables)
{
    std::for_each(begin(variables_map), end(variables_map), [&variables](auto&& variable_value_pair)
    {
        if (!variable_value_pair.second.defaulted() &&
            !std::any_of(begin(variables), end(variables), [&variable_value_pair](auto&& variable)
            {
                return variable_value_pair.first == variable;
            }))
        {
            raise<InappropriateOption>([msg = "inappropriate option '" + variable_value_pair.first + "'"]()
            {
                return msg.c_str();
            });
        }
    });
}

void conflicting_options(
        const boost::program_options::variables_map& variables_map, 
        const char* opt1,
        const char* opt2)
{
    if (variables_map.count(opt1) && !variables_map.at(opt1).defaulted() &&
        variables_map.count(opt2) && !variables_map.at(opt2).defaulted())
    {
        raise<ConflictingOptions>([msg = std::string{"conflicting options '"} + opt1 + "' and '" + opt2 + "'."]()
        {
            return msg.c_str();
        });
    }
}

constexpr int invalid_argc = -1;
constexpr const char* const invalid_argv[0] = { };

static constexpr unsigned terminal_width = 120;

template <typename Cmd>
struct HasCommand : boost::static_visitor<bool>
{
    template <typename T>
    constexpr bool operator()(T&&) const noexcept
    {
        return std::is_same<std::decay_t<T>, Cmd>::value;
    }
};

template <typename Cmd>
bool has_command(const Options::Command& command)
{
    return boost::apply_visitor(HasCommand<Cmd>{}, command);
}

bool has_no_command(const Options::Command& command)
{
    return has_command<Options::NoCommand>(command);
}

template <typename Cmd>
struct GetCommand : boost::static_visitor<Cmd&>
{
    Cmd& operator()(Cmd& value) const noexcept
    {
        return value;
    }
    template <typename T>
    [[noreturn]] Cmd& operator()(const T& value) const
    {
        throw value;
    }
};

template <typename Cmd>
Cmd& set_command(Options::Command& command)
{
    if (has_no_command(command))
    {
        command = Cmd{};
    }
    return boost::apply_visitor(GetCommand<Cmd>{}, command);
}

class CheckOptions : boost::static_visitor<void>
{
public:
    explicit CheckOptions(
            std::function<void()> check_help,
            std::function<void()> check_version,
            std::function<void()> check_import)
        : check_help_command_{check_help},
          check_version_command_{check_version},
          check_import_command_{check_import}
    { }
    void operator()(const Options::NoCommand&) const { }
    void operator()(const Options::Help&) const { check_help_command_(); }
    void operator()(const Options::Version&) const { check_version_command_(); }
    void operator()(const Options::Import&) const { check_import_command_(); }
private:
    std::function<void()> check_help_command_;
    std::function<void()> check_version_command_;
    std::function<void()> check_import_command_;
};

}//namespace CdnskeyProcessor::DbLoader::Cfg::{anonymous}

const Options& Options::get()
{
    return init(invalid_argc, invalid_argv);
}

const Options& Options::init(int argc, const char* const* argv)
{
    static const Options* singleton_ptr = nullptr;
    const bool init_requested = (argc != invalid_argc) || (argv != invalid_argv);
    const bool first_run = singleton_ptr == nullptr;
    if (first_run)
    {
        if (!init_requested)
        {
            throw std::runtime_error("First call of Cfg::Options::init must contain valid arguments");
        }
        static const Options singleton(argc, argv);
        singleton_ptr = &singleton;
    }
    else if (init_requested)
    {
        throw std::runtime_error("Only first call of Cfg::Options::init can contain valid arguments");
    }
    return *singleton_ptr;
}

Options::Options(int argc, const char* const* argv)
    : command{NoCommand{}}
{
    boost::program_options::options_description generic_options{"Generic options", terminal_width};
    std::string opt_config_file_name;
    generic_options.add_options()
        ("help,h", "produce help message")
        ("version,v", "display version information")
        ("config,c",
#ifdef DEFAULT_CONFIG_FILE
         boost::program_options::value<std::string>(&opt_config_file_name)->default_value(DEFAULT_CONFIG_FILE),
#else
         boost::program_options::value<std::string>(&opt_config_file_name),
#endif
         "name of a configuration file.");

    boost::program_options::options_description command_options{"Command options", terminal_width};
    command_options.add_options()
        (
            "command",
            boost::program_options::value<std::string>()->notifier([this](auto&& value)
            {
                try
                {
                    if (value == "help")
                    {
                        set_command<Help>(command);
                        return;
                    }
                    if (value == "version")
                    {
                        set_command<Version>(command);
                        return;
                    }
                    if (value == "import")
                    {
                        set_command<Import>(command);
                        return;
                    }
                }
                catch (const Help&)
                {
                    raise<InvalidCommand>([msg = "invalid option for \"" + value + "\" command"]() { return msg.c_str(); });
                }
                catch (const Version&)
                {
                    raise<InvalidCommand>([msg = "invalid option for \"" + value + "\" command"]() { return msg.c_str(); });
                }
                catch (const Import&)
                {
                    raise<InvalidCommand>([msg = "invalid option for \"" + value + "\" command"]() { return msg.c_str(); });
                }
                raise<InvalidCommand>([msg = "invalid command \"" + value + "\""]() { return msg.c_str(); });
            }),
            "run specific command; possible values are help, version and import");
    boost::program_options::positional_options_description pos_options;
    pos_options.add("command", -1);

    boost::program_options::options_description database_options{"Database options", terminal_width};
    std::string host;
    std::uint16_t port;
    std::string user;
    std::string dbname;
    std::string password;
    database_options.add_options()
        ("database.host", boost::program_options::value<std::string>(&host), "name of host to connect to")
        ("database.port", boost::program_options::value<std::uint16_t>(&port), "port number to connect to at the server host")
        ("database.user", boost::program_options::value<std::string>(&user), "PostgreSQL user name to connect as")
        ("database.dbname", boost::program_options::value<std::string>(&dbname), "the database name")
        ("database.password", boost::program_options::value<std::string>(&password), "password used for password authentication.");

    boost::program_options::options_description log_options{"Logging options", terminal_width};
    std::vector<std::string> devices;
    std::string default_min_severity;
    log_options.add_options()
        ("log.device", boost::program_options::value<std::vector<std::string>>(&devices), "where to log (console/file/syslog)")
        ("log.min_severity", boost::program_options::value<std::string>(&default_min_severity),
         "do not log more trivial events; "
         "severity in descending order: emerg/alert/crit/err/warning/notice/info/debug/trace");

    boost::program_options::options_description log_console_options{"Logging on console options", terminal_width};
    std::string log_console_min_severity;
    log_console_options.add_options()
        ("log.console.min_severity",
         boost::program_options::value<std::string>(&log_console_min_severity),
         "do not log more trivial events; "
         "severity in descending order: emerg/alert/crit/err/warning/notice/info/debug/trace");

    boost::program_options::options_description log_file_options{"Logging into file options", terminal_width};
    std::string log_file_name;
    std::string log_file_min_severity;
    log_file_options.add_options()
        ("log.file.file_name",
         boost::program_options::value<std::string>(&log_file_name),
         "what file to log into")
        ("log.file.min_severity",
         boost::program_options::value<std::string>(&log_file_min_severity),
         "do not log more trivial events; "
         "severity in descending order: emerg/alert/crit/err/warning/notice/info/debug/trace");

    boost::program_options::options_description log_syslog_options{"Logging into syslog options", terminal_width};
    std::string ident;
    int facility_local_offset;
    std::string log_syslog_min_severity;
    log_syslog_options.add_options()
        ("log.syslog.ident", boost::program_options::value<std::string>(&ident), "what ident to log with (default is empty string)")
        ("log.syslog.facility",
         boost::program_options::value<int>(&facility_local_offset),
         "what LOG_LOCALx facility to log with (x in range 0..7, default means facility LOG_USER)")
        ("log.syslog.min_severity",
         boost::program_options::value<std::string>(&log_syslog_min_severity),
         "do not log more trivial events; "
         "severity in descending order: emerg/alert/crit/err/warning/notice/info/debug/trace");

    boost::program_options::options_description command_line_options{"cdnskey-processor-db-loader options", terminal_width};
    command_line_options
        .add(generic_options)
        .add(command_options)
        .add(log_options)
        .add(log_console_options)
        .add(log_file_options)
        .add(log_syslog_options)
        .add(database_options);
    boost::program_options::options_description config_file_options;
    config_file_options
        .add(log_options)
        .add(log_console_options)
        .add(log_file_options)
        .add(log_syslog_options)
        .add(database_options);

    boost::program_options::variables_map variables_map;
    const auto has_option = [&variables_map](const char* option)
    {
        return 0 < variables_map.count(option) &&
               !variables_map.at(option).defaulted();
    };
    try
    {
        boost::program_options::store(
                boost::program_options::command_line_parser(argc, argv)
                    .options(command_line_options)
                    .positional(pos_options)
                    .run(),
                variables_map);
    }
    catch (const boost::program_options::unknown_option& unknown_option)
    {
        std::ostringstream out;
        out << unknown_option.what() << "\n\n" << command_line_options;
        raise<UnknownOption>([msg = out.str()]() { return msg.c_str(); });
    }
    boost::program_options::notify(variables_map);
    conflicting_options(variables_map, "config", "version");
    conflicting_options(variables_map, "config", "help");
    conflicting_options(variables_map, "help", "version");

    const auto options_from = [&variables_map](const std::initializer_list<const char*>& list)
    {
        only_from(variables_map, list);
    };
    boost::apply_visitor(
            CheckOptions{
                    [&]() { options_from({"command"}); },
                    [&]() { options_from({"command"}); },
                    [&]() { options_from({"command", "config"}); }
            },
            command);
    if (has_command<Help>(command) || has_option("help"))
    {
        std::ostringstream out;
        out << command_line_options;
        raise<AllDone>([msg = out.str()]() { return msg.c_str(); });
    }
    if (has_command<Version>(command) || has_option("version"))
    {
        raise<AllDone>([msg = std::string{"Version: "} + SERVER_VERSION]() { return msg.c_str(); });
    }
    required_one_of(variables_map, {"command"});

    const bool config_file_name_presents = 0 < variables_map.count("config");
    if (config_file_name_presents)
    {
        this->config_file_name = opt_config_file_name;
        std::ifstream config_file{opt_config_file_name};
        if (!config_file)
        {
            raise<Exception>([msg = "can not open config file \"" + opt_config_file_name + "\""]() { return msg.c_str(); });
        }
        try
        {
            boost::program_options::store(
                    boost::program_options::parse_config_file(
                            config_file,
                            config_file_options),
                    variables_map);
        }
        catch (const boost::program_options::unknown_option& unknown_option)
        {
            std::ostringstream out;
            out << unknown_option.what() << "\n\n" << command_line_options;
            raise<UnknownOption>([msg = out.str()]() { return msg.c_str(); });
        }
        boost::program_options::notify(variables_map);
    }

    if (has_option("database.host"))
    {
        this->database.host = std::move(host);
    }
    if (has_option("database.port"))
    {
        this->database.port = port;
    }
    if (has_option("database.user"))
    {
        this->database.user = std::move(user);
    }
    if (has_option("database.dbname"))
    {
        this->database.dbname = std::move(dbname);
    }
    if (has_option("database.password"))
    {
        this->database.password = std::move(password);
    }

    if (has_option("log.device"))
    {
        boost::optional<LogSeverity> log_min_severity = LogSeverity::trace;
        if (has_option("log.min_severity"))
        {
            log_min_severity = make_severity(default_min_severity);
        }
        else
        {
            log_min_severity = boost::none;
        }
        std::for_each(begin(devices), end(devices), [&](auto&& device)
        {
            if (device == "console")
            {
                required_one_of(variables_map, {"log.min_severity", "log.console.min_severity"});
                this->log.console = Console{
                        .min_severity = has_option("log.console.min_severity")
                            ? make_severity(log_console_min_severity)
                            : *log_min_severity};
            }
            else if (device == "file")
            {
                required_one_of(variables_map, {"log.file.file_name"});
                required_one_of(variables_map, {"log.min_severity", "log.file.min_severity"});
                this->log.file = LogFile{
                        .file_name = log_file_name,
                        .min_severity = has_option("log.file.min_severity")
                            ? make_severity(log_file_min_severity)
                            : *log_min_severity};
            }
            else if (device == "syslog")
            {
                static constexpr int default_facility = LOG_USER;
                static const auto make_facility = [](int offset)
                {
                    switch (offset)
                    {
                        case 0: return LOG_LOCAL0;
                        case 1: return LOG_LOCAL1;
                        case 2: return LOG_LOCAL2;
                        case 3: return LOG_LOCAL3;
                        case 4: return LOG_LOCAL4;
                        case 5: return LOG_LOCAL5;
                        case 6: return LOG_LOCAL6;
                        case 7: return LOG_LOCAL7;
                    }
                    raise<Exception>([]() { return "option 'log.syslog.facility' out of range [0, 7]"; });
                };
                required_one_of(variables_map, {"log.min_severity", "log.syslog.min_severity"});
                this->log.syslog = SysLog{
                        .ident = has_option("log.syslog.ident") ? ident
                                                                : std::string{},
                        .facility = has_option("log.syslog.facility") ? make_facility(facility_local_offset)
                                                                      : default_facility,
                        .min_severity = has_option("log.syslog.min_severity")
                            ? make_severity(log_syslog_min_severity)
                            : *log_min_severity};
            }
            else
            {
                raise<Exception>([]() { return "Invalid value of log.device"; });
            }
        });
    }
}

Exception::Exception()
    : std::runtime_error{std::string{}}
{ }

} // namespace CdnskeyProcessor::DbLoader::Cfg
} // namespace CdnskeyProcessor::DbLoader
} // namespace CdnskeyProcessor

using namespace CdnskeyProcessor::DbLoader::Cfg;

namespace {

class Run : public boost::static_visitor<int>
{
public:
    explicit Run(std::function<int(const Options::Import&)> import)
        : import_{import}
    {}
    int operator()(const Options::Import& cfg) const
    {
        return import_(cfg);
    }
    template <typename T>
    [[noreturn]] int operator()(T&&) const
    {
        throw std::runtime_error{"no command specified"};
    }
private:
    std::function<int(const Options::Import&)> import_;
};

}//namespace {anonymous}

int CdnskeyProcessor::DbLoader::Cfg::run(
        const Options::Command& cmd,
        std::function<int(const Options::Import&)> import)
{
    return boost::apply_visitor(Run{import}, cmd);
}
