/*
 * Copyright (C) 2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef OPTIONS_HH_F615C636232F6D268D46554A81147AE9//date "+%s.%N"|md5sum|tr "[a-f]" "[A-F]"
#define OPTIONS_HH_F615C636232F6D268D46554A81147AE9

/**
 * config file example:

################################################################################
#                                   FRED-AKM                                   #
#                 Cdnskey Processor Database Loader tool for FRED              #
################################################################################

## For commented out options value present is the default

###
### Database connection settings
###
[database]
## Host name of the machine on which the database server is running
## or the directory for the Unix-domain socket (starting with ``/``)
host = {{ db.host }}

## TCP port or the Unix-domain socket file extension
## on which the database server is listening for connections
port = {{ db.port }}

## Connect to the database as this user instead of the default
## The user must have appropriate permissions.
user =  fred

## Name of the database intended for the domain registry, usually "fred"
dbname = {{ db.name }}

## Password for authentication by the database server (if required)
password = {{ db.app_pass }}


##
## Logging settings
##
[log]
## Minimal severity of messages to log
## Options (by severity in descending order):
##    emerg / alert / crit / err / warning / notice / info / debug / trace
min_severity = debug

## use the following option for each device: console, file, syslog
## (use the option multiple times for multiple devices)
device = syslog

## Device-specific settings - syslog
## User should specify facility in order to log
[log.syslog]

# Syslog facility - possible values: <0,7>
facility = 7

# Minimal severity of messages to log - stricter overrides the setting in [log] section
min_severity = debug

## Device-specific settings - file
## User should specify file name in order to log
[log.file]
# Path to log file
file_name = cdnskey-processor.log

# Minimal severity of messages to log - stricter overrides the setting in [log] section
min_severity = trace

## Device-specific settings - console
## user should specify severity in order to log
[log.console]
# Minimal severity of messages to log - stricter overrides the setting in [log] section
min_severity = error
 */

#include <boost/optional.hpp>
#include <boost/variant.hpp>

#include <cstdint>
#include <functional>
#include <stdexcept>
#include <string>

namespace CdnskeyProcessor {
namespace DbLoader {
namespace Cfg {

struct AllDone : std::exception { };

struct Exception : std::runtime_error
{
    explicit Exception();
};

struct UnknownOption : Exception { };

struct MissingOption : Exception { };

struct InappropriateOption : Exception { };

struct ConflictingOptions : Exception { };

struct InvalidCommand : Exception { };

struct InvalidValue : Exception { };

enum class LogSeverity
{
    trace,
    debug,
    info,
    warning,
    error,
    critical
};

class Options
{
public:
    Options() = delete;
    Options(const Options&) = delete;
    Options(Options&&) = delete;
    Options& operator=(const Options&) = delete;
    Options& operator=(Options&&) = delete;

    static const Options& get();
    static const Options& init(int argc, const char* const* argv);

    boost::optional<std::string> config_file_name;
    struct NoCommand { };
    struct Help { };
    struct Version { };
    struct Import { };
    using Command = boost::variant<NoCommand, Help, Version, Import>;
    Command command;
    struct Database
    {
        boost::optional<std::string> host;
        boost::optional<std::uint16_t> port;
        boost::optional<std::string> user;
        boost::optional<std::string> dbname;
        boost::optional<std::string> password;
    } database;
    struct Console
    {
        LogSeverity min_severity;
    };
    struct LogFile
    {
        std::string file_name;
        LogSeverity min_severity;
    };
    struct SysLog
    {
        std::string ident;
        int facility;
        LogSeverity min_severity;
    };
    struct Log
    {
        LogSeverity min_severity;
        boost::optional<Console> console;
        boost::optional<LogFile> file;
        boost::optional<SysLog> syslog;
    } log;
private:
    Options(int argc, const char* const* argv);
};

int run(const Options::Command& cmd, std::function<int(const Options::Import&)> import);

} // namespace CdnskeyProcessor::DbLoader::Cfg
} // namespace CdnskeyProcessor::DbLoader
} // namespace CdnskeyProcessor

#endif//OPTIONS_HH_F615C636232F6D268D46554A81147AE9
