/*
 * Copyright (C) 2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "db_loader/import.hh"
#include "src/storage.hh"

#include <boost/optional.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/uuid/string_generator.hpp>
#include <boost/uuid/uuid.hpp>
#include <boost/variant.hpp>

#include <liblog/liblog.hh>
#include <spdlog/fmt/ostr.h>

#include <libpg/query.hh>

#include <nlohmann/json.hpp>

#include <algorithm>
#include <chrono>
#include <cstdlib>
#include <set>
#include <string>
#include <tuple>
#include <utility>
#include <vector>

namespace boost {
namespace asio {
namespace ip {

std::string wrap_into_psql_representation(const boost::asio::ip::address&);

}//namespace boost::asio::ip
}//namespace boost::asio
}//namespace boost

namespace CdnskeyProcessor {
namespace DbLoader {

#if 0
JSON data example:

[
    {
        "queue_import_time_offset": "-7DAYS -22HOURS -30MINUTES",
        "batches":
        [
            {
                "items":
                [
                    {
                        "domain.fqdn": "secure-a.cz",
                        "domain.uid": "81569a01-f55e-4246-afb9-7f973d50af52",
                        "result":
                        {
                            "status": "secure",
                            "keys":
                            [
                                {
                                    "flags": 257,
                                    "proto": 3,
                                    "alg": 13,
                                    "key": "mfD5EGj0lhChuyNVDfF4K+vUvM72RZe4O4El1ti32MO9Jy12pWPDjbhoeiwvd5TiIerEXE/00MvfpMEKiqs/dA=="
                                }
                            ]
                        }
                    },
                    {
                        "domain.fqdn": "secure-b.cz",
                        "domain.uid": "9d2ba0dc-d6ef-46f6-8b8a-441516757ad3",
                        "result":
                        {
                            "status": "secure-empty"
                        }
                    },
                    {
                        "domain.fqdn": "secure-c.cz",
                        "domain.uid": "77cda79c-f886-40f8-be50-59a442b58bba",
                        "result":
                        {
                            "status": "untrustworthy"
                        }
                    },
                    {
                        "domain.fqdn": "secure-d.cz",
                        "domain.uid": "04173462-6e0b-46cf-826a-8225449c843a",
                        "result":
                        {
                            "status": "unknown"
                        }
                    }
                ]
            },
            {
                "items":
                [
                    {
                        "domain.fqdn": "insecure-a.cz",
                        "domain.uid": "c20346b2-c4e7-453b-9a8b-5bf3ddf4408e",
                        "nameservers":
                        [
                            {
                                "fqdn": "a.ns.cz",
                                "results":
                                [
                                    {
                                        "status": "insecure",
                                        "ip": "108.162.192.132",
                                        "flags": 257,
                                        "proto": 3,
                                        "alg": 13,
                                        "key": "mdsswUyr3DPW132mOi8V9xESWE8jTo0dxCjjnopKl+GqJxpVXckHAeF+KkxLbxILfDLUT0rAK9iUzy1L53eKGQ=="
                                    },
                                    {
                                        "status": "insecure",
                                        "ip": "2606:4700:50::adf5:3a84",
                                        "flags": 257,
                                        "proto": 3,
                                        "alg": 13,
                                        "key": "mdsswUyr3DPW132mOi8V9xESWE8jTo0dxCjjnopKl+GqJxpVXckHAeF+KkxLbxILfDLUT0rAK9iUzy1L53eKGQ=="
                                    }
                                ]
                            },
                            {
                                "fqdn": "b.ns.cz",
                                "results":
                                [
                                    {
                                        "status": "insecure",
                                        "ip": "108.162.193.240",
                                        "flags": 257,
                                        "proto": 3,
                                        "alg": 13,
                                        "key": "mdsswUyr3DPW132mOi8V9xESWE8jTo0dxCjjnopKl+GqJxpVXckHAeF+KkxLbxILfDLUT0rAK9iUzy1L53eKGQ=="
                                    },
                                    {
                                        "status": "insecure",
                                        "ip": "2606:4700:58::adf5:3bf0",
                                        "flags": 257,
                                        "proto": 3,
                                        "alg": 13,
                                        "key": "mdsswUyr3DPW132mOi8V9xESWE8jTo0dxCjjnopKl+GqJxpVXckHAeF+KkxLbxILfDLUT0rAK9iUzy1L53eKGQ=="
                                    }
                                ]
                            }
                        ]
                    },
                    {
                        "domain.fqdn": "insecure-b.cz",
                        "domain.uid": "a4869a0c-c01f-4972-be78-146281264d23",
                        "nameservers":
                        [
                            {
                                "fqdn": "a.ns.cz",
                                "results":
                                [
                                    {
                                        "status": "insecure-empty",
                                        "ip": "108.162.194.78"
                                    },
                                    {
                                        "status": "insecure-empty",
                                        "ip": "2606:4700:50::a29f:264e"
                                    }
                                ]
                            }
                        ]
                    },
                    {
                        "domain.fqdn": "insecure-c.cz",
                        "domain.uid": "d759a6fc-cfc6-41fe-ba1d-d2255de3b1f3",
                        "nameservers":
                        [
                            {
                                "fqdn": "a.ns.cz",
                                "results":
                                [
                                    {
                                        "status": "unresolved",
                                        "ip": "218.98.111.202"
                                    },
                                    {
                                        "status": "unresolved",
                                        "ip": "2001:df2:f080:1::10"
                                    }
                                ]
                            }
                        ]
                    },
                    {
                        "domain.fqdn": "insecure-d.cz",
                        "domain.uid": "6c19101a-a90c-41f9-8b50-6315e64169f2",
                        "nameservers":
                        [
                            {
                                "fqdn": "a.ns.cz",
                                "results":
                                [
                                    {
                                        "status": "insecure-empty",
                                        "ip": "81.2.209.185"
                                    },
                                    {
                                        "status": "insecure-empty",
                                        "ip": "2001:15e8:201:1::d1b9"
                                    }
                                ]
                            },
                            {
                                "fqdn": "b.ns.cz",
                                "results":
                                [
                                    {
                                        "status": "unresolved-ip"
                                    }
                                ]
                            }
                        ]
                    }
                ]
            }
        ]
    },
    {
        "queue_import_time_offset": "-6DAYS -22HOURS -30MINUTES",
        "batches":
        [
            {
                "items":
                [
                    {
                        "domain.fqdn": "secure-a.cz",
                        "domain.uid": "81569a01-f55e-4246-afb9-7f973d50af52",
                        "result":
                        {
                            "status": "secure",
                            "keys":
                            [
                                {
                                    "flags": 257,
                                    "proto": 3,
                                    "alg": 13,
                                    "key": "mfD5EGj0lhChuyNVDfF4K+vUvM72RZe4O4El1ti32MO9Jy12pWPDjbhoeiwvd5TiIerEXE/00MvfpMEKiqs/dA=="
                                }
                            ]
                        }
                    },
                    {
                        "domain.fqdn": "secure-b.cz",
                        "domain.uid": "9d2ba0dc-d6ef-46f6-8b8a-441516757ad3",
                        "result":
                        {
                            "status": "secure-empty"
                        }
                    },
                    {
                        "domain.fqdn": "secure-c.cz",
                        "domain.uid": "77cda79c-f886-40f8-be50-59a442b58bba",
                        "result":
                        {
                            "status": "untrustworthy"
                        }
                    },
                    {
                        "domain.fqdn": "secure-d.cz",
                        "domain.uid": "04173462-6e0b-46cf-826a-8225449c843a",
                        "result":
                        {
                            "status": "unknown"
                        }
                    }
                ]
            },
            {
                "items":
                [
                    {
                        "domain.fqdn": "insecure-a.cz",
                        "domain.uid": "c20346b2-c4e7-453b-9a8b-5bf3ddf4408e",
                        "nameservers":
                        [
                            {
                                "fqdn": "a.ns.cz",
                                "results":
                                [
                                    {
                                        "status": "insecure",
                                        "ip": "108.162.192.132",
                                        "flags": 257,
                                        "proto": 3,
                                        "alg": 13,
                                        "key": "mdsswUyr3DPW132mOi8V9xESWE8jTo0dxCjjnopKl+GqJxpVXckHAeF+KkxLbxILfDLUT0rAK9iUzy1L53eKGQ=="
                                    }
                                ]
                            },
                            {
                                "fqdn": "b.ns.cz",
                                "results":
                                [
                                    {
                                        "status": "insecure",
                                        "ip": "108.162.193.240",
                                        "flags": 257,
                                        "proto": 3,
                                        "alg": 13,
                                        "key": "mdsswUyr3DPW132mOi8V9xESWE8jTo0dxCjjnopKl+GqJxpVXckHAeF+KkxLbxILfDLUT0rAK9iUzy1L53eKGQ=="
                                    }
                                ]
                            }
                        ]
                    },
                    {
                        "domain.fqdn": "insecure-b.cz",
                        "domain.uid": "a4869a0c-c01f-4972-be78-146281264d23",
                        "nameservers":
                        [
                            {
                                "fqdn": "a.ns.cz",
                                "results":
                                [
                                    {
                                        "status": "insecure-empty",
                                        "ip": "108.162.194.78"
                                    }
                                ]
                            }
                        ]
                    },
                    {
                        "domain.fqdn": "insecure-c.cz",
                        "domain.uid": "d759a6fc-cfc6-41fe-ba1d-d2255de3b1f3",
                        "nameservers":
                        [
                            {
                                "fqdn": "a.ns.cz",
                                "results":
                                [
                                    {
                                        "status": "unresolved",
                                        "ip": "218.98.111.202"
                                    }
                                ]
                            }
                        ]
                    },
                    {
                        "domain.fqdn": "insecure-d.cz",
                        "domain.uid": "6c19101a-a90c-41f9-8b50-6315e64169f2",
                        "nameservers":
                        [
                            {
                                "fqdn": "a.ns.cz",
                                "results":
                                [
                                    {
                                        "status": "insecure-empty",
                                        "ip": "81.2.209.185"
                                    },
                                    {
                                        "status": "insecure-empty",
                                        "ip": "2001:15e8:201:1::d1b9"
                                    }
                                ]
                            },
                            {
                                "fqdn": "b.ns.cz",
                                "results":
                                [
                                    {
                                        "status": "unresolved-ip"
                                    }
                                ]
                            }
                        ]
                    }
                ]
            }
        ]
    }
]
#endif

namespace {

struct InsecureItem
{
    struct InsecureResult
    {
        std::string ip;
        unsigned flags;
        unsigned proto;
        unsigned alg;
        std::string key;
    };
    struct InsecureEmptyResult
    {
        std::string ip;
    };
    struct UnresolvedResult
    {
        std::string ip;
    };
    struct UnresolvedIpResult { };
    using Result = boost::variant<boost::blank, InsecureResult, InsecureEmptyResult, UnresolvedResult, UnresolvedIpResult>;
    struct Nameserver
    {
        std::string fqdn;
        std::vector<Result> results;
    };
    unsigned long long id = 0;
    std::string fqdn;
    boost::uuids::uuid uuid;
    std::vector<Nameserver> nameservers;
};

struct SecureItem
{
    struct DnsKey
    {
        unsigned flags;
        unsigned proto;
        unsigned alg;
        std::string key;
        friend bool operator<(const DnsKey& lhs, const DnsKey& rhs) noexcept
        {
            static const auto make_dns_key_tuple = [](const DnsKey& dns_key)
            {
                return std::make_tuple(dns_key.flags, dns_key.proto, dns_key.alg, dns_key.key);
            };
            return make_dns_key_tuple(lhs) < make_dns_key_tuple(rhs);
        }
    };
    using SecureResult = std::set<DnsKey>;
    struct SecureEmptyResult { };
    struct UntrustworthyResult { };
    struct UnknownResult { };
    unsigned long long id = 0;
    std::string fqdn;
    boost::uuids::uuid uuid;
    using Result = boost::variant<boost::blank, SecureResult, SecureEmptyResult, UntrustworthyResult, UnknownResult>;
    Result result;
};

struct Job
{
    unsigned long long id = 0;
    using Item = boost::variant<InsecureItem, SecureItem>;
    Item item;
};

struct Batch
{
    unsigned long long id = 0;
    std::vector<Job> jobs;
};

struct Queue
{
    std::string time_offset;
    unsigned long long id = 0;
    std::vector<Batch> batches;
};

InsecureItem::Result get_nameserver_ip(const nlohmann::json& json_data)
{
    const std::string status = json_data["status"];
    if (status == "unresolved-ip")
    {
        return InsecureItem::UnresolvedIpResult{};
    }
    std::string ip = json_data["ip"];
    if (status == "insecure")
    {
        InsecureItem::InsecureResult result;
        result.ip = std::move(ip);
        result.flags = json_data["flags"];
        result.proto = json_data["proto"];
        result.alg = json_data["alg"];
        result.key = json_data["key"];
        return result;
    }
    if (status == "insecure-empty")
    {
        InsecureItem::InsecureEmptyResult result;
        result.ip = std::move(ip);
        return result;
    }
    if (status == "unresolved")
    {
        InsecureItem::UnresolvedResult result;
        result.ip = std::move(ip);
        return result;
    }
    throw std::runtime_error{"unexpected status " + status};
}

InsecureItem::Nameserver get_nameserver(const nlohmann::json& json_data)
{
    InsecureItem::Nameserver result;
    result.fqdn = json_data["fqdn"];
    const auto ip_iter = json_data.find("results");
    if (ip_iter != end(json_data))
    {
        std::for_each(begin(*ip_iter), end(*ip_iter), [&](auto&& nameserver_ip)
        {
            result.results.push_back(get_nameserver_ip(nameserver_ip));
        });
    }
    return result;
}

InsecureItem get_insecure_item(const std::string& domain, const std::string& uid, const nlohmann::json& json_data)
{
    InsecureItem item;
    item.fqdn = domain;
    item.uuid = boost::uuids::string_generator{}(uid);
    std::for_each(begin(json_data), end(json_data), [&](auto&& nameserver)
    {
        item.nameservers.push_back(get_nameserver(nameserver));
    });
    return item;
}

SecureItem get_secure_item(const std::string& domain, const std::string& uid, const nlohmann::json& json_data)
{
    SecureItem item;
    item.fqdn = domain;
    item.uuid = boost::uuids::string_generator{}(uid);
    const std::string status = json_data["status"];
    if (status == "secure")
    {
        const auto keys_iter = json_data.find("keys");
        const bool has_keys = keys_iter != end(json_data);
        if (has_keys)
        {
            SecureItem::SecureResult keys;
            std::for_each(begin(*keys_iter), end(*keys_iter), [&keys](auto&& json_key)
            {
                SecureItem::DnsKey dns_key;
                dns_key.flags = json_key["flags"];
                dns_key.proto = json_key["proto"];
                dns_key.alg = json_key["alg"];
                dns_key.key = json_key["key"];
                keys.insert(std::move(dns_key));
            });
            item.result = keys;
        }
        return item;
    }
    if (status == "secure-empty")
    {
        item.result = SecureItem::SecureEmptyResult{};
        return item;
    }
    if (status == "untrustworthy")
    {
        item.result = SecureItem::UntrustworthyResult{};
        return item;
    }
    if (status == "unknown")
    {
        item.result = SecureItem::UnknownResult{};
        return item;
    }
    throw std::runtime_error{"unexpected status " + status};
}

SecureItem get_secure_item(const std::string& domain, const std::string& uid)
{
    SecureItem result;
    result.fqdn = domain;
    result.uuid = boost::uuids::string_generator{}(uid);
    return result;
}

Job get_item(const nlohmann::json& json_data)
{
    const std::string domain = json_data["domain.fqdn"];
    const std::string uid = json_data["domain.uid"];
    const auto result_iter = json_data.find("result");
    const bool has_result = result_iter != end(json_data);
    Job job;
    if (has_result)
    {
        job.item = get_secure_item(domain, uid, *result_iter);
        return job;
    }
    const auto nameservers_iter = json_data.find("nameservers");
    const bool is_insecure_item = nameservers_iter != end(json_data);
    if (is_insecure_item)
    {
        job.item = get_insecure_item(domain, uid, *nameservers_iter);
        return job;
    }
    job.item = get_secure_item(domain, uid);
    return job;
}

Batch get_batch(const nlohmann::json& json_data)
{
    Batch result;
    const auto& items = json_data["items"];
    std::for_each(begin(items), end(items), [&result](auto&& item)
    {
        result.jobs.push_back(get_item(item));
    });
    return result;
}

Queue get_queue(const nlohmann::json& json_data)
{
    Queue result;
    result.time_offset = json_data["queue_import_time_offset"];
    const auto& batches = json_data["batches"];
    std::for_each(begin(batches), end(batches), [&result](auto&& batch)
    {
        result.batches.push_back(get_batch(batch));
    });
    return result;
}

std::vector<Queue> get_queues(const nlohmann::json& json_data)
{
    std::vector<Queue> result;
    std::for_each(begin(json_data), end(json_data), [&result](auto&& queue)
    {
        result.push_back(get_queue(queue));
    });
    return result;
}

struct InsertJob : boost::static_visitor<unsigned long long>
{
    explicit InsertJob(const LibPg::PgRwTransaction& tx, const Queue& queue, const Batch& batch, const Job& job)
        : tx{tx},
          queue{queue},
          batch{batch},
          job{job}
    { }
    unsigned long long operator()(InsecureItem& item) const
    {
        using Nameserver0 = StrongString<struct Nameserver0Tag_>;
        using Nameserver1 = StrongString<struct Nameserver1Tag_>;
        using Nameserver2 = StrongString<struct Nameserver2Tag_>;
        using Nameserver3 = StrongString<struct Nameserver3Tag_>;
        using Nameserver4 = StrongString<struct Nameserver4Tag_>;
        using Nameserver5 = StrongString<struct Nameserver5Tag_>;
        using Nameserver6 = StrongString<struct Nameserver6Tag_>;
        using Nameserver7 = StrongString<struct Nameserver7Tag_>;
        using Nameserver8 = StrongString<struct Nameserver8Tag_>;
        using Nameserver9 = StrongString<struct Nameserver9Tag_>;
        const auto db_res = exec(tx, LibPg::make_query(
                "WITH od AS "
                "("
                    "SELECT domain.id, d.fqdn, d.uid, d.ns "
                      "FROM (VALUES (") << LibPg::parameter<DomainFqdn>().as_text() << ", "
                                        << LibPg::parameter<DomainUid>().as_text() << ", "
                                    "ARRAY[" << LibPg::parameter<Nameserver0::Optional>().as_text() << ", "
                                             << LibPg::parameter<Nameserver1::Optional>().as_text() << ", "
                                             << LibPg::parameter<Nameserver2::Optional>().as_text() << ", "
                                             << LibPg::parameter<Nameserver3::Optional>().as_text() << ", "
                                             << LibPg::parameter<Nameserver4::Optional>().as_text() << ", "
                                             << LibPg::parameter<Nameserver5::Optional>().as_text() << ", "
                                             << LibPg::parameter<Nameserver6::Optional>().as_text() << ", "
                                             << LibPg::parameter<Nameserver7::Optional>().as_text() << ", "
                                             << LibPg::parameter<Nameserver8::Optional>().as_text() << ", "
                                             << LibPg::parameter<Nameserver9::Optional>().as_text() << "])"
                           ") d(fqdn, uid, ns) "
                 "LEFT JOIN domain ON domain.fqdn = d.fqdn AND "
                                     "domain.uid = d.uid"
                "), "
                "nd AS "
                "("
                    "INSERT INTO domain (fqdn, uid) "
                         "SELECT fqdn, uid "
                           "FROM od "
                          "WHERE id IS NULL "
                      "RETURNING id, fqdn, uid"
                "), "
                "d AS "
                "("
                    "SELECT * FROM od WHERE id IS NOT NULL "
                    "UNION "
                    "SELECT nd.id, nd.fqdn, nd.uid, od.ns "
                      "FROM nd "
                      "JOIN od ON od.fqdn = nd.fqdn AND od.uid = nd.uid"
                "), "
                "s AS "
                "("
                    "INSERT INTO scan (scan_queue_id, domain_id) "
                         "SELECT " << LibPg::parameter<ScanQueueId>().as_big_int() << ", "
                                "id "
                           "FROM d "
                      "RETURNING id, domain_id"
                "), "
                "ons AS "
                "("
                    "SELECT n.id, ans.fqdn "
                      "FROM (SELECT DISTINCT UNNEST(ns) AS fqdn FROM od) ans "
                 "LEFT JOIN nameserver n ON n.fqdn = ans.fqdn "
                     "WHERE ans.fqdn IS NOT NULL"
                "), "
                "nns AS "
                "("
                    "INSERT INTO nameserver (fqdn) SELECT fqdn FROM ons WHERE id IS NULL "
                      "RETURNING id, fqdn"
                "), "
                "ns AS "
                "("
                    "SELECT * FROM ons WHERE id IS NOT NULL "
                    "UNION "
                    "SELECT * FROM nns"
                "), "
                "i AS "
                "("
                    "INSERT INTO insecure_scan (scan_id, nameserver_id) "
                         "SELECT s.id, ns.id "
                           "FROM (SELECT id, fqdn, uid, UNNEST(ns) "
                                   "FROM d) AS dns(domain_id, domain_fqdn, uid, ns_fqdn) "
                           "JOIN s ON s.domain_id = dns.domain_id "
                           "JOIN ns ON ns.fqdn = dns.ns_fqdn"
                "), "
                "j AS "
                "("
                    "INSERT INTO scan_job (scan_batch_id, scan_id) "
                         "SELECT " << LibPg::parameter<ScanBatchId>().as_big_int() << ", id "
                           "FROM s "
                      "RETURNING id"
                ") "
                "SELECT" << LibPg::item<unsigned long long>() << "s.id"
                         << LibPg::item<ScanJobId>() << "j.id "
                  "FROM s, "
                       "j",
                ScanQueueId{queue.id},
                ScanBatchId{static_cast<std::int64_t>(batch.id)},
                DomainFqdn{item.fqdn},
                DomainUid{to_string(item.uuid)},
                0 < item.nameservers.size() ? Nameserver0::Optional{Nameserver0{item.nameservers.at(0).fqdn}}
                                            : Nameserver0::nullopt,
                1 < item.nameservers.size() ? Nameserver1::Optional{Nameserver1{item.nameservers.at(1).fqdn}}
                                            : Nameserver1::nullopt,
                2 < item.nameservers.size() ? Nameserver2::Optional{Nameserver2{item.nameservers.at(2).fqdn}}
                                            : Nameserver2::nullopt,
                3 < item.nameservers.size() ? Nameserver3::Optional{Nameserver3{item.nameservers.at(3).fqdn}}
                                            : Nameserver3::nullopt,
                4 < item.nameservers.size() ? Nameserver4::Optional{Nameserver4{item.nameservers.at(4).fqdn}}
                                            : Nameserver4::nullopt,
                5 < item.nameservers.size() ? Nameserver5::Optional{Nameserver5{item.nameservers.at(5).fqdn}}
                                            : Nameserver5::nullopt,
                6 < item.nameservers.size() ? Nameserver6::Optional{Nameserver6{item.nameservers.at(6).fqdn}}
                                            : Nameserver6::nullopt,
                7 < item.nameservers.size() ? Nameserver7::Optional{Nameserver7{item.nameservers.at(7).fqdn}}
                                            : Nameserver7::nullopt,
                8 < item.nameservers.size() ? Nameserver8::Optional{Nameserver8{item.nameservers.at(8).fqdn}}
                                            : Nameserver8::nullopt,
                9 < item.nameservers.size() ? Nameserver9::Optional{Nameserver9{item.nameservers.at(9).fqdn}}
                                            : Nameserver9::nullopt);
        item.id = db_res.front().template get<unsigned long long>();
        LIBLOG_DEBUG("scan_id = {}", item.id);
        return *db_res.front().template get<ScanJobId>();
    }
    unsigned long long operator()(SecureItem& item) const
    {
        const auto db_res = exec(tx, LibPg::make_query(
                "WITH od AS "
                "("
                    "SELECT domain.id, d.fqdn, d.uid "
                      "FROM (VALUES (") << LibPg::parameter<DomainFqdn>().as_text() << ", "
                                        << LibPg::parameter<DomainUid>().as_text() << ")) d(fqdn, uid) "
                 "LEFT JOIN domain ON domain.fqdn = d.fqdn AND "
                                     "domain.uid = d.uid"
                "), "
                "nd AS "
                "("
                    "INSERT INTO domain (fqdn, uid) "
                        "SELECT fqdn, uid "
                        "FROM od "
                        "WHERE id IS NULL "
                    "RETURNING id, fqdn, uid"
                "), "
                "d AS "
                "("
                    "SELECT * FROM od WHERE id IS NOT NULL "
                    "UNION "
                    "SELECT * FROM nd"
                "), "
                "s AS "
                "("
                    "INSERT INTO scan (scan_queue_id, domain_id) "
                         "SELECT " << LibPg::parameter<ScanQueueId>().as_big_int() << ", "
                                "id "
                           "FROM d "
                      "RETURNING id"
                "), "
                "j AS "
                "("
                    "INSERT INTO scan_job (scan_batch_id, scan_id) "
                         "SELECT " << LibPg::parameter<ScanBatchId>().as_big_int() << ", id "
                           "FROM s "
                      "RETURNING id"
                ") "
                "SELECT" << LibPg::item<unsigned long long>() << "s.id"
                         << LibPg::item<ScanJobId>() << "j.id "
                  "FROM s, "
                       "j",
                ScanQueueId{queue.id},
                ScanBatchId{static_cast<std::int64_t>(batch.id)},
                DomainFqdn{item.fqdn},
                DomainUid{to_string(item.uuid)});
        item.id = db_res.front().get<unsigned long long>();
        LIBLOG_DEBUG("scan_id = {}", item.id);
        return *db_res.front().get<ScanJobId>();
    }
    const LibPg::PgRwTransaction& tx;
    const Queue& queue;
    const Batch& batch;
    const Job& job;
};

void set_job(const LibPg::PgRwTransaction& tx, const Queue& queue, const Batch& batch, Job& job)
{
    job.id = [&]()
    {
        return boost::apply_visitor(InsertJob{tx, queue, batch, job}, job.item);
    }();
    LIBLOG_DEBUG("job_id = {}", job.id);
}

void set_batch(const LibPg::PgRwTransaction& tx, const Queue& queue, Batch& batch)
{
    if (batch.jobs.empty())
    {
        return;
    }
    if (batch.id == 0)
    {
        const auto db_res = exec(tx, LibPg::make_query(
                "WITH b AS "
                "("
                    "INSERT INTO scan_batch DEFAULT VALUES "
                      "RETURNING id"
                "), "
                "p AS "
                "("
                    "INSERT INTO scan_batch_progress ("
                                        "scan_batch_id, "
                                        "worker_id, "
                                        "sent_at) "
                         "SELECT b.id, "
                                "w.id, "
                                "q.import_started_at + '2SECONDS'::INTERVAL "
                           "FROM worker w "
                     "CROSS JOIN b "
                     "CROSS JOIN scan_queue q "
                          "WHERE w.removed_at IS NULL AND "
                                "q.id = ") << LibPg::parameter<unsigned long long>().as_big_int() <<
                ") "
                "SELECT" << LibPg::item<unsigned long long>() << "b.id "
                  "FROM b",
                queue.id);
        if (db_res.size() != LibPg::RowIndex{1})
        {
            struct UnexpectedResult : std::exception
            {
                const char* what() const noexcept override { return "unexpected number of inserted scan_batch rows"; }
            };
            throw UnexpectedResult{};
        }
        batch.id = db_res.front().get<unsigned long long>();
        LIBLOG_DEBUG("batch_id = {}", batch.id);
    }
    std::for_each(begin(batch.jobs), end(batch.jobs), [&](auto&& job)
    {
        set_job(tx, queue, batch, job);
    });
}

unsigned long long open_scan_queue(const LibPg::PgRwTransaction& tx, const std::string& time_offset)
{
    const auto db_res = exec(tx, LibPg::make_query(
            "INSERT INTO scan_queue (import_started_at) VALUES (NOW()::DATE + ") <<
                LibPg::parameter<std::string>().as_type("INTERVAL") << ") "
              "RETURNING" << LibPg::item<unsigned long long>() << "id",
            time_offset);
    if (db_res.size() != LibPg::RowIndex{1})
    {
        struct UnexpectedResult : std::exception
        {
            const char* what() const noexcept override { return "unexpected number of inserted scan_queue rows"; }
        };
        throw UnexpectedResult{};
    }
    return db_res.front().get<unsigned long long>();
}

void close_scan_queue(const LibPg::PgRwTransaction& tx, unsigned long long scan_queue_id)
{
    const auto db_res = exec(tx, LibPg::make_query(
            "UPDATE scan_queue "
               "SET import_finished_at = import_started_at + '1SECOND'::INTERVAL "
             "WHERE id = ") << LibPg::parameter<unsigned long long>().as_big_int() << " AND "
                   "import_finished_at IS NULL", scan_queue_id);
    if (db_res.get_number_of_rows_affected() == LibPg::RowIndex{0})
    {
        struct UnexpectedResult : std::exception
        {
            const char* what() const noexcept override { return "no open scan queue found"; }
        };
        throw UnexpectedResult{};
    }
    if (LibPg::RowIndex{1} < db_res.get_number_of_rows_affected())
    {
        struct UnexpectedResult : std::exception
        {
            const char* what() const noexcept override { return "too many scan queues found"; }
        };
        throw UnexpectedResult{};
    }
}

void set_queue(const LibPg::PgRwTransaction& tx, Queue& queue)
{
    queue.id = open_scan_queue(tx, queue.time_offset);
    LIBLOG_DEBUG("queue_id = {}", queue.id);
    std::for_each(begin(queue.batches), end(queue.batches), [&tx, &queue](auto&& batch)
    {
        set_batch(tx, queue, batch);
    });
    close_scan_queue(tx, queue.id);
}

struct SetResult;

struct SetInsecureResult : boost::static_visitor<void>
{
    explicit SetInsecureResult(const SetResult& set_result, const std::string& nameserver)
        : set_result{set_result},
          nameserver{nameserver}
    { }
    void operator()(const boost::blank&) const { }
    void operator()(const InsecureItem::InsecureResult&) const;
    void operator()(const InsecureItem::InsecureEmptyResult&) const;
    void operator()(const InsecureItem::UnresolvedResult&) const;
    void operator()(const InsecureItem::UnresolvedIpResult&) const;
    const SetResult& set_result;
    const std::string& nameserver;
};

struct SetSecureResult : boost::static_visitor<void>
{
    explicit SetSecureResult(const SetResult& set_result)
        : set_result{set_result}
    { }
    void operator()(const boost::blank&) const { }
    void operator()(const SecureItem::SecureResult&) const;
    void operator()(const SecureItem::SecureEmptyResult&) const;
    void operator()(const SecureItem::UntrustworthyResult&) const;
    void operator()(const SecureItem::UnknownResult&) const;
    const SetResult& set_result;
};

struct SecureResultIsEmpty : boost::static_visitor<bool>
{
    bool operator()(const boost::blank&) const { return true; }
    bool operator()(const SecureItem::SecureResult&) const { return false; }
    bool operator()(const SecureItem::SecureEmptyResult&) const { return false; }
    bool operator()(const SecureItem::UntrustworthyResult&) const { return false; }
    bool operator()(const SecureItem::UnknownResult&) const { return false; }
};

struct SetResult : boost::static_visitor<void>
{
    explicit SetResult(const LibPg::PgRwTransaction& tx, const Queue& queue, const Batch& batch, unsigned long long job_id)
        : tx{tx},
          queue{queue},
          batch{batch},
          job_id{job_id}
    { }
    void operator()(const InsecureItem& item) const
    {
        std::for_each(begin(item.nameservers), end(item.nameservers), [&](auto&& nameserver)
        {
            std::for_each(begin(nameserver.results), end(nameserver.results), [&](auto&& result)
            {
                boost::apply_visitor(SetInsecureResult{*this, nameserver.fqdn}, result);
            });
        });
    }
    void operator()(const SecureItem& item) const
    {
        if (boost::apply_visitor(SecureResultIsEmpty{}, item.result))
        {
            return;
        }
        boost::apply_visitor(SetSecureResult{*this}, item.result);
        set_times();
    }
    void set_times() const
    {
        exec(tx, LibPg::make_query(
                "WITH q AS "
                "("
                    "UPDATE scan_queue "
                       "SET solving_started_at = import_finished_at + '3SECONDS'::INTERVAL, "
                           "solving_finished_at = import_finished_at + '33SECONDS'::INTERVAL "
                     "WHERE solving_started_at IS NULL AND "
                            "solving_finished_at IS NULL AND "
                            "import_finished_at IS NOT NULL AND "
                           "id = ") << LibPg::parameter<ScanQueueId>().as_big_int() <<
                ") "
                "UPDATE scan_batch_progress "
                   "SET started_at = q.import_finished_at + '4SECONDS'::INTERVAL, "
                       "finished_at = q.import_finished_at + '32SECONDS'::INTERVAL "
                  "FROM scan_queue q "
                 "WHERE started_at IS NULL AND "
                       "finished_at IS NULL AND "
                       "scan_batch_id = " << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                       "q.id = " << LibPg::parameter<ScanQueueId>().as_big_int() << " AND "
                       "q.import_finished_at IS NOT NULL",
                ScanQueueId{queue.id},
                ScanBatchId{static_cast<std::int64_t>(batch.id)});
    }
    const LibPg::PgRwTransaction& tx;
    const Queue& queue;
    const Batch& batch;
    unsigned long long job_id;
};

void SetInsecureResult::operator()(const InsecureItem::InsecureResult& result) const
{
    const auto dbres = exec(set_result.tx, LibPg::make_query(
            "WITH d AS "
            "("
                "SELECT w.id AS worker_id, "
                       "s.domain_id, "
                       "i.nameserver_id "
                  "FROM worker w "
                  "JOIN scan_batch_progress p ON p.worker_id = w.id "
                  "JOIN scan_job j ON j.scan_batch_id = p.scan_batch_id "
                  "JOIN scan s ON s.id = j.scan_id "
                  "JOIN insecure_scan i ON i.scan_id = s.id "
                  "JOIN nameserver n ON n.id = i.nameserver_id "
                 "WHERE w.removed_at IS NULL AND "
                       "p.scan_batch_id = ") << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                       "j.id = " << LibPg::parameter<ScanJobId>().as_big_int() << " AND "
                       "n.fqdn = " << LibPg::parameter<Nameserver>().as_text() <<
            ") "
            "INSERT INTO scan_result "
            "("
                "scan_job_id, "
                "worker_id, "
                "nameserver_id, "
                "nameserver_ip, "
                "scanned_at, "
                "scan_status, "
                "cdnskey_flags, "
                "cdnskey_proto, "
                "cdnskey_alg, "
                "cdnskey_public_key"
            ") SELECT " << LibPg::parameter<ScanJobId>().as_big_int() << ", "
                     "d.worker_id, "
                     "d.nameserver_id, "
                        << LibPg::parameter<IpAddress>().as_type("INET") << ", "
                     "q.import_started_at + '10SECONDS'::INTERVAL, "
                     "'insecure'::cdnskey_status, "
                        << LibPg::parameter<CdnskeyFlags>().as_integer() << ", "
                        << LibPg::parameter<CdnskeyProto>().as_integer() << ", "
                        << LibPg::parameter<CdnskeyAlg>().as_integer() << ", "
                        << LibPg::parameter<CdnskeyPublicKey>().as_text() << " "
                "FROM d "
          "CROSS JOIN scan_queue q "
               "WHERE q.id = " << LibPg::parameter<ScanQueueId>().as_big_int() << " AND "
                     "q.import_started_at IS NOT NULL "
              "RETURNING" << LibPg::item<std::int64_t>() << "id",
            ScanBatchId{static_cast<std::int64_t>(set_result.batch.id)},
            ScanJobId{static_cast<std::int64_t>(set_result.job_id)},
            Nameserver{nameserver},
            IpAddress{boost::asio::ip::address::from_string(result.ip)},
            ScanQueueId{static_cast<std::uint64_t>(set_result.queue.id)},
            CdnskeyFlags{result.flags},
            CdnskeyProto{result.proto},
            CdnskeyAlg{result.alg},
            CdnskeyPublicKey{result.key});
    if (dbres.empty())
    {
        throw std::runtime_error{"no job joinable with this result"};
    }
    set_result.set_times();
}

void store_result(
        const SetResult& set_result,
        const std::string& nameserver,
        const std::string& nameserver_ip,
        const char* cdnskey_status)
{
    const auto dbres = exec(set_result.tx, LibPg::make_query(
            "WITH d AS "
            "("
                "SELECT w.id AS worker_id, "
                       "s.domain_id, "
                       "i.nameserver_id "
                  "FROM worker w "
                  "JOIN scan_batch_progress p ON p.worker_id = w.id "
                  "JOIN scan_job j ON j.scan_batch_id = p.scan_batch_id "
                  "JOIN scan s ON s.id = j.scan_id "
                  "JOIN insecure_scan i ON i.scan_id = s.id "
                  "JOIN nameserver n ON n.id = i.nameserver_id "
                 "WHERE w.removed_at IS NULL AND "
                       "p.scan_batch_id = ") << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                       "j.id = " << LibPg::parameter<ScanJobId>().as_big_int() << " AND "
                       "n.fqdn = " << LibPg::parameter<Nameserver>().as_text() <<
            ") "
            "INSERT INTO scan_result "
            "("
                "scan_job_id, "
                "worker_id, "
                "nameserver_id, "
                "nameserver_ip, "
                "scanned_at, "
                "scan_status"
            ") SELECT " << LibPg::parameter<ScanJobId>().as_big_int() << ", "
                     "d.worker_id, "
                     "d.nameserver_id, "
                        << LibPg::parameter<IpAddress>().as_type("INET") << ", "
                     "q.import_started_at + '10SECONDS'::INTERVAL, "
                        << LibPg::parameter<const char*>().as_type("cdnskey_status") << " "
                "FROM d "
          "CROSS JOIN scan_queue q "
               "WHERE q.id = " << LibPg::parameter<ScanQueueId>().as_big_int() << " AND "
                     "q.import_started_at IS NOT NULL "
              "RETURNING" << LibPg::item<std::int64_t>() << "id",
            ScanBatchId{static_cast<std::int64_t>(set_result.batch.id)},
            ScanJobId{static_cast<std::int64_t>(set_result.job_id)},
            Nameserver{nameserver},
            IpAddress{boost::asio::ip::address::from_string(nameserver_ip)},
            ScanQueueId{static_cast<std::uint64_t>(set_result.queue.id)},
            cdnskey_status);
    if (dbres.empty())
    {
        throw std::runtime_error{"no job joinable with this result"};
    }
    set_result.set_times();
}

void store_result(
        const SetResult& set_result,
        const std::string& nameserver,
        const char* cdnskey_status)
{
    const auto dbres = exec(set_result.tx, LibPg::make_query(
            "WITH d AS "
            "("
                "SELECT w.id AS worker_id, "
                       "s.domain_id, "
                       "i.nameserver_id "
                  "FROM worker w "
                  "JOIN scan_batch_progress p ON p.worker_id = w.id "
                  "JOIN scan_job j ON j.scan_batch_id = p.scan_batch_id "
                  "JOIN scan s ON s.id = j.scan_id "
                  "JOIN insecure_scan i ON i.scan_id = s.id "
                  "JOIN nameserver n ON n.id = i.nameserver_id "
                 "WHERE w.removed_at IS NULL AND "
                       "p.scan_batch_id = ") << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                       "j.id = " << LibPg::parameter<ScanJobId>().as_big_int() << " AND "
                       "n.fqdn = " << LibPg::parameter<Nameserver>().as_text() <<
            ") "
            "INSERT INTO scan_result "
            "("
                "scan_job_id, "
                "worker_id, "
                "nameserver_id, "
                "scanned_at, "
                "scan_status"
            ") SELECT " << LibPg::parameter<ScanJobId>().as_big_int() << ", "
                     "d.worker_id, "
                     "d.nameserver_id, "
                     "q.import_started_at + '10SECONDS'::INTERVAL, "
                        << LibPg::parameter<const char*>().as_type("cdnskey_status") << " "
                "FROM d "
          "CROSS JOIN scan_queue q "
               "WHERE q.id = " << LibPg::parameter<ScanQueueId>().as_big_int() << " AND "
                     "q.import_started_at IS NOT NULL "
              "RETURNING" << LibPg::item<std::int64_t>() << "id",
            ScanBatchId{static_cast<std::int64_t>(set_result.batch.id)},
            ScanJobId{static_cast<std::int64_t>(set_result.job_id)},
            Nameserver{nameserver},
            ScanQueueId{static_cast<std::uint64_t>(set_result.queue.id)},
            cdnskey_status);
    if (dbres.empty())
    {
        throw std::runtime_error{"no job joinable with this result"};
    }
    set_result.set_times();
}

void SetInsecureResult::operator()(const InsecureItem::InsecureEmptyResult& result) const
{
    store_result(set_result, nameserver, result.ip, "insecure-empty");
}

void SetInsecureResult::operator()(const InsecureItem::UnresolvedResult& result) const
{
    store_result(set_result, nameserver, result.ip, "unresolved");
}

void SetInsecureResult::operator()(const InsecureItem::UnresolvedIpResult&) const
{
    store_result(set_result, nameserver, "unresolved-ip");
}

void SetSecureResult::operator()(const SecureItem::SecureResult& result) const
{
    std::for_each(begin(result), end(result), [this](auto&& dns_key)
    {
        const auto dbres = exec(set_result.tx, LibPg::make_query(
                "WITH d AS "
                "("
                    "SELECT w.id AS worker_id, "
                        "s.domain_id "
                    "FROM worker w "
                    "JOIN scan_batch_progress p ON p.worker_id = w.id "
                    "JOIN scan_job j ON j.scan_batch_id = p.scan_batch_id "
                    "JOIN scan s ON s.id = j.scan_id "
                    "WHERE w.removed_at IS NULL AND "
                        "p.scan_batch_id = ") << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                        "j.id = " << LibPg::parameter<ScanJobId>().as_big_int() <<
                ") "
                "INSERT INTO scan_result "
                "("
                    "scan_job_id, "
                    "worker_id, "
                    "scanned_at, "
                    "scan_status, "
                    "cdnskey_flags, "
                    "cdnskey_proto, "
                    "cdnskey_alg, "
                    "cdnskey_public_key"
                ") SELECT " << LibPg::parameter<ScanJobId>().as_big_int() << ", "
                        "d.worker_id, "
                        "q.import_started_at + '10SECONDS'::INTERVAL, "
                        "'secure'::cdnskey_status, "
                            << LibPg::parameter<CdnskeyFlags>().as_integer() << ", "
                            << LibPg::parameter<CdnskeyProto>().as_integer() << ", "
                            << LibPg::parameter<CdnskeyAlg>().as_integer() << ", "
                            << LibPg::parameter<CdnskeyPublicKey>().as_text() << " "
                    "FROM d "
              "CROSS JOIN scan_queue q "
                   "WHERE q.id = " << LibPg::parameter<ScanQueueId>().as_big_int() << " AND "
                         "q.import_started_at IS NOT NULL "
                  "RETURNING" << LibPg::item<std::int64_t>() << "id",
                ScanBatchId{static_cast<std::int64_t>(set_result.batch.id)},
                ScanJobId{static_cast<std::int64_t>(set_result.job_id)},
                ScanQueueId{static_cast<std::uint64_t>(set_result.queue.id)},
                CdnskeyFlags{dns_key.flags},
                CdnskeyProto{dns_key.proto},
                CdnskeyAlg{dns_key.alg},
                CdnskeyPublicKey{dns_key.key});
        if (dbres.empty())
        {
            throw std::runtime_error{"no job joinable with this result"};
        }
    });
}

void store_result(const SetResult& set_result, const char* cdnskey_status)
{
    const auto dbres = exec(set_result.tx, LibPg::make_query(
            "WITH d AS "
            "("
                "SELECT w.id AS worker_id, "
                       "s.domain_id "
                  "FROM worker w "
                  "JOIN scan_batch_progress p ON p.worker_id = w.id "
                  "JOIN scan_job j ON j.scan_batch_id = p.scan_batch_id "
                  "JOIN scan s ON s.id = j.scan_id "
                 "WHERE w.removed_at IS NULL AND "
                       "p.scan_batch_id = ") << LibPg::parameter<ScanBatchId>().as_big_int() << " AND "
                       "j.id = " << LibPg::parameter<ScanJobId>().as_big_int() <<
            ") "
            "INSERT INTO scan_result "
            "("
                "scan_job_id, "
                "worker_id, "
                "scanned_at, "
                "scan_status"
            ") SELECT " << LibPg::parameter<ScanJobId>().as_big_int() << ", "
                     "d.worker_id, "
                     "q.import_started_at + '10SECONDS'::INTERVAL, "
                        << LibPg::parameter<const char*>().as_type("cdnskey_status") << " "
                "FROM d "
          "CROSS JOIN scan_queue q "
               "WHERE q.id = " << LibPg::parameter<ScanQueueId>().as_big_int() << " AND "
                     "q.import_started_at IS NOT NULL "
              "RETURNING" << LibPg::item<std::int64_t>() << "id",
            ScanBatchId{static_cast<std::int64_t>(set_result.batch.id)},
            ScanJobId{static_cast<std::int64_t>(set_result.job_id)},
            ScanQueueId{static_cast<std::uint64_t>(set_result.queue.id)},
            cdnskey_status);
    if (dbres.empty())
    {
        throw std::runtime_error{"no job joinable with this result"};
    }
}

void SetSecureResult::operator()(const SecureItem::SecureEmptyResult&) const
{
    store_result(set_result, "secure-empty");
}

void SetSecureResult::operator()(const SecureItem::UntrustworthyResult&) const
{
    store_result(set_result, "untrustworthy");
}

void SetSecureResult::operator()(const SecureItem::UnknownResult&) const
{
    store_result(set_result, "unknown");
}

void set_result(const LibPg::PgRwTransaction& tx, const Queue& queue, const Batch& batch, const Job& job)
{
    boost::apply_visitor(SetResult{tx, queue, batch, job.id}, job.item);
}

void set_result(const LibPg::PgRwTransaction& tx, const Queue& queue, const Batch& batch)
{
    std::for_each(begin(batch.jobs), end(batch.jobs), [&](auto&& job)
    {
        set_result(tx, queue, batch, job);
    });
}

void set_result(const LibPg::PgRwTransaction& tx, const Queue& queue)
{
    std::for_each(begin(queue.batches), end(queue.batches), [&tx, &queue](auto&& batch)
    {
        set_result(tx, queue, batch);
    });
}

}//namespace CdnskeyProcessor::DbLoader::{anonymous}

}//namespace CdnskeyProcessor::DbLoader
}//namespace CdnskeyProcessor

using namespace CdnskeyProcessor::DbLoader;

int CdnskeyProcessor::DbLoader::import(std::istream& in)
{
    auto queues = [&in]()
    {
        nlohmann::json json_data;
        in >> json_data;
        return get_queues(json_data);
    }();
    auto tx = CdnskeyProcessor::make_rw_transaction();
    std::for_each(begin(queues), end(queues), [&tx](auto&& queue)
    {
        set_queue(tx, queue);
    });
    std::for_each(begin(queues), end(queues), [&tx](auto&& queue)
    {
        set_result(tx, queue);
    });
    clean_old_records(tx, MaxScanPeriod{std::chrono::hours{2 * 24}}, MinPeriodFixedCdnskey{std::chrono::hours{7 * 24}});
    auto conn = commit(std::move(tx));
    vacuum(conn);
    return EXIT_SUCCESS;
}
