/*
 * Copyright (C) 2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "db_loader/cfg/options.hh"
#include "db_loader/import.hh"
#include "src/database.hh"

#include "liblog/liblog.hh"
#include "liblog/log.hh"

#include <nlohmann/json.hpp>

#include <cstdlib>
#include <iostream>

namespace {

auto severity_to_level(CdnskeyProcessor::DbLoader::Cfg::LogSeverity severity)
{
    using namespace CdnskeyProcessor::DbLoader;
    switch (severity)
    {
        case Cfg::LogSeverity::critical:
            return LibLog::Level::critical;
        case Cfg::LogSeverity::error:
            return LibLog::Level::error;
        case Cfg::LogSeverity::warning:
            return LibLog::Level::warning;
        case Cfg::LogSeverity::info:
            return LibLog::Level::info;
        case Cfg::LogSeverity::debug:
            return LibLog::Level::debug;
        case Cfg::LogSeverity::trace:
            return LibLog::Level::trace;
    }
    return LibLog::Level::trace;
}

auto setup_logging(const CdnskeyProcessor::DbLoader::Cfg::Options::Log& log_options)
{
    LibLog::LogConfig log_config;
    struct Result
    {
        bool log_enabled = false;
    } result;
    if (log_options.console != boost::none)
    {
        LibLog::Sink::ConsoleSinkConfig config;
        const auto& option = *log_options.console;
        config.set_level(severity_to_level(option.min_severity))
            .set_output_stream(LibLog::Sink::ConsoleSinkConfig::OutputStream::stderr)
            .set_color_mode(LibLog::ColorMode::automatic);
        log_config.add_sink_config(config);
        result.log_enabled = true;
    }
    if (log_options.file != boost::none)
    {
        const auto& option = *log_options.file;
        LibLog::Sink::FileSinkConfig config{option.file_name};
        config.set_level(severity_to_level(option.min_severity));
        log_config.add_sink_config(config);
        result.log_enabled = true;
    }
    if (log_options.syslog != boost::none)
    {
        const auto& option = *log_options.syslog;
        LibLog::Sink::SyslogSinkConfig config;
        if (!option.ident.empty())
        {
            config.set_syslog_ident(option.ident);
        }
        config.set_level(severity_to_level(option.min_severity));
        config.set_syslog_facility(option.facility);
        log_config.add_sink_config(config);
        result.log_enabled = true;
    }
    if (result.log_enabled)
    {
        LibLog::Log::start<LibLog::ThreadMode::multi_threaded>(log_config);
    }
    return result;
}

void setup_database(const CdnskeyProcessor::DbLoader::Cfg::Options::Database& cfg)
{
    CdnskeyProcessor::set_up_database({
            *cfg.host,
            {},
            cfg.port,
            *cfg.dbname,
            *cfg.user,
            *cfg.password,
            {}
    });
}

struct Command
{
    using Options = CdnskeyProcessor::DbLoader::Cfg::Options;
    static int import(const Options::Import&)
    {
        LIBLOG_SET_CONTEXT(LogCtx, log_ctx, __func__);
        LIBLOG_TRACE("enter");
        const auto result = CdnskeyProcessor::DbLoader::import(std::cin);
        LIBLOG_TRACE("leave");
        return result;
    }
};

}//namespace {anonymous}

int main(int argc, char* argv[])
{
    using namespace CdnskeyProcessor::DbLoader;
    try
    {
        Cfg::Options::init(argc, argv);
    }
    catch (const Cfg::AllDone& e)
    {
        std::cout << e.what() << std::endl;
        return EXIT_SUCCESS;
    }
    catch (const Cfg::Exception& e)
    {
        std::cerr << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (const std::exception& e)
    {
        std::cerr << "Exception caught: " << e.what() << std::endl;
        return EXIT_FAILURE;
    }
    catch (...)
    {
        std::cerr << "Unexpected exception caught" << std::endl;
        return EXIT_FAILURE;
    }

    const auto log_enabled = setup_logging(Cfg::Options::get().log).log_enabled;
    if (log_enabled)
    {
        LIBLOG_DEBUG("cdnskey-processor-db-loader configuration successfully loaded");
    }
    else
    {
        std::cout << "cdnskey-processor-db-loader configuration successfully loaded" << std::endl;
    }
    try
    {
        setup_database(Cfg::Options::get().database);
        return run(Cfg::Options::get().command, Command::import);
    }
    catch (const std::exception& e)
    {
        if (log_enabled)
        {
            LIBLOG_ERROR("Exception caught: {}", e.what());
        }
        else
        {
            std::cerr << "Exception caught: " << e.what() << std::endl;
        }
    }
    catch (...)
    {
        if (log_enabled)
        {
            LIBLOG_ERROR("Unexpected exception caught");
        }
        else
        {
            std::cerr << "Unexpected exception caught" << std::endl;
        }
    }
    return EXIT_FAILURE;
}
