/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "result_importer/pair_scan_jobs_with_scan_results.hh"

namespace CdnskeyProcessor {
namespace Test {

namespace {

void add_paired_scan_results(
        std::vector<ScanResultCollection>& _scan_result_collections,
        ScanResultsPerWorker& _output_scan_results,
        const ScanBatchId& _batch_id)
{
    for (auto& scan_result_per_worker : _output_scan_results)
    {
        auto& scan_result_collection = scan_result_per_worker.second;
        scan_result_collection.scan_batch_id = _batch_id;
        scan_result_collection.worker_name = scan_result_per_worker.first;
        scan_result_collection.sent_at = scan_result_per_worker.second.finished_at.get();
        _scan_result_collections.emplace_back(scan_result_per_worker.second);
    }
}

void pair_scan_job_with_scan_result(
        const ScanJobId& _scan_job_id,
        ScanResultsPerWorker& _input_results_for_domain,
        ScanResultsPerWorker& _output_scan_results)
{
    // looking for each worker's results for scan_job
    for (auto& input_results_for_worker : _input_results_for_domain)
    {
        auto emplace_result = _output_scan_results.emplace(input_results_for_worker.first, ScanResultCollection{});
        ScanResultCollection& output_scan_results = emplace_result.first->second;
        auto& input_scan_results = input_results_for_worker.second;
        if (emplace_result.second) // emplacement happened
        {
            input_scan_results.worker_name = input_results_for_worker.first;
            output_scan_results.worker_name = input_results_for_worker.first;
            output_scan_results.start_at = Timestamp{Timestamp::UnderlyingType::max()};
            output_scan_results.finished_at = Timestamp{Timestamp::UnderlyingType::min()};
        }


        if (input_scan_results.start_at < output_scan_results.start_at)
        {
            output_scan_results.start_at = input_scan_results.start_at;
        }
        if (input_scan_results.finished_at > output_scan_results.finished_at)
        {
            output_scan_results.sent_at = input_scan_results.sent_at;
            output_scan_results.finished_at = input_scan_results.finished_at;
        }

        output_scan_results.scan_results.reserve(
                output_scan_results.scan_results.size() + input_scan_results.scan_results.size());
        for (auto& scan_result : input_scan_results)
        {
            scan_result.scan_job_id = _scan_job_id;
            output_scan_results.scan_results.emplace_back(scan_result);
        }
    }
}

std::vector<ScanResultCollection> pair_scan_jobs_with_scan_results(
        const ScanBatches<SecureScanTaskCollection>& _batches,
        ScanResultsPerDomain& _results)
{
    std::vector<ScanResultCollection> scan_result_collection;
    for (const auto& batch_pair : _batches)
    {
        auto& batch = batch_pair.second;
        // for this batch id
        ScanResultsPerWorker output_scan_results;
        for (const auto& secure_scan_task : batch.scan_tasks)
        {
            const auto& scan_job = secure_scan_task.scan_job;
            auto& input_results_for_domain = _results[scan_job.domain];
            pair_scan_job_with_scan_result(scan_job.id, input_results_for_domain, output_scan_results);
        }

        add_paired_scan_results(scan_result_collection, output_scan_results, batch.id);
    }
    return scan_result_collection;
}

std::vector<ScanResultCollection> pair_scan_jobs_with_scan_results(
        const ScanBatches<InsecureScanTaskCollection>& _batches,
        InsecureScanResults& _results)
{
    std::vector<ScanResultCollection> scan_result_collection;
    for (const auto& batch_pair : _batches)
    {
        auto& batch = batch_pair.second;
        // for this batch id
        ScanResultsPerWorker output_scan_results;
        for (const auto& insecure_scan_task : batch.scan_tasks)
        {
            // for nameserver's scan_jobs
            auto& input_results_for_nameserver = _results[insecure_scan_task.nameserver];
            for (const auto& scan_job_pair : insecure_scan_task.scan_jobs)
            {
                const auto& scan_job = scan_job_pair.second;
                // looking for each worker's results for scan_job
                auto& input_results_for_domain = input_results_for_nameserver[scan_job.domain];
                pair_scan_job_with_scan_result(scan_job.id, input_results_for_domain, output_scan_results);
            }
        }

        add_paired_scan_results(scan_result_collection, output_scan_results, batch.id);
    }
    return scan_result_collection;
}

} // namespace CdnskeyProcessor::Test::{anonymous}

std::vector<ScanResultCollection> pair_scan_jobs_with_scan_results(
        const UnprocessedScanBatches& _scan_batches,
        ScanQueue& _scan_queue)
{
    auto all_results = pair_scan_jobs_with_scan_results(_scan_batches.insecure, _scan_queue.get_insecure_scan_results());
    auto secure_results = pair_scan_jobs_with_scan_results(_scan_batches.secure, _scan_queue.get_secure_scan_results());
    all_results.insert(all_results.end(), secure_results.cbegin(), secure_results.cend());
    return all_results;
}

} // namespace CdnskeyProcessor::Test
} // namespace CdnskeyProcessor
