/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "result_importer/log/log.hh"
#include "result_importer/read_input.hh"

#include <src/util/to_string.hh>

#include <nlohmann/json.hpp>

namespace CdnskeyProcessor {
namespace Test {

std::string to_string(const Domain& _domain)
{
    return  "domain " + _domain.fqdn.get()
         + " uid: " + _domain.uid.get()
         + " scan_type: " + to_string(_domain.scan_type);
}

Timestamp get_timestamp_from_time_delta(const int64_t& _input)
{
    auto timestamp_now = now();
    return Timestamp{timestamp_now.get() - std::chrono::minutes(_input)};
}

ScanResult read_scan_result_from_input(const nlohmann::json& _json_data)
{
    boost::optional<NameserverIp> nameserver_ip;
    if (not _json_data["ip"].is_null())
    {
        nameserver_ip = from_string<NameserverIp>(_json_data["ip"]);
    }

    auto& json_cdnskey = _json_data["cdnskey"];
    Cdnskey cdnskey (from_string<CdnskeyStatus>(json_cdnskey["status"]));
    if (not json_cdnskey["key"].is_null())
    {
        cdnskey.alg = CdnskeyAlg{json_cdnskey["alg"]};
        cdnskey.flags = CdnskeyFlags{json_cdnskey["flag"]};
        cdnskey.proto = CdnskeyProto{json_cdnskey["proto"]};
        cdnskey.public_key = CdnskeyPublicKey{json_cdnskey["key"]};
    }

    ScanResult scan_result {
            ScanJobId{0},
            get_timestamp_from_time_delta(_json_data["time_delta"]),
            nameserver_ip,
            cdnskey,
            boost::none,
            boost::none};
    return scan_result;
}

void read_input_for_domain(ScanQueues& _scan_queues, const nlohmann::json& _json_data)
{
    Domain domain(
            DomainUid{_json_data["domain_uuid"]},
            DomainFqdn{_json_data["domain_fqdn"]},
            from_string<ScanType>(_json_data["scan_type"]));

    for (const auto& json_nameserver_response : _json_data["nameservers"].items())
    {
        boost::optional<Nameserver> nameserver;
        Nameservers nameservers;

        CP_TEST_SET_CONTEXT(Context, log_ctx, "[read input] fqdn:", domain.fqdn.get() + ",");

        if (domain.scan_type == ScanType::insecure)
        {
            if (json_nameserver_response.value()["name"].is_null())
            {
                const std::string message = "no nameserver name present for insecure domain";
                LIBLOG_ERROR(message);
                throw InvalidInput{message};
            }
            nameserver = Nameserver{json_nameserver_response.value()["name"]};
            nameservers.insert(nameserver.get());
            log_ctx.ctx.add({"ns:", nameserver.get().get() + ","});
        }

        for (const auto& json_iteration : json_nameserver_response.value()["iterations"].items())
        {
            ScanQueueId scan_queue_id{json_iteration.value()["iter_id"]};
            CP_TEST_SET_CONTEXT(Context, ctx, "iter: " + to_string(scan_queue_id));

            auto& scan_queue = _scan_queues[scan_queue_id];
            scan_queue.set_id(scan_queue_id);
            scan_queue.add_scan_task(DomainScanTask(domain, nameservers));

            for (const auto& json_result : json_iteration.value()["results"].items())
            {
                auto scan_result = read_scan_result_from_input(json_result.value());
                scan_result.nameserver = nameserver;
                scan_result.domain_fqdn = domain.fqdn;
                log_scan_result(scan_result);

                WorkerName worker{json_result.value()["worker"]};
                scan_queue.add_scan_result(domain, worker, scan_result);
            }
        }
    }
}

ScanQueues read_input(std::ifstream& is)
{
    using json = nlohmann::json;

    json json_data;
    is >> json_data;

    ScanQueues scan_queues;
    std::for_each(json_data.begin(), json_data.end(),
            [&scan_queues](const json& item) { read_input_for_domain(scan_queues, item); });

    return scan_queues;
}

} // namespace CdnskeyProcessor::Test
} // namespace CdnskeyProcessor
