/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "result_importer/scan_queue.hh"
#include "result_importer/log/log.hh"

#include <src/util/to_string.hh>

#include <liblog/liblog.hh>

namespace CdnskeyProcessor {
namespace Test {

ScanQueue::ScanQueue()
    : batch_size(0),
      start_at(Timestamp::UnderlyingType::max()),
      finished_at(Timestamp::UnderlyingType::min())
{
}

ScanQueueId ScanQueue::get_id() const
{
    return id;
}

void ScanQueue::set_id(const ScanQueueId& _id)
{
    id = _id;
}

Timestamp ScanQueue::get_start_at() const
{
    return start_at;
}

Timestamp ScanQueue::get_finished_at() const
{
    return finished_at;
}

Interface::ScanBatchSize ScanQueue::get_scan_batch_size() const
{
    return batch_size;
}

const DomainScanTaskCollection& ScanQueue::get_scan_tasks() const
{
    return tasks;
}

void ScanQueue::add_scan_task(const DomainScanTask& _task)
{
    if (_task.nameservers.empty())
    {
        batch_size = Interface::ScanBatchSize{batch_size.get() + 1};
    }
    else
    {
        batch_size = Interface::ScanBatchSize{batch_size.get() + _task.nameservers.size()};
    }
    tasks.insert_or_update(_task);
}

InsecureScanResults& ScanQueue::get_insecure_scan_results()
{
    return insecure_results;
}

SecureScanResults& ScanQueue::get_secure_scan_results()
{
    return secure_results;
}

namespace {

ScanResultsPerDomain& scan_queue_results_for_scan_type(
        ScanQueue& _scan_queue,
        const ScanType& _scan_type,
        const boost::optional<Nameserver>& _nameserver)
{
    switch (_scan_type)
    {
        case ScanType::insecure:
            return _scan_queue.get_insecure_scan_results()[_nameserver.get()];
        case ScanType::secure:
            return _scan_queue.get_secure_scan_results();
    }
    throw InvalidScanType{};
}

} // namespace CdnskeyProcessor::Test::{anonymous}

void ScanQueue::add_scan_result(
        const Domain& _domain,
        const WorkerName& _worker,
        const ScanResult& _scan_result)
{
    auto& scan_results = scan_queue_results_for_scan_type(*this, _domain.scan_type, _scan_result.nameserver);
    auto& scan_result_collection = scan_results[_domain][_worker];
    scan_result_collection.scan_results.push_back(_scan_result);

    if (not scan_result_collection.finished_at or (_scan_result.scan_at > scan_result_collection.finished_at))
    {
        scan_result_collection.sent_at = _scan_result.scan_at;
        scan_result_collection.finished_at = _scan_result.scan_at;
    }
    if (not scan_result_collection.start_at or (_scan_result.scan_at < scan_result_collection.start_at))
    {
        scan_result_collection.start_at = _scan_result.scan_at;
    }

    if (_scan_result.scan_at > finished_at)
    {
        finished_at = _scan_result.scan_at;
    }
    if (_scan_result.scan_at < start_at)
    {
        start_at = _scan_result.scan_at;
    }
}

void log_scan_result(const ScanResult& _scan_result)
{
    std::string nameserver_ip;
    if (_scan_result.nameserver_ip)
    {
        nameserver_ip = to_string(_scan_result.nameserver_ip.get());
    }
    else
    {
        nameserver_ip = "<none>";
    }
    std::string nameserver;
    if (_scan_result.nameserver)
    {
        nameserver = to_string(_scan_result.nameserver.get());
    }
    else
    {
        nameserver = "<none>";
    }
    std::string domain_fqdn;
    if (_scan_result.domain_fqdn)
    {
        domain_fqdn = to_string(_scan_result.domain_fqdn.get());
    }
    else
    {
        domain_fqdn = "<none>";
    }

    std::string log_cdnskey = "[ status: " + to_string(_scan_result.cdnskey.status);

    if (_scan_result.cdnskey.public_key)
    {
        log_cdnskey +=
                ", flags: " + to_string(_scan_result.cdnskey.flags.get()) +
                ", proto: " + to_string(_scan_result.cdnskey.proto.get()) +
                ", alg: "   + to_string(_scan_result.cdnskey.alg.get()) +
                ", pk: "    + to_string(_scan_result.cdnskey.public_key.get());
    }
    log_cdnskey += " ]";

    CP_TEST_INFO("domain: {}, nameserver: {}, nameserver-ip: {}, scan_job_id: {}, scan_at: {}, cdnskey: {}",
            domain_fqdn,
            nameserver,
            nameserver_ip,
            _scan_result.scan_job_id.get(),
            to_string(_scan_result.scan_at),
            log_cdnskey);

}

void log_scan_result_collection(const ScanResultCollection& _scan_result_collection)
{
    CP_TEST_INFO("scan result collection [scan_batch_id: {}, worker: {}, sent_at: {}, start_at: {}, finished_at: {}]",
            to_string(_scan_result_collection.scan_batch_id),
            _scan_result_collection.worker_name.get(),
            to_string(_scan_result_collection.sent_at),
            to_string(_scan_result_collection.start_at.get()),
            to_string(_scan_result_collection.finished_at.get()));
    std::for_each(
            _scan_result_collection.begin(),
            _scan_result_collection.end(),
            log_scan_result);
}

namespace {

using ScanResultsPerWorkerPair = ScanResultsPerWorker::value_type;
using ScanResultsPerDomainPair = ScanResultsPerDomain::value_type;
using InsecureScanResultsPair = InsecureScanResults::value_type;
using ScanQueuesPair = std::pair<ScanQueueId, ScanQueue>;

void log_domain_scan_task(const DomainScanTask& _task)
{
    std::string nameservers;
    if (not _task.nameservers.empty())
    {
        auto nameserver_it = _task.nameservers.begin();
        nameservers += nameserver_it->get();
        ++nameserver_it;
        for (; nameserver_it != _task.nameservers.end() ; ++nameserver_it)
        {
            nameservers += ", " + nameserver_it->get();
        }
        CP_TEST_INFO("domain: {}, scan_type: {}, uid: {}, nameservers: {}",
                _task.domain.fqdn.get(),
                to_string(_task.domain.scan_type),
                _task.domain.uid.get(),
                nameservers);
    }
    else
    {
        CP_TEST_INFO("domain: {}, scan_type: {}, uid: {}",
                _task.domain.fqdn.get(),
                to_string(_task.domain.scan_type),
                _task.domain.uid.get());
    }
}

void log_scan_results_per_worker(const ScanResultsPerWorkerPair& _scan_result_per_worker_pair)
{
    CP_TEST_SET_CONTEXT(Context, ctx, "worker:", _scan_result_per_worker_pair.first.get());
    log_scan_result_collection(_scan_result_per_worker_pair.second);
}

void log_scan_results_per_domain_pair(const ScanResultsPerDomainPair& _scan_results_per_domain)
{
    std::for_each(
            _scan_results_per_domain.second.begin(),
            _scan_results_per_domain.second.end(),
            log_scan_results_per_worker);
}

void log_scan_results_per_domain(const ScanResultsPerDomain& _scan_results_per_domain)
{
    std::for_each(
            _scan_results_per_domain.begin(),
            _scan_results_per_domain.end(),
            log_scan_results_per_domain_pair);
}

void log_insecure_scan_results(const InsecureScanResultsPair& _nameserver_scan_result_per_domain_pair)
{
    CP_TEST_SET_CONTEXT(Context, ctx, "type: insecure,");
    log_scan_results_per_domain(_nameserver_scan_result_per_domain_pair.second);
}

} // namespace CdnskeyProcessor::Test::{anonymous}

void ScanQueue::log() const
{
    std::for_each(tasks.begin(), tasks.end(), log_domain_scan_task);
    std::for_each(insecure_results.begin(), insecure_results.end(), log_insecure_scan_results);
    log_scan_results_per_domain(secure_results);
}

void log_scan_queues(const ScanQueues& _scan_queues)
{
    std::for_each(_scan_queues.begin(), _scan_queues.end(),
            [](const ScanQueuesPair& _scan_queue_pair)
            {
                CP_TEST_SET_CONTEXT(Context, ctx, "[scan queue "
                        "file:", to_string(_scan_queue_pair.first) + ",",
                        "imported:", to_string(_scan_queue_pair.second.get_id()) + "]");
                _scan_queue_pair.second.log();
            });
}

} // namespace CdnskeyProcessor::Test
} // namespace CdnskeyProcessor
