/*
 * Copyright (C) 2021-2024  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "result_importer/log/log.hh"
#include "result_importer/pair_scan_jobs_with_scan_results.hh"
#include "result_importer/storage.hh"

#include <src/api/storage_supervisor.hh>
#include <src/internal/master/storage.hh>
#include <src/storage.hh>
#include <src/util/to_string.hh>

#include <libpg/pg_exception.hh>

namespace CdnskeyProcessor {
namespace Test {
namespace Storage {

namespace {

using Transaction = CdnskeyProcessor::Storage::Transaction;

void start_scan_queue_with_a_timestamp(
        const Transaction& _transaction,
        const ScanQueueId& _scan_queue_id,
        const Timestamp& _timestamp)
{
    const auto query = LibPg::make_query() <<
            // clang-format off
            "UPDATE scan_queue "
               "SET start_at = " << LibPg::parameter<Timestamp>().as_timestamp_with_time_zone() << " "
             "WHERE id = " << LibPg::parameter<ScanQueueId>().as_big_int();
            // clang-format on
    const auto cmdres = exec(_transaction, query, _scan_queue_id, _timestamp);
}

void finish_scan_queue_with_a_timestamp(
        const Transaction& _transaction,
        const ScanQueueId& _scan_queue_id,
        const Timestamp& _timestamp)
{
    const auto query = LibPg::make_query() <<
            // clang-format off
            "UPDATE scan_queue "
               "SET finished_at = " << LibPg::parameter<Timestamp>().as_timestamp_with_time_zone() << " "
             "WHERE id = " << LibPg::parameter<ScanQueueId>().as_big_int();
            // clang-format on
    const auto cmdres = exec(_transaction, query, _scan_queue_id, _timestamp);
}

} // namespace CdnskeyProcessor::Test::{anonymous}

void import_scan_tasks(const LibPg::Dsn& _dsn, const ScanQueues& _scan_queues)
try
{
    CP_TEST_INFO("importing scan tasks:");
    auto transaction = CdnskeyProcessor::Storage::make_transaction(_dsn);
    std::for_each(_scan_queues.begin(), _scan_queues.end(),
            [&_dsn, &transaction](const ScanQueues::value_type& _scan_queue_pair)
            {
                const auto& scan_queue = _scan_queue_pair.second;
                CP_TEST_INFO("importing scan task for queue file: {}, task count: {}",
                        to_string(scan_queue.get_id()),
                        to_string(scan_queue.get_scan_batch_size()));
                const auto imported_scan_queue_id = Interface::Supervisor::Storage::create_scan_queue_from_scan_tasks(
                        transaction,
                        scan_queue.get_scan_tasks(),
                        scan_queue.get_scan_batch_size());
                CP_TEST_INFO("imported scan queue id [file: {}, imported: {}]",
                        to_string(scan_queue.get_id()),
                        to_string(imported_scan_queue_id));
            });
    commit(std::move(transaction));
}
catch (const LibPg::PgException& e)
{
    throw DatabaseError(std::string("at scan task import: ") + e.what());
}

void import_scan_results(const LibPg::Dsn& _dsn, ScanQueues& _scan_queues)
try
{
    CP_TEST_INFO("importing scan results:");
    auto transaction = CdnskeyProcessor::Storage::make_transaction(_dsn);
    Internal::Master::Storage::purge_scan_queues(transaction);
    for (auto& scan_queue_pair : _scan_queues)
    {
        auto& scan_queue = scan_queue_pair.second;

        CP_TEST_INFO("[scan queue file: {}]", to_string(scan_queue_pair.first));

        const auto optional_scan_queue_id = Internal::Master::Storage::prepare_new_scan_queue(transaction);
        if (not optional_scan_queue_id)
        {
            throw ImportedScanQueueMissing{};
        }
        scan_queue.set_id(optional_scan_queue_id.get());
        CP_TEST_INFO("[scan queue imported: {}]", to_string(scan_queue.get_id()));

        CP_TEST_SET_CONTEXT(Context, ctx, "[scan queue",
                "file:", to_string(scan_queue_pair.first) + ",",
                "imported:", to_string(scan_queue.get_id()) + "]");

        start_scan_queue_with_a_timestamp(transaction, scan_queue.get_id(), scan_queue.get_start_at());
        auto scan_batches = Internal::Master::Storage::get_unprocessed_scan_batches(transaction, scan_queue.get_id());
        if (scan_batches.insecure.empty() and scan_batches.secure.empty())
        {
            throw ImportedScanBatchesAreEmpty{};
        }

        const auto scan_results = Test::pair_scan_jobs_with_scan_results(scan_batches, scan_queue);
        std::for_each(scan_results.cbegin(), scan_results.cend(),
                [&transaction](const ScanResultCollection& _scan_results)
                {
                    log_scan_result_collection(_scan_results);
                    Internal::Master::Storage::insert_scan_results(transaction, _scan_results);
                });
        finish_scan_queue_with_a_timestamp(transaction, scan_queue.get_id(), scan_queue.get_finished_at());
    }
    commit(std::move(transaction));
}
catch(const LibPg::PgException& e)
{
    throw DatabaseError(std::string("at scan result import: ") + e.what());
}

} // namespace CdnskeyProcessor::Test::Storage
} // namespace CdnskeyProcessor::Test
} // namespace CdnskeyProcessor
